@@ -7,12 +7,11 @@
 #include "nsDOMNavigationTiming.h"
 
 #include "GeckoProfiler.h"
-#include "ipc/IPCMessageUtilsSpecializations.h"
 #include "mozilla/ProfilerMarkers.h"
 #include "mozilla/Telemetry.h"
 #include "mozilla/TimeStamp.h"
+#include "mozilla/glean/GleanMetrics.h"
 #include "mozilla/dom/Document.h"
-#include "mozilla/dom/PerformanceNavigation.h"
 #include "mozilla/ipc/IPDLParamTraits.h"
 #include "mozilla/ipc/URIUtils.h"
 #include "nsCOMPtr.h"
@@ -56,7 +55,8 @@ void nsDOMNavigationTiming::Clear() {
   mDOMContentLoadedEventStart = TimeStamp();
   mDOMContentLoadedEventEnd = TimeStamp();
   mDOMComplete = TimeStamp();
-  mContentfulPaint = TimeStamp();
+  mContentfulComposite = TimeStamp();
+  mLargestContentfulRender = TimeStamp();
   mNonBlankPaint = TimeStamp();
 
   mDocShellHasBeenActiveSinceNavigationStart = false;
@@ -140,8 +140,8 @@ void nsDOMNavigationTiming::NotifyLoadEventStart() {
   if (IsTopLevelContentDocumentInContentProcess()) {
     TimeStamp now = TimeStamp::Now();
 
-    Telemetry::AccumulateTimeDelta(Telemetry::TIME_TO_LOAD_EVENT_START_MS,
-                                   mNavigationStart, now);
+    glean::performance_time::load_event_start.AccumulateRawDuration(
+        now - mNavigationStart);
 
     if (mDocShellHasBeenActiveSinceNavigationStart) {
       if (net::nsHttp::IsBeforeLastActiveTabLoadOptimization(
@@ -170,15 +170,13 @@ void nsDOMNavigationTiming::NotifyLoadEventEnd() {
                   Tracing, "Navigation");
 
   if (IsTopLevelContentDocumentInContentProcess()) {
-    if (profiler_can_accept_markers() || PAGELOAD_LOG_ENABLED()) {
+    if (profiler_thread_is_being_profiled_for_markers() ||
+        PAGELOAD_LOG_ENABLED()) {
       TimeDuration elapsed = mLoadEventEnd - mNavigationStart;
       TimeDuration duration = mLoadEventEnd - mLoadEventStart;
-      nsAutoCString spec;
-      if (mLoadedURI) {
-        mLoadedURI->GetSpec(spec);
-      }
       nsPrintfCString marker(
-          "Document %s loaded after %dms, load event duration %dms", spec.get(),
+          "Document %s loaded after %dms, load event duration %dms",
+          nsContentUtils::TruncatedURLForDisplay(mLoadedURI).get(),
           int(elapsed.ToMilliseconds()), int(duration.ToMilliseconds()));
       PAGELOAD_LOG(("%s", marker.get()));
       PROFILER_MARKER_TEXT(
@@ -187,8 +185,8 @@ void nsDOMNavigationTiming::NotifyLoadEventEnd() {
                         MarkerInnerWindowIdFromDocShell(mDocShell)),
           marker);
     }
-    Telemetry::AccumulateTimeDelta(Telemetry::TIME_TO_LOAD_EVENT_END_MS,
-                                   mNavigationStart);
+    glean::performance_time::load_event_end.AccumulateRawDuration(
+        TimeStamp::Now() - mNavigationStart);
   }
 }
 
@@ -250,8 +248,8 @@ void nsDOMNavigationTiming::NotifyDOMContentLoadedStart(nsIURI* aURI) {
   if (IsTopLevelContentDocumentInContentProcess()) {
     TimeStamp now = TimeStamp::Now();
 
-    Telemetry::AccumulateTimeDelta(
-        Telemetry::TIME_TO_DOM_CONTENT_LOADED_START_MS, mNavigationStart, now);
+    glean::performance_time::dom_content_loaded_start.AccumulateRawDuration(
+        now - mNavigationStart);
 
     if (mDocShellHasBeenActiveSinceNavigationStart) {
       if (net::nsHttp::IsBeforeLastActiveTabLoadOptimization(
@@ -282,8 +280,8 @@ void nsDOMNavigationTiming::NotifyDOMContentLoadedEnd(nsIURI* aURI) {
                   Tracing, "Navigation");
 
   if (IsTopLevelContentDocumentInContentProcess()) {
-    Telemetry::AccumulateTimeDelta(Telemetry::TIME_TO_DOM_CONTENT_LOADED_END_MS,
-                                   mNavigationStart);
+    glean::performance_time::dom_content_loaded_end.AccumulateRawDuration(
+        TimeStamp::Now() - mNavigationStart);
   }
 }
 
@@ -299,20 +297,20 @@ void nsDOMNavigationTiming::TTITimeoutCallback(nsITimer* aTimer,
 void nsDOMNavigationTiming::TTITimeout(nsITimer* aTimer) {
   // Check TTI: see if it's been 5 seconds since the last Long Task
   TimeStamp now = TimeStamp::Now();
-  MOZ_RELEASE_ASSERT(!mContentfulPaint.IsNull(),
-                     "TTI timeout with no contentful-paint?");
+  MOZ_RELEASE_ASSERT(!mContentfulComposite.IsNull(),
+                     "TTI timeout with no contentful-composite?");
 
   nsCOMPtr<nsIThread> mainThread = do_GetMainThread();
   TimeStamp lastLongTaskEnded;
   mainThread->GetLastLongNonIdleTaskEnd(&lastLongTaskEnded);
-  // Window starts at mContentfulPaint; any long task before that is ignored
-  if (lastLongTaskEnded.IsNull() || lastLongTaskEnded < mContentfulPaint) {
+  // Window starts at mContentfulComposite; any long task before that is ignored
+  if (lastLongTaskEnded.IsNull() || lastLongTaskEnded < mContentfulComposite) {
     PAGELOAD_LOG(
-        ("no longtask (last was %g ms before ContentfulPaint)",
+        ("no longtask (last was %g ms before ContentfulComposite)",
          lastLongTaskEnded.IsNull()
              ? 0
-             : (mContentfulPaint - lastLongTaskEnded).ToMilliseconds()));
-    lastLongTaskEnded = mContentfulPaint;
+             : (mContentfulComposite - lastLongTaskEnded).ToMilliseconds()));
+    lastLongTaskEnded = mContentfulComposite;
   }
   TimeDuration delta = now - lastLongTaskEnded;
   PAGELOAD_LOG(("TTI delta: %g ms", delta.ToMilliseconds()));
@@ -346,7 +344,7 @@ void nsDOMNavigationTiming::TTITimeout(nsITimer* aTimer) {
   // is >= FCP here.
 
   if (mTTFI.IsNull()) {
-    // lastLongTaskEnded is >= mContentfulPaint
+    // lastLongTaskEnded is >= mContentfulComposite
     mTTFI = (mDOMContentLoadedEventEnd.IsNull() ||
              lastLongTaskEnded > mDOMContentLoadedEventEnd)
                 ? lastLongTaskEnded
@@ -368,18 +366,16 @@ void nsDOMNavigationTiming::TTITimeout(nsITimer* aTimer) {
 
   mTTITimer = nullptr;
 
-  if (profiler_can_accept_markers() || PAGELOAD_LOG_ENABLED()) {
+  if (profiler_thread_is_being_profiled_for_markers() ||
+      PAGELOAD_LOG_ENABLED()) {
     TimeDuration elapsed = mTTFI - mNavigationStart;
     MOZ_ASSERT(elapsed.ToMilliseconds() > 0);
     TimeDuration elapsedLongTask =
         lastLongTaskEnded.IsNull() ? 0 : lastLongTaskEnded - mNavigationStart;
-    nsAutoCString spec;
-    if (mLoadedURI) {
-      mLoadedURI->GetSpec(spec);
-    }
-    nsPrintfCString marker("TTFI after %dms (LongTask was at %dms) for URL %s",
-                           int(elapsed.ToMilliseconds()),
-                           int(elapsedLongTask.ToMilliseconds()), spec.get());
+    nsPrintfCString marker(
+        "TTFI after %dms (LongTask was at %dms) for URL %s",
+        int(elapsed.ToMilliseconds()), int(elapsedLongTask.ToMilliseconds()),
+        nsContentUtils::TruncatedURLForDisplay(mLoadedURI).get());
 
     PROFILER_MARKER_TEXT(
         "TimeToFirstInteractive (TTFI)", DOM,
@@ -399,15 +395,13 @@ void nsDOMNavigationTiming::NotifyNonBlankPaintForRootContentDocument() {
 
   mNonBlankPaint = TimeStamp::Now();
 
-  if (profiler_thread_is_being_profiled() || PAGELOAD_LOG_ENABLED()) {
+  if (profiler_thread_is_being_profiled_for_markers() ||
+      PAGELOAD_LOG_ENABLED()) {
     TimeDuration elapsed = mNonBlankPaint - mNavigationStart;
-    nsAutoCString spec;
-    if (mLoadedURI) {
-      mLoadedURI->GetSpec(spec);
-    }
     nsPrintfCString marker(
         "Non-blank paint after %dms for URL %s, %s",
-        int(elapsed.ToMilliseconds()), spec.get(),
+        int(elapsed.ToMilliseconds()),
+        nsContentUtils::TruncatedURLForDisplay(mLoadedURI).get(),
         mDocShellHasBeenActiveSinceNavigationStart
             ? "foreground tab"
             : "this tab was inactive some of the time between navigation start "
@@ -431,40 +425,38 @@ void nsDOMNavigationTiming::NotifyNonBlankPaintForRootContentDocument() {
           mNonBlankPaint);
     }
 
-    Telemetry::AccumulateTimeDelta(Telemetry::TIME_TO_NON_BLANK_PAINT_MS,
-                                   mNavigationStart, mNonBlankPaint);
+    glean::performance_page::non_blank_paint.AccumulateRawDuration(
+        mNonBlankPaint - mNavigationStart);
   }
 }
 
-void nsDOMNavigationTiming::NotifyContentfulPaintForRootContentDocument(
+void nsDOMNavigationTiming::NotifyContentfulCompositeForRootContentDocument(
     const mozilla::TimeStamp& aCompositeEndTime) {
   MOZ_ASSERT(NS_IsMainThread());
   MOZ_ASSERT(!mNavigationStart.IsNull());
 
-  if (!mContentfulPaint.IsNull()) {
+  if (!mContentfulComposite.IsNull()) {
     return;
   }
 
-  mContentfulPaint = aCompositeEndTime;
+  mContentfulComposite = aCompositeEndTime;
 
-  if (profiler_can_accept_markers() || PAGELOAD_LOG_ENABLED()) {
-    TimeDuration elapsed = mContentfulPaint - mNavigationStart;
-    nsAutoCString spec;
-    if (mLoadedURI) {
-      mLoadedURI->GetSpec(spec);
-    }
+  if (profiler_thread_is_being_profiled_for_markers() ||
+      PAGELOAD_LOG_ENABLED()) {
+    TimeDuration elapsed = mContentfulComposite - mNavigationStart;
     nsPrintfCString marker(
-        "Contentful paint after %dms for URL %s, %s",
-        int(elapsed.ToMilliseconds()), spec.get(),
+        "Contentful composite after %dms for URL %s, %s",
+        int(elapsed.ToMilliseconds()),
+        nsContentUtils::TruncatedURLForDisplay(mLoadedURI).get(),
         mDocShellHasBeenActiveSinceNavigationStart
             ? "foreground tab"
             : "this tab was inactive some of the time between navigation start "
               "and first non-blank paint");
     PAGELOAD_LOG(("%s", marker.get()));
     PROFILER_MARKER_TEXT(
-        "FirstContentfulPaint", DOM,
+        "FirstContentfulComposite", DOM,
         MarkerOptions(
-            MarkerTiming::Interval(mNavigationStart, mContentfulPaint),
+            MarkerTiming::Interval(mNavigationStart, mContentfulComposite),
             MarkerInnerWindowIdFromDocShell(mDocShell)),
         marker);
   }
@@ -482,10 +474,20 @@ void nsDOMNavigationTiming::NotifyContentfulPaintForRootContentDocument(
 
   if (mDocShellHasBeenActiveSinceNavigationStart) {
     Telemetry::AccumulateTimeDelta(Telemetry::TIME_TO_FIRST_CONTENTFUL_PAINT_MS,
-                                   mNavigationStart, mContentfulPaint);
+                                   mNavigationStart, mContentfulComposite);
   }
 }
 
+void nsDOMNavigationTiming::NotifyLargestContentfulRenderForRootContentDocument(
+    const DOMHighResTimeStamp& aRenderTime) {
+  MOZ_ASSERT(NS_IsMainThread());
+  MOZ_ASSERT(!mNavigationStart.IsNull());
+
+  // This can get called multiple times and updates over time.
+  mLargestContentfulRender =
+      mNavigationStart + TimeDuration::FromMilliseconds(aRenderTime);
+}
+
 void nsDOMNavigationTiming::NotifyDOMContentFlushedForRootContentDocument() {
   MOZ_ASSERT(NS_IsMainThread());
   MOZ_ASSERT(!mNavigationStart.IsNull());
@@ -496,15 +498,13 @@ void nsDOMNavigationTiming::NotifyDOMContentFlushedForRootContentDocument() {
 
   mDOMContentFlushed = TimeStamp::Now();
 
-  if (profiler_thread_is_being_profiled() || PAGELOAD_LOG_ENABLED()) {
+  if (profiler_thread_is_being_profiled_for_markers() ||
+      PAGELOAD_LOG_ENABLED()) {
     TimeDuration elapsed = mDOMContentFlushed - mNavigationStart;
-    nsAutoCString spec;
-    if (mLoadedURI) {
-      mLoadedURI->GetSpec(spec);
-    }
     nsPrintfCString marker(
         "DOMContentFlushed after %dms for URL %s, %s",
-        int(elapsed.ToMilliseconds()), spec.get(),
+        int(elapsed.ToMilliseconds()),
+        nsContentUtils::TruncatedURLForDisplay(mLoadedURI).get(),
         mDocShellHasBeenActiveSinceNavigationStart
             ? "foreground tab"
             : "this tab was inactive some of the time between navigation start "
@@ -525,6 +525,34 @@ void nsDOMNavigationTiming::NotifyDocShellStateChanged(
       (aDocShellState == DocShellState::eActive);
 }
 
+void nsDOMNavigationTiming::MaybeAddLCPProfilerMarker(
+    MarkerInnerWindowId aInnerWindowID) {
+  // This method might get called from outside of the main thread, so can't
+  // check `profiler_thread_is_being_profiled_for_markers()` here.
+  if (!profiler_is_active_and_unpaused()) {
+    return;
+  }
+
+  TimeStamp navStartTime = GetNavigationStartTimeStamp();
+  TimeStamp lcpTime = GetLargestContentfulRenderTimeStamp();
+
+  if (!navStartTime || !lcpTime) {
+    return;
+  }
+
+  TimeDuration elapsed = lcpTime - navStartTime;
+  nsPrintfCString marker("Largest contentful paint after %dms",
+                         int(elapsed.ToMilliseconds()));
+  PROFILER_MARKER_TEXT(
+      "LargestContentfulPaint", DOM,
+      // Putting this marker to the main thread even if it's called from another
+      // one.
+      MarkerOptions(MarkerThreadId::MainThread(),
+                    MarkerTiming::Interval(navStartTime, lcpTime),
+                    std::move(aInnerWindowID)),
+      marker);
+}
+
 mozilla::TimeStamp nsDOMNavigationTiming::GetUnloadEventStartTimeStamp() const {
   nsIScriptSecurityManager* ssm = nsContentUtils::GetSecurityManager();
   // todo: if you intend to update CheckSameOriginURI to log the error to the
@@ -566,7 +594,7 @@ nsDOMNavigationTiming::nsDOMNavigationTiming(nsDocShell* aDocShell,
       mNavigationStartHighRes(aOther->mNavigationStartHighRes),
       mNavigationStart(aOther->mNavigationStart),
       mNonBlankPaint(aOther->mNonBlankPaint),
-      mContentfulPaint(aOther->mContentfulPaint),
+      mContentfulComposite(aOther->mContentfulComposite),
       mDOMContentFlushed(aOther->mDOMContentFlushed),
       mBeforeUnloadStart(aOther->mBeforeUnloadStart),
       mUnloadStart(aOther->mUnloadStart),
@@ -584,61 +612,76 @@ nsDOMNavigationTiming::nsDOMNavigationTiming(nsDocShell* aDocShell,
 
 /* static */
 void mozilla::ipc::IPDLParamTraits<nsDOMNavigationTiming*>::Write(
-    IPC::Message* aMsg, IProtocol* aActor, nsDOMNavigationTiming* aParam) {
+    IPC::MessageWriter* aWriter, IProtocol* aActor,
+    nsDOMNavigationTiming* aParam) {
+  bool isNull = !aParam;
+  WriteIPDLParam(aWriter, aActor, isNull);
+  if (isNull) {
+    return;
+  }
+
   RefPtr<nsIURI> unloadedURI = aParam->mUnloadedURI.get();
   RefPtr<nsIURI> loadedURI = aParam->mLoadedURI.get();
-  WriteIPDLParam(aMsg, aActor, unloadedURI ? Some(unloadedURI) : Nothing());
-  WriteIPDLParam(aMsg, aActor, loadedURI ? Some(loadedURI) : Nothing());
-  WriteIPDLParam(aMsg, aActor, uint32_t(aParam->mNavigationType));
-  WriteIPDLParam(aMsg, aActor, aParam->mNavigationStartHighRes);
-  WriteIPDLParam(aMsg, aActor, aParam->mNavigationStart);
-  WriteIPDLParam(aMsg, aActor, aParam->mNonBlankPaint);
-  WriteIPDLParam(aMsg, aActor, aParam->mContentfulPaint);
-  WriteIPDLParam(aMsg, aActor, aParam->mDOMContentFlushed);
-  WriteIPDLParam(aMsg, aActor, aParam->mBeforeUnloadStart);
-  WriteIPDLParam(aMsg, aActor, aParam->mUnloadStart);
-  WriteIPDLParam(aMsg, aActor, aParam->mUnloadEnd);
-  WriteIPDLParam(aMsg, aActor, aParam->mLoadEventStart);
-  WriteIPDLParam(aMsg, aActor, aParam->mLoadEventEnd);
-  WriteIPDLParam(aMsg, aActor, aParam->mDOMLoading);
-  WriteIPDLParam(aMsg, aActor, aParam->mDOMInteractive);
-  WriteIPDLParam(aMsg, aActor, aParam->mDOMContentLoadedEventStart);
-  WriteIPDLParam(aMsg, aActor, aParam->mDOMContentLoadedEventEnd);
-  WriteIPDLParam(aMsg, aActor, aParam->mDOMComplete);
-  WriteIPDLParam(aMsg, aActor, aParam->mTTFI);
-  WriteIPDLParam(aMsg, aActor,
+  WriteIPDLParam(aWriter, aActor, unloadedURI ? Some(unloadedURI) : Nothing());
+  WriteIPDLParam(aWriter, aActor, loadedURI ? Some(loadedURI) : Nothing());
+  WriteIPDLParam(aWriter, aActor, uint32_t(aParam->mNavigationType));
+  WriteIPDLParam(aWriter, aActor, aParam->mNavigationStartHighRes);
+  WriteIPDLParam(aWriter, aActor, aParam->mNavigationStart);
+  WriteIPDLParam(aWriter, aActor, aParam->mNonBlankPaint);
+  WriteIPDLParam(aWriter, aActor, aParam->mContentfulComposite);
+  WriteIPDLParam(aWriter, aActor, aParam->mDOMContentFlushed);
+  WriteIPDLParam(aWriter, aActor, aParam->mBeforeUnloadStart);
+  WriteIPDLParam(aWriter, aActor, aParam->mUnloadStart);
+  WriteIPDLParam(aWriter, aActor, aParam->mUnloadEnd);
+  WriteIPDLParam(aWriter, aActor, aParam->mLoadEventStart);
+  WriteIPDLParam(aWriter, aActor, aParam->mLoadEventEnd);
+  WriteIPDLParam(aWriter, aActor, aParam->mDOMLoading);
+  WriteIPDLParam(aWriter, aActor, aParam->mDOMInteractive);
+  WriteIPDLParam(aWriter, aActor, aParam->mDOMContentLoadedEventStart);
+  WriteIPDLParam(aWriter, aActor, aParam->mDOMContentLoadedEventEnd);
+  WriteIPDLParam(aWriter, aActor, aParam->mDOMComplete);
+  WriteIPDLParam(aWriter, aActor, aParam->mTTFI);
+  WriteIPDLParam(aWriter, aActor,
                  aParam->mDocShellHasBeenActiveSinceNavigationStart);
 }
 
 /* static */
 bool mozilla::ipc::IPDLParamTraits<nsDOMNavigationTiming*>::Read(
-    const IPC::Message* aMsg, PickleIterator* aIter, IProtocol* aActor,
+    IPC::MessageReader* aReader, IProtocol* aActor,
     RefPtr<nsDOMNavigationTiming>* aResult) {
+  bool isNull;
+  if (!ReadIPDLParam(aReader, aActor, &isNull)) {
+    return false;
+  }
+  if (isNull) {
+    *aResult = nullptr;
+    return true;
+  }
+
   auto timing = MakeRefPtr<nsDOMNavigationTiming>(nullptr);
   uint32_t type;
   Maybe<RefPtr<nsIURI>> unloadedURI;
   Maybe<RefPtr<nsIURI>> loadedURI;
-  if (!ReadIPDLParam(aMsg, aIter, aActor, &unloadedURI) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &loadedURI) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &type) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mNavigationStartHighRes) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mNavigationStart) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mNonBlankPaint) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mContentfulPaint) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mDOMContentFlushed) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mBeforeUnloadStart) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mUnloadStart) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mUnloadEnd) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mLoadEventStart) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mLoadEventEnd) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mDOMLoading) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mDOMInteractive) ||
-      !ReadIPDLParam(aMsg, aIter, aActor,
-                     &timing->mDOMContentLoadedEventStart) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mDOMContentLoadedEventEnd) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mDOMComplete) ||
-      !ReadIPDLParam(aMsg, aIter, aActor, &timing->mTTFI) ||
-      !ReadIPDLParam(aMsg, aIter, aActor,
+  if (!ReadIPDLParam(aReader, aActor, &unloadedURI) ||
+      !ReadIPDLParam(aReader, aActor, &loadedURI) ||
+      !ReadIPDLParam(aReader, aActor, &type) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mNavigationStartHighRes) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mNavigationStart) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mNonBlankPaint) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mContentfulComposite) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mDOMContentFlushed) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mBeforeUnloadStart) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mUnloadStart) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mUnloadEnd) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mLoadEventStart) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mLoadEventEnd) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mDOMLoading) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mDOMInteractive) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mDOMContentLoadedEventStart) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mDOMContentLoadedEventEnd) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mDOMComplete) ||
+      !ReadIPDLParam(aReader, aActor, &timing->mTTFI) ||
+      !ReadIPDLParam(aReader, aActor,
                      &timing->mDocShellHasBeenActiveSinceNavigationStart)) {
     return false;
   }
