@@ -13,13 +13,11 @@
 #include "nsWindowRoot.h"
 #include "nsPIDOMWindow.h"
 #include "nsPresContext.h"
-#include "nsLayoutCID.h"
-#include "nsContentCID.h"
 #include "nsString.h"
 #include "nsFrameLoaderOwner.h"
 #include "nsFrameLoader.h"
 #include "nsQueryActor.h"
-#include "nsGlobalWindow.h"
+#include "nsGlobalWindowOuter.h"
 #include "nsFocusManager.h"
 #include "nsIContent.h"
 #include "nsIControllers.h"
@@ -34,16 +32,14 @@
 #include "mozilla/dom/JSActorService.h"
 #include "mozilla/dom/WindowGlobalParent.h"
 
-#ifdef MOZ_XUL
-#  include "nsXULElement.h"
-#endif
+#include "nsXULElement.h"
 
 using namespace mozilla;
 using namespace mozilla::dom;
 
 nsWindowRoot::nsWindowRoot(nsPIDOMWindowOuter* aWindow) {
+  SetIsOnMainThread();
   mWindow = aWindow;
-  mShowFocusRings = StaticPrefs::browser_display_show_focus_rings();
 }
 
 nsWindowRoot::~nsWindowRoot() {
@@ -51,17 +47,13 @@ nsWindowRoot::~nsWindowRoot() {
     mListenerManager->Disconnect();
   }
 
-  if (XRE_IsContentProcess()) {
-    JSActorService::UnregisterChromeEventTarget(this);
-  }
+  JSActorService::UnregisterChromeEventTarget(this);
 }
 
-NS_IMPL_CYCLE_COLLECTION_CLASS(nsWindowRoot)
+NS_IMPL_CYCLE_COLLECTION_WRAPPERCACHE_CLASS(nsWindowRoot)
 
 NS_IMPL_CYCLE_COLLECTION_UNLINK_BEGIN(nsWindowRoot)
-  if (XRE_IsContentProcess()) {
-    JSActorService::UnregisterChromeEventTarget(tmp);
-  }
+  JSActorService::UnregisterChromeEventTarget(tmp);
 
   NS_IMPL_CYCLE_COLLECTION_UNLINK(mWindow)
   NS_IMPL_CYCLE_COLLECTION_UNLINK(mListenerManager)
@@ -75,8 +67,6 @@ NS_IMPL_CYCLE_COLLECTION_TRAVERSE_BEGIN(nsWindowRoot)
   NS_IMPL_CYCLE_COLLECTION_TRAVERSE(mParent)
 NS_IMPL_CYCLE_COLLECTION_TRAVERSE_END
 
-NS_IMPL_CYCLE_COLLECTION_TRACE_WRAPPERCACHE(nsWindowRoot)
-
 NS_INTERFACE_MAP_BEGIN_CYCLE_COLLECTION(nsWindowRoot)
   NS_WRAPPERCACHE_INTERFACE_MAP_ENTRY
   NS_INTERFACE_MAP_ENTRY(nsISupports)
@@ -156,7 +146,6 @@ nsresult nsWindowRoot::GetControllers(bool aForVisibleWindow,
   nsIContent* focusedContent = nsFocusManager::GetFocusedDescendant(
       mWindow, searchRange, getter_AddRefs(focusedWindow));
   if (focusedContent) {
-#ifdef MOZ_XUL
     RefPtr<nsXULElement> xulElement = nsXULElement::FromNode(focusedContent);
     if (xulElement) {
       ErrorResult rv;
@@ -164,7 +153,6 @@ nsresult nsWindowRoot::GetControllers(bool aForVisibleWindow,
       NS_IF_ADDREF(*aResult);
       return rv.StealNSResult();
     }
-#endif
 
     HTMLTextAreaElement* htmlTextArea =
         HTMLTextAreaElement::FromNode(focusedContent);
@@ -395,10 +383,8 @@ void nsWindowRoot::EnumerateBrowsers(BrowserEnumerator aEnumFunc, void* aArg) {
 already_AddRefed<EventTarget> NS_NewWindowRoot(nsPIDOMWindowOuter* aWindow) {
   nsCOMPtr<EventTarget> result = new nsWindowRoot(aWindow);
 
-  if (XRE_IsContentProcess()) {
-    RefPtr<JSActorService> wasvc = JSActorService::GetSingleton();
-    wasvc->RegisterChromeEventTarget(result);
-  }
+  RefPtr<JSActorService> wasvc = JSActorService::GetSingleton();
+  wasvc->RegisterChromeEventTarget(result);
 
   return result.forget();
 }
