@@ -16,13 +16,14 @@
 #include "mozilla/ErrorResult.h"
 #include "mozilla/HashFunctions.h"
 #include "mozilla/Maybe.h"
+#include "mozilla/dom/BindingNames.h"
 #include "mozilla/dom/DOMJSClass.h"
-#include "mozilla/dom/DOMJSProxyHandler.h"
 #include "mozilla/dom/Exceptions.h"
 #include "mozilla/dom/JSSlots.h"
 #include "mozilla/dom/PrototypeList.h"
+#include "mozilla/dom/ProxyHandlerUtils.h"
 #include "mozilla/dom/RegisterBindings.h"
-#include "nsGlobalWindow.h"
+#include "nsGlobalWindowInner.h"
 #include "nsTHashtable.h"
 #include "WrapperFactory.h"
 
@@ -37,20 +38,25 @@ static JSObject* FindNamedConstructorForXray(
     return nullptr;
   }
 
-  // This is a call over Xrays, so we will actually use the return value
-  // (instead of just having it defined on the global now).  Check for named
-  // constructors with this id, in case that's what the caller is asking for.
-  for (unsigned slot = DOM_INTERFACE_SLOTS_BASE;
-       slot < JSCLASS_RESERVED_SLOTS(JS::GetClass(interfaceObject)); ++slot) {
-    JSObject* constructor =
-        &JS::GetReservedSlot(interfaceObject, slot).toObject();
-    if (JS_GetFunctionId(JS_GetObjectFunction(constructor)) ==
-        JSID_TO_STRING(aId)) {
-      return constructor;
+  if (IsInterfaceObject(interfaceObject)) {
+    // This is a call over Xrays, so we will actually use the return value
+    // (instead of just having it defined on the global now).  Check for named
+    // constructors with this id, in case that's what the caller is asking for.
+    for (unsigned slot = INTERFACE_OBJECT_FIRST_LEGACY_FACTORY_FUNCTION;
+         slot < INTERFACE_OBJECT_MAX_SLOTS; ++slot) {
+      const JS::Value& v = js::GetFunctionNativeReserved(interfaceObject, slot);
+      if (!v.isObject()) {
+        break;
+      }
+      JSObject* constructor = &v.toObject();
+      if (JS_GetMaybePartialFunctionId(JS_GetObjectFunction(constructor)) ==
+          aId.toString()) {
+        return constructor;
+      }
     }
   }
 
-  // None of the named constructors match, so the caller must want the
+  // None of the legacy factory functions match, so the caller must want the
   // interface object itself.
   return interfaceObject;
 }
@@ -60,9 +66,9 @@ bool WebIDLGlobalNameHash::DefineIfEnabled(
     JSContext* aCx, JS::Handle<JSObject*> aObj, JS::Handle<jsid> aId,
     JS::MutableHandle<mozilla::Maybe<JS::PropertyDescriptor>> aDesc,
     bool* aFound) {
-  MOZ_ASSERT(JSID_IS_STRING(aId), "Check for string id before calling this!");
+  MOZ_ASSERT(aId.isString(), "Check for string id before calling this!");
 
-  const WebIDLNameTableEntry* entry = GetEntry(JSID_TO_LINEAR_STRING(aId));
+  const WebIDLNameTableEntry* entry = GetEntry(aId.toLinearString());
   if (!entry) {
     *aFound = false;
     return true;
@@ -174,7 +180,7 @@ bool WebIDLGlobalNameHash::DefineIfEnabled(
 
 /* static */
 bool WebIDLGlobalNameHash::MayResolve(jsid aId) {
-  return GetEntry(JSID_TO_LINEAR_STRING(aId)) != nullptr;
+  return GetEntry(aId.toLinearString()) != nullptr;
 }
 
 /* static */
@@ -190,9 +196,9 @@ bool WebIDLGlobalNameHash::GetNames(JSContext* aCx, JS::Handle<JSObject*> aObj,
     if ((aNameType == AllNames ||
          !cache->HasEntryInSlot(entry.mConstructorId)) &&
         (!entry.mEnabled || entry.mEnabled(aCx, aObj))) {
-      JSString* str =
-          JS_AtomizeStringN(aCx, sNames + entry.mNameOffset, entry.mNameLength);
-      if (!str || !aNames.append(JS::PropertyKey::fromNonIntAtom(str))) {
+      JSString* str = JS_AtomizeStringN(aCx, BindingName(entry.mNameOffset),
+                                        entry.mNameLength);
+      if (!str || !aNames.append(JS::PropertyKey::NonIntAtom(str))) {
         return false;
       }
     }
@@ -217,7 +223,7 @@ bool WebIDLGlobalNameHash::ResolveForSystemGlobal(JSContext* aCx,
   }
 
   // We don't resolve any non-string entries.
-  if (!JSID_IS_STRING(aId)) {
+  if (!aId.isString()) {
     return true;
   }
 
@@ -227,7 +233,7 @@ bool WebIDLGlobalNameHash::ResolveForSystemGlobal(JSContext* aCx,
   MOZ_ASSERT(!xpc::WrapperFactory::IsXrayWrapper(aObj), "Xrays not supported!");
 
   // Look up the corresponding entry in the name table, and resolve if enabled.
-  const WebIDLNameTableEntry* entry = GetEntry(JSID_TO_LINEAR_STRING(aId));
+  const WebIDLNameTableEntry* entry = GetEntry(aId.toLinearString());
   if (entry && (!entry->mEnabled || entry->mEnabled(aCx, aObj))) {
     if (NS_WARN_IF(!GetPerInterfaceObjectHandle(
             aCx, entry->mConstructorId, entry->mCreate,
@@ -261,9 +267,9 @@ bool WebIDLGlobalNameHash::NewEnumerateSystemGlobal(
   for (size_t i = 0; i < sCount; ++i) {
     const WebIDLNameTableEntry& entry = sEntries[i];
     if (!entry.mEnabled || entry.mEnabled(aCx, aObj)) {
-      JSString* str =
-          JS_AtomizeStringN(aCx, sNames + entry.mNameOffset, entry.mNameLength);
-      if (!str || !aProperties.append(JS::PropertyKey::fromNonIntAtom(str))) {
+      JSString* str = JS_AtomizeStringN(aCx, BindingName(entry.mNameOffset),
+                                        entry.mNameLength);
+      if (!str || !aProperties.append(JS::PropertyKey::NonIntAtom(str))) {
         return false;
       }
     }
