@@ -6,12 +6,15 @@
 
 #include "OffscreenCanvas.h"
 
+#include "mozilla/Atomics.h"
+#include "mozilla/CheckedInt.h"
 #include "mozilla/dom/BlobImpl.h"
 #include "mozilla/dom/OffscreenCanvasBinding.h"
 #include "mozilla/dom/OffscreenCanvasDisplayHelper.h"
 #include "mozilla/dom/OffscreenCanvasRenderingContext2D.h"
 #include "mozilla/dom/Promise.h"
 #include "mozilla/dom/WorkerPrivate.h"
+#include "mozilla/dom/WorkerRef.h"
 #include "mozilla/dom/WorkerScope.h"
 #include "mozilla/layers/ImageBridgeChild.h"
 #include "mozilla/Telemetry.h"
@@ -63,10 +66,17 @@ OffscreenCanvas::OffscreenCanvas(
       mDisplay(aDisplay) {}
 
 OffscreenCanvas::~OffscreenCanvas() {
+  Destroy();
   NS_ReleaseOnMainThread("OffscreenCanvas::mExpandedReader",
                          mExpandedReader.forget());
 }
 
+void OffscreenCanvas::Destroy() {
+  if (mDisplay) {
+    mDisplay->DestroyCanvas();
+  }
+}
+
 JSObject* OffscreenCanvas::WrapObject(JSContext* aCx,
                                       JS::Handle<JSObject*> aGivenProto) {
   return OffscreenCanvas_Binding::Wrap(aCx, this, aGivenProto);
@@ -74,7 +84,25 @@ JSObject* OffscreenCanvas::WrapObject(JSContext* aCx,
 
 /* static */
 already_AddRefed<OffscreenCanvas> OffscreenCanvas::Constructor(
-    const GlobalObject& aGlobal, uint32_t aWidth, uint32_t aHeight) {
+    const GlobalObject& aGlobal, uint32_t aWidth, uint32_t aHeight,
+    ErrorResult& aRv) {
+  // CanvasRenderingContextHelper::GetWidthHeight wants us to return
+  // an nsIntSize, so make sure that that will work.
+  if (!CheckedInt<int32_t>(aWidth).isValid()) {
+    aRv.ThrowRangeError(
+        nsPrintfCString("OffscreenCanvas width %u is out of range: must be no "
+                        "greater than 2147483647.",
+                        aWidth));
+    return nullptr;
+  }
+  if (!CheckedInt<int32_t>(aHeight).isValid()) {
+    aRv.ThrowRangeError(
+        nsPrintfCString("OffscreenCanvas height %u is out of range: must be no "
+                        "greater than 2147483647.",
+                        aHeight));
+    return nullptr;
+  }
+
   nsCOMPtr<nsIGlobalObject> global = do_QueryInterface(aGlobal.GetAsSupports());
   RefPtr<OffscreenCanvas> offscreenCanvas =
       new OffscreenCanvas(global, aWidth, aHeight);
@@ -83,28 +111,60 @@ already_AddRefed<OffscreenCanvas> OffscreenCanvas::Constructor(
 
 void OffscreenCanvas::SetWidth(uint32_t aWidth, ErrorResult& aRv) {
   if (mNeutered) {
-    aRv.ThrowInvalidStateError(
-        "Cannot set width of placeholder canvas transferred to worker.");
+    aRv.ThrowInvalidStateError("Cannot set width of detached OffscreenCanvas.");
     return;
   }
 
-  if (mWidth != aWidth) {
-    mWidth = aWidth;
-    CanvasAttrChanged();
+  // CanvasRenderingContextHelper::GetWidthHeight wants us to return
+  // an nsIntSize, so make sure that that will work.
+  if (!CheckedInt<int32_t>(aWidth).isValid()) {
+    aRv.ThrowRangeError(
+        nsPrintfCString("OffscreenCanvas width %u is out of range: must be no "
+                        "greater than 2147483647.",
+                        aWidth));
+    return;
   }
+
+  mWidth = aWidth;
+  CanvasAttrChanged();
 }
 
 void OffscreenCanvas::SetHeight(uint32_t aHeight, ErrorResult& aRv) {
   if (mNeutered) {
     aRv.ThrowInvalidStateError(
-        "Cannot set height of placeholder canvas transferred to worker.");
+        "Cannot set height of detached OffscreenCanvas.");
     return;
   }
 
-  if (mHeight != aHeight) {
-    mHeight = aHeight;
-    CanvasAttrChanged();
+  // CanvasRenderingContextHelper::GetWidthHeight wants us to return
+  // an nsIntSize, so make sure that that will work.
+  if (!CheckedInt<int32_t>(aHeight).isValid()) {
+    aRv.ThrowRangeError(
+        nsPrintfCString("OffscreenCanvas height %u is out of range: must be no "
+                        "greater than 2147483647.",
+                        aHeight));
+    return;
   }
+
+  mHeight = aHeight;
+  CanvasAttrChanged();
+}
+
+void OffscreenCanvas::SetSize(const nsIntSize& aSize, ErrorResult& aRv) {
+  if (mNeutered) {
+    aRv.ThrowInvalidStateError(
+        "Cannot set dimensions of detached OffscreenCanvas.");
+    return;
+  }
+
+  if (NS_WARN_IF(aSize.IsEmpty())) {
+    aRv.ThrowRangeError("OffscreenCanvas size is empty, must be non-empty.");
+    return;
+  }
+
+  mWidth = aSize.width;
+  mHeight = aSize.height;
+  CanvasAttrChanged();
 }
 
 void OffscreenCanvas::GetContext(
@@ -114,7 +174,7 @@ void OffscreenCanvas::GetContext(
   if (mNeutered) {
     aResult.SetNull();
     aRv.ThrowInvalidStateError(
-        "Cannot create context for placeholder canvas transferred to worker.");
+        "Cannot create context for detached OffscreenCanvas.");
     return;
   }
 
@@ -142,6 +202,27 @@ void OffscreenCanvas::GetContext(
       return;
   }
 
+  // If we are on a worker, we need to give our OffscreenCanvasDisplayHelper
+  // object access to a worker ref so we can dispatch properly during painting
+  // if we need to flush our contents to its ImageContainer for display.
+  RefPtr<ThreadSafeWorkerRef> workerRef;
+  if (mDisplay) {
+    if (WorkerPrivate* workerPrivate = GetCurrentThreadWorkerPrivate()) {
+      RefPtr<StrongWorkerRef> strongRef = StrongWorkerRef::Create(
+          workerPrivate, "OffscreenCanvas::GetContext",
+          [display = mDisplay]() { display->DestroyCanvas(); });
+      if (NS_WARN_IF(!strongRef)) {
+        aResult.SetNull();
+        aRv.ThrowUnknownError("Worker shutting down");
+        return;
+      }
+
+      workerRef = new ThreadSafeWorkerRef(strongRef);
+    } else {
+      MOZ_ASSERT(NS_IsMainThread());
+    }
+  }
+
   RefPtr<nsISupports> result = CanvasRenderingContextHelper::GetOrCreateContext(
       aCx, contextType, aContextOptions, aRv);
   if (!result) {
@@ -183,7 +264,8 @@ void OffscreenCanvas::GetContext(
   }
 
   if (mDisplay) {
-    mDisplay->UpdateContext(mCurrentContextType, childId);
+    mDisplay->UpdateContext(this, std::move(workerRef), mCurrentContextType,
+                            childId);
   }
 }
 
@@ -191,11 +273,27 @@ already_AddRefed<nsICanvasRenderingContextInternal>
 OffscreenCanvas::CreateContext(CanvasContextType aContextType) {
   RefPtr<nsICanvasRenderingContextInternal> ret =
       CanvasRenderingContextHelper::CreateContext(aContextType);
+  if (NS_WARN_IF(!ret)) {
+    return nullptr;
+  }
 
   ret->SetOffscreenCanvas(this);
   return ret.forget();
 }
 
+Maybe<uint64_t> OffscreenCanvas::GetWindowID() {
+  if (NS_IsMainThread()) {
+    if (nsIGlobalObject* global = GetOwnerGlobal()) {
+      if (auto* window = global->GetAsInnerWindow()) {
+        return Some(window->WindowID());
+      }
+    }
+  } else if (auto* workerPrivate = GetCurrentThreadWorkerPrivate()) {
+    return Some(workerPrivate->WindowID());
+  }
+  return Nothing();
+}
+
 void OffscreenCanvas::UpdateDisplayData(
     const OffscreenCanvasDisplayData& aData) {
   if (!mDisplay) {
@@ -243,17 +341,50 @@ void OffscreenCanvas::CommitFrameToCompositor() {
   mDisplay->CommitFrameToCompositor(mCurrentContext, mTextureType, update);
 }
 
-OffscreenCanvasCloneData* OffscreenCanvas::ToCloneData() {
-  return new OffscreenCanvasCloneData(mDisplay, mWidth, mHeight,
-                                      mCompositorBackendType, mTextureType,
-                                      mNeutered, mIsWriteOnly, mExpandedReader);
+UniquePtr<OffscreenCanvasCloneData> OffscreenCanvas::ToCloneData(
+    JSContext* aCx) {
+  if (NS_WARN_IF(mNeutered)) {
+    ErrorResult rv;
+    rv.ThrowDataCloneError(
+        "Cannot clone OffscreenCanvas that is already transferred.");
+    MOZ_ALWAYS_TRUE(rv.MaybeSetPendingException(aCx));
+    return nullptr;
+  }
+
+  if (NS_WARN_IF(mCurrentContext)) {
+    ErrorResult rv;
+    rv.ThrowInvalidStateError("Cannot clone canvas with context.");
+    MOZ_ALWAYS_TRUE(rv.MaybeSetPendingException(aCx));
+    return nullptr;
+  }
+
+  // Check if we are using HTMLCanvasElement::captureStream. This is not
+  // defined by the spec yet, so it is better to fail now than implement
+  // something not compliant:
+  // https://github.com/w3c/mediacapture-fromelement/issues/65
+  // https://github.com/w3c/mediacapture-extensions/pull/26
+  // https://github.com/web-platform-tests/wpt/issues/21102
+  if (mDisplay && NS_WARN_IF(mDisplay->UsingElementCaptureStream())) {
+    ErrorResult rv;
+    rv.ThrowNotSupportedError(
+        "Cannot transfer OffscreenCanvas bound to element using "
+        "captureStream.");
+    MOZ_ALWAYS_TRUE(rv.MaybeSetPendingException(aCx));
+    return nullptr;
+  }
+
+  auto cloneData = MakeUnique<OffscreenCanvasCloneData>(
+      mDisplay, mWidth, mHeight, mCompositorBackendType, mTextureType,
+      mNeutered, mIsWriteOnly, mExpandedReader);
+  SetNeutered();
+  return cloneData;
 }
 
 already_AddRefed<ImageBitmap> OffscreenCanvas::TransferToImageBitmap(
     ErrorResult& aRv) {
   if (mNeutered) {
     aRv.ThrowInvalidStateError(
-        "Cannot get bitmap from placeholder canvas transferred to worker.");
+        "Cannot get bitmap from detached OffscreenCanvas.");
     return nullptr;
   }
 
@@ -265,39 +396,45 @@ already_AddRefed<ImageBitmap> OffscreenCanvas::TransferToImageBitmap(
 
   RefPtr<ImageBitmap> result =
       ImageBitmap::CreateFromOffscreenCanvas(GetOwnerGlobal(), *this, aRv);
-  if (aRv.Failed()) {
+  if (!result) {
     return nullptr;
   }
 
-  if (result && mCurrentContext) {
-    // FIXME(aosmond): The spec is unclear about the state of the canvas after
-    // clearing. Does it expect to preserve the WebGL state, other than the
-    // buffer state? Once we have clarity, we should ensure we clear the WebGL
-    // canvas as desired.
-    mCurrentContext->Reset();
-    mCurrentContext->SetDimensions(result->Width(), result->Height());
+  if (mCurrentContext) {
+    mCurrentContext->ResetBitmap();
   }
   return result.forget();
 }
 
 already_AddRefed<EncodeCompleteCallback>
-OffscreenCanvas::CreateEncodeCompleteCallback(
-    nsCOMPtr<nsIGlobalObject>&& aGlobal, Promise* aPromise) {
+OffscreenCanvas::CreateEncodeCompleteCallback(Promise* aPromise) {
   // Encoder callback when encoding is complete.
   class EncodeCallback : public EncodeCompleteCallback {
    public:
-    EncodeCallback(nsCOMPtr<nsIGlobalObject>&& aGlobal, Promise* aPromise)
-        : mGlobal(std::move(aGlobal)), mPromise(aPromise) {}
+    explicit EncodeCallback(Promise* aPromise)
+        : mPromise(aPromise), mCanceled(false) {}
+
+    void MaybeInitWorkerRef() {
+      WorkerPrivate* wp = GetCurrentThreadWorkerPrivate();
+      if (wp) {
+        mWorkerRef = WeakWorkerRef::Create(
+            wp, [self = RefPtr{this}]() { self->Cancel(); });
+        if (!mWorkerRef) {
+          Cancel();
+        }
+      }
+    }
 
-    // This is called on main thread.
     nsresult ReceiveBlobImpl(already_AddRefed<BlobImpl> aBlobImpl) override {
       RefPtr<BlobImpl> blobImpl = aBlobImpl;
+      mWorkerRef = nullptr;
 
       if (mPromise) {
-        if (NS_WARN_IF(!blobImpl)) {
+        RefPtr<nsIGlobalObject> global = mPromise->GetGlobalObject();
+        if (NS_WARN_IF(!global) || NS_WARN_IF(!blobImpl)) {
           mPromise->MaybeReject(NS_ERROR_FAILURE);
         } else {
-          RefPtr<Blob> blob = Blob::Create(mGlobal, blobImpl);
+          RefPtr<Blob> blob = Blob::Create(global, blobImpl);
           if (NS_WARN_IF(!blob)) {
             mPromise->MaybeReject(NS_ERROR_FAILURE);
           } else {
@@ -306,17 +443,27 @@ OffscreenCanvas::CreateEncodeCompleteCallback(
         }
       }
 
-      mGlobal = nullptr;
       mPromise = nullptr;
 
       return NS_OK;
     }
 
-    nsCOMPtr<nsIGlobalObject> mGlobal;
+    bool CanBeDeletedOnAnyThread() override { return mCanceled; }
+
+    void Cancel() {
+      mPromise = nullptr;
+      mWorkerRef = nullptr;
+      mCanceled = true;
+    }
+
     RefPtr<Promise> mPromise;
+    RefPtr<WeakWorkerRef> mWorkerRef;
+    Atomic<bool> mCanceled;
   };
 
-  return MakeAndAddRef<EncodeCallback>(std::move(aGlobal), aPromise);
+  RefPtr<EncodeCallback> p = MakeAndAddRef<EncodeCallback>(aPromise);
+  p->MaybeInitWorkerRef();
+  return p.forget();
 }
 
 already_AddRefed<Promise> OffscreenCanvas::ConvertToBlob(
@@ -329,7 +476,7 @@ already_AddRefed<Promise> OffscreenCanvas::ConvertToBlob(
 
   if (mNeutered) {
     aRv.ThrowInvalidStateError(
-        "Cannot get blob from placeholder canvas transferred to worker.");
+        "Cannot get blob from detached OffscreenCanvas.");
     return nullptr;
   }
 
@@ -358,8 +505,9 @@ already_AddRefed<Promise> OffscreenCanvas::ConvertToBlob(
   }
 
   RefPtr<EncodeCompleteCallback> callback =
-      CreateEncodeCompleteCallback(std::move(global), promise);
-  bool usePlaceholder = ShouldResistFingerprinting();
+      CreateEncodeCompleteCallback(promise);
+  bool usePlaceholder =
+      ShouldResistFingerprinting(RFPTarget::CanvasImageExtractionPrompt);
   CanvasRenderingContextHelper::ToBlob(callback, type, encodeOptions,
                                        /* aUsingCustomOptions */ false,
                                        usePlaceholder, aRv);
@@ -382,7 +530,7 @@ already_AddRefed<Promise> OffscreenCanvas::ToBlob(JSContext* aCx,
 
   if (mNeutered) {
     aRv.ThrowInvalidStateError(
-        "Cannot get blob from placeholder canvas transferred to worker.");
+        "Cannot get blob from detached OffscreenCanvas.");
     return nullptr;
   }
 
@@ -399,8 +547,9 @@ already_AddRefed<Promise> OffscreenCanvas::ToBlob(JSContext* aCx,
   }
 
   RefPtr<EncodeCompleteCallback> callback =
-      CreateEncodeCompleteCallback(std::move(global), promise);
-  bool usePlaceholder = ShouldResistFingerprinting();
+      CreateEncodeCompleteCallback(promise);
+  bool usePlaceholder =
+      ShouldResistFingerprinting(RFPTarget::CanvasImageExtractionPrompt);
   CanvasRenderingContextHelper::ToBlob(aCx, callback, aType, aParams,
                                        usePlaceholder, aRv);
 
@@ -423,28 +572,24 @@ void OffscreenCanvas::SetWriteOnly(RefPtr<nsIPrincipal>&& aExpandedReader) {
   mIsWriteOnly = true;
 }
 
-bool OffscreenCanvas::CallerCanRead(nsIPrincipal* aPrincipal) const {
+bool OffscreenCanvas::CallerCanRead(nsIPrincipal& aPrincipal) const {
   if (!mIsWriteOnly) {
     return true;
   }
 
-  if (!aPrincipal) {
-    return false;
-  }
-
   // If mExpandedReader is set, this canvas was tainted only by
   // mExpandedReader's resources. So allow reading if the subject
   // principal subsumes mExpandedReader.
-  if (mExpandedReader && aPrincipal->Subsumes(mExpandedReader)) {
+  if (mExpandedReader && aPrincipal.Subsumes(mExpandedReader)) {
     return true;
   }
 
-  return nsContentUtils::PrincipalHasPermission(*aPrincipal,
+  return nsContentUtils::PrincipalHasPermission(aPrincipal,
                                                 nsGkAtoms::all_urlsPermission);
 }
 
-bool OffscreenCanvas::ShouldResistFingerprinting() const {
-  return nsContentUtils::ShouldResistFingerprinting(GetOwnerGlobal());
+bool OffscreenCanvas::ShouldResistFingerprinting(RFPTarget aTarget) const {
+  return nsContentUtils::ShouldResistFingerprinting(GetOwnerGlobal(), aTarget);
 }
 
 /* static */
@@ -466,11 +611,7 @@ already_AddRefed<OffscreenCanvas> OffscreenCanvas::CreateFromCloneData(
 /* static */
 bool OffscreenCanvas::PrefEnabledOnWorkerThread(JSContext* aCx,
                                                 JSObject* aObj) {
-  if (NS_IsMainThread()) {
-    return true;
-  }
-
-  return CanvasUtils::IsOffscreenCanvasEnabled(aCx, aObj);
+  return NS_IsMainThread() || StaticPrefs::gfx_offscreencanvas_enabled();
 }
 
 NS_IMPL_CYCLE_COLLECTION_INHERITED(OffscreenCanvas, DOMEventTargetHelper,
@@ -480,7 +621,7 @@ NS_IMPL_ADDREF_INHERITED(OffscreenCanvas, DOMEventTargetHelper)
 NS_IMPL_RELEASE_INHERITED(OffscreenCanvas, DOMEventTargetHelper)
 
 NS_INTERFACE_MAP_BEGIN_CYCLE_COLLECTION(OffscreenCanvas)
-  NS_INTERFACE_MAP_ENTRY(nsISupports)
+  NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsISupports, EventTarget)
 NS_INTERFACE_MAP_END_INHERITING(DOMEventTargetHelper)
 
 }  // namespace mozilla::dom
