@@ -8,21 +8,28 @@
 
 #include <limits>
 #include <string>
+#include <string_view>
+#include <tuple>
 #include <type_traits>
 #include <unordered_map>
 #include <vector>
 
+#include "GLContextTypes.h"
 #include "GLDefs.h"
 #include "ImageContainer.h"
 #include "mozilla/Casting.h"
 #include "mozilla/CheckedInt.h"
+#include "mozilla/EnumTypeTraits.h"
 #include "mozilla/MathAlgorithms.h"
 #include "mozilla/Range.h"
 #include "mozilla/RefCounted.h"
 #include "mozilla/Result.h"
 #include "mozilla/ResultVariant.h"
+#include "mozilla/Span.h"
+#include "mozilla/TypedEnumBits.h"
 #include "mozilla/gfx/2D.h"
 #include "mozilla/gfx/BuildConstants.h"
+#include "mozilla/gfx/Logging.h"
 #include "mozilla/gfx/Point.h"
 #include "mozilla/gfx/Rect.h"
 #include "mozilla/ipc/Shmem.h"
@@ -33,6 +40,7 @@
 #include "nsString.h"
 #include "mozilla/dom/WebGLRenderingContextBinding.h"
 #include "mozilla/ipc/SharedMemoryBasic.h"
+#include "TiedFields.h"
 
 // Manual reflection of WebIDL typedefs that are different from their
 // OpenGL counterparts.
@@ -80,26 +88,10 @@ inline void* calloc(const ForbidNarrowing<size_t> n,
 
 // -
 
-namespace detail {
-
-template <typename From>
-class AutoAssertCastT final {
-  const From mVal;
-
- public:
-  explicit AutoAssertCastT(const From val) : mVal(val) {}
-
-  template <typename To>
-  operator To() const {
-    return AssertedCast<To>(mVal);
-  }
-};
-
-}  // namespace detail
-
+// TODO: Remove this now-mere-alias.
 template <typename From>
 inline auto AutoAssertCast(const From val) {
-  return detail::AutoAssertCastT<From>(val);
+  return LazyAssertedCast(val);
 }
 
 const char* GetEnumName(GLenum val, const char* defaultRet = "<unknown>");
@@ -256,6 +248,7 @@ enum class WebGLExtensionID : uint8_t {
   WEBGL_draw_buffers,
   WEBGL_explicit_present,
   WEBGL_lose_context,
+  WEBGL_provoking_vertex,
   Max
 };
 
@@ -343,37 +336,68 @@ enum class BufferKind : uint8_t {
 struct FloatOrInt final  // For TexParameter[fi] and friends.
 {
   bool isFloat = false;
+  uint8_t padding[3] = {};
   GLfloat f = 0;
   GLint i = 0;
 
   explicit FloatOrInt(GLint x = 0) : isFloat(false), f(x), i(x) {}
 
   explicit FloatOrInt(GLfloat x) : isFloat(true), f(x), i(roundf(x)) {}
+
+  auto MutTiedFields() { return std::tie(isFloat, padding, f, i); }
 };
 
-struct WebGLContextOptions {
+// -
+
+struct WebGLContextOptions final {
   bool alpha = true;
   bool depth = true;
   bool stencil = false;
   bool premultipliedAlpha = true;
+
   bool antialias = true;
   bool preserveDrawingBuffer = false;
   bool failIfMajorPerformanceCaveat = false;
   bool xrCompatible = false;
+
   dom::WebGLPowerPreference powerPreference =
       dom::WebGLPowerPreference::Default;
-  dom::PredefinedColorSpace colorSpace = dom::PredefinedColorSpace::Srgb;
-  bool ignoreColorSpace = true;  // Our legacy behavior.
+  std::optional<dom::PredefinedColorSpace> colorSpace;
   bool shouldResistFingerprinting = true;
+
   bool enableDebugRendererInfo = false;
 
+  auto MutTiedFields() {
+    // clang-format off
+    return std::tie(
+      alpha,
+      depth,
+      stencil,
+      premultipliedAlpha,
+
+      antialias,
+      preserveDrawingBuffer,
+      failIfMajorPerformanceCaveat,
+      xrCompatible,
+
+      powerPreference,
+      colorSpace,
+      shouldResistFingerprinting,
+
+      enableDebugRendererInfo);
+    // clang-format on
+  }
+
+  // -
+
   WebGLContextOptions();
   WebGLContextOptions(const WebGLContextOptions&) = default;
 
-  bool operator==(const WebGLContextOptions&) const;
-  bool operator!=(const WebGLContextOptions& rhs) const {
-    return !(*this == rhs);
+  using Self = WebGLContextOptions;
+  friend bool operator==(const Self& a, const Self& b) {
+    return TiedFields(a) == TiedFields(b);
   }
+  friend bool operator!=(const Self& a, const Self& b) { return !(a == b); }
 };
 
 namespace gfx {
@@ -384,8 +408,6 @@ inline ColorSpace2 ToColorSpace2(const dom::PredefinedColorSpace cs) {
       return ColorSpace2::SRGB;
     case dom::PredefinedColorSpace::Display_p3:
       return ColorSpace2::DISPLAY_P3;
-    case dom::PredefinedColorSpace::EndGuard_:
-      break;
   }
   MOZ_CRASH("Exhaustive switch");
 }
@@ -401,6 +423,8 @@ struct avec2 {
   T x = T();
   T y = T();
 
+  auto MutTiedFields() { return std::tie(x, y); }
+
   template <typename U, typename V>
   static Maybe<avec2> From(const U _x, const V _y) {
     const auto x = CheckedInt<T>(_x);
@@ -476,6 +500,8 @@ struct avec3 {
   T y = T();
   T z = T();
 
+  auto MutTiedFields() { return std::tie(x, y, z); }
+
   template <typename U, typename V>
   static Maybe<avec3> From(const U _x, const V _y, const V _z) {
     const auto x = CheckedInt<T>(_x);
@@ -514,14 +540,21 @@ struct PackingInfo final {
   GLenum format = 0;
   GLenum type = 0;
 
-  bool operator<(const PackingInfo& x) const {
-    if (format != x.format) return format < x.format;
+  auto MutTiedFields() { return std::tie(format, type); }
 
-    return type < x.type;
+  using Self = PackingInfo;
+  friend bool operator<(const Self& a, const Self& b) {
+    return TiedFields(a) < TiedFields(b);
+  }
+  friend bool operator==(const Self& a, const Self& b) {
+    return TiedFields(a) == TiedFields(b);
   }
 
-  bool operator==(const PackingInfo& x) const {
-    return (format == x.format && type == x.type);
+  template <class T>
+  friend T& operator<<(T& s, const PackingInfo& pi) {
+    s << "PackingInfo{format: " << EnumString(pi.format)
+      << ", type: " << EnumString(pi.type) << "}";
+    return s;
   }
 };
 
@@ -564,9 +597,13 @@ class EnumMask {
  public:
   BitRef operator[](const E i) { return {*this, Mask(i)}; }
   bool operator[](const E i) const { return mBits & Mask(i); }
+
+  // -
+
+  auto MutTiedFields() { return std::tie(mBits); }
 };
 
-class ExtensionBits : public EnumMask<WebGLExtensionID> {};
+using ExtensionBits = EnumMask<WebGLExtensionID>;
 
 // -
 
@@ -590,9 +627,16 @@ inline bool ReadContextLossReason(const uint8_t val,
 struct InitContextDesc final {
   bool isWebgl2 = false;
   bool resistFingerprinting = false;
+  std::array<uint8_t, 2> _padding;
+  uint32_t principalKey = 0;
   uvec2 size = {};
   WebGLContextOptions options;
-  uint32_t principalKey = 0;
+  std::array<uint8_t, 3> _padding2;
+
+  auto MutTiedFields() {
+    return std::tie(isWebgl2, resistFingerprinting, _padding, principalKey,
+                    size, options, _padding2);
+  }
 };
 
 constexpr uint32_t kMaxTransformFeedbackSeparateAttribs = 4;
@@ -617,17 +661,85 @@ struct Limits final {
 
   // Exts
   bool astcHdr = false;
+  std::array<uint8_t, 3> _padding;
   uint32_t maxColorDrawBuffers = 1;
+  uint32_t maxMultiviewLayers = 0;
   uint64_t queryCounterBitsTimeElapsed = 0;
   uint64_t queryCounterBitsTimestamp = 0;
-  uint32_t maxMultiviewLayers = 0;
+
+  auto MutTiedFields() {
+    return std::tie(supportedExtensions,
+
+                    maxTexUnits, maxTex2dSize, maxTexCubeSize, maxVertexAttribs,
+                    maxViewportDim, pointSizeRange, lineWidthRange,
+
+                    maxTexArrayLayers, maxTex3dSize, maxUniformBufferBindings,
+                    uniformBufferOffsetAlignment,
+
+                    astcHdr, _padding, maxColorDrawBuffers, maxMultiviewLayers,
+                    queryCounterBitsTimeElapsed, queryCounterBitsTimestamp);
+  }
 };
 
+// -
+
+template <class T, size_t PaddedSize>
+struct Padded {
+ private:
+  T val = {};
+  uint8_t padding[PaddedSize - sizeof(T)] = {};
+
+ public:
+  operator T&() { return val; }
+  operator const T&() const { return val; }
+
+  auto& operator=(const T& rhs) { return val = rhs; }
+  auto& operator=(T&& rhs) { return val = std::move(rhs); }
+
+  auto& operator*() { return val; }
+  auto& operator*() const { return val; }
+  auto operator->() { return &val; }
+  auto operator->() const { return &val; }
+};
+
+// -
+
+enum class OptionalRenderableFormatBits : uint8_t {
+  RGB8 = (1 << 0),
+  SRGB8 = (1 << 1),
+};
+MOZ_MAKE_ENUM_CLASS_BITWISE_OPERATORS(OptionalRenderableFormatBits)
+inline constexpr bool IsEnumCase(const OptionalRenderableFormatBits raw) {
+  auto rawWithoutValidBits = UnderlyingValue(raw);
+  auto bit = decltype(rawWithoutValidBits){1};
+  while (bit) {
+    switch (OptionalRenderableFormatBits{bit}) {
+      // -Werror=switch ensures exhaustive.
+      case OptionalRenderableFormatBits::RGB8:
+      case OptionalRenderableFormatBits::SRGB8:
+        rawWithoutValidBits &= ~bit;
+        break;
+    }
+    bit <<= 1;
+  }
+  return rawWithoutValidBits == 0;
+}
+
+// -
+
 struct InitContextResult final {
-  std::string error;
+  Padded<std::string, 32> error;  // MINGW 32-bit needs this padding.
   WebGLContextOptions options;
-  webgl::Limits limits;
+  gl::GLVendor vendor;
+  OptionalRenderableFormatBits optionalRenderableFormatBits;
+  uint8_t _padding = {};
+  Limits limits;
   EnumMask<layers::SurfaceDescriptor::Type> uploadableSdTypes;
+
+  auto MutTiedFields() {
+    return std::tie(error, options, vendor, optionalRenderableFormatBits,
+                    _padding, limits, uploadableSdTypes);
+  }
 };
 
 // -
@@ -667,14 +779,30 @@ struct CompileResult final {
 struct OpaqueFramebufferOptions final {
   bool depthStencil = true;
   bool antialias = true;
+  std::array<uint8_t, 2> _padding;
   uint32_t width = 0;
   uint32_t height = 0;
+
+  auto MutTiedFields() {
+    return std::tie(depthStencil, antialias, _padding, width, height);
+  }
 };
 
 // -
 
 struct SwapChainOptions final {
+  layers::RemoteTextureId remoteTextureId;
+  layers::RemoteTextureOwnerId remoteTextureOwnerId;
   bool bgra = false;
+  bool forceAsyncPresent = false;
+  // Pad to sizeof(u64):
+  uint16_t padding1 = 0;
+  uint32_t padding2 = 0;
+
+  auto MutTiedFields() {
+    return std::tie(remoteTextureId, remoteTextureOwnerId, bgra,
+                    forceAsyncPresent, padding1, padding2);
+  }
 };
 
 // -
@@ -728,8 +856,11 @@ struct LinkResult final {
 
 /// 4x32-bit primitives, with a type tag.
 struct TypedQuad final {
-  alignas(alignof(float)) uint8_t data[4 * sizeof(float)] = {};
+  alignas(alignof(float)) std::array<uint8_t, 4 * sizeof(float)> data = {};
   webgl::AttribBaseType type = webgl::AttribBaseType::Float;
+  uint8_t padding[3] = {};
+
+  constexpr auto MutTiedFields() { return std::tie(data, type, padding); }
 };
 
 /// [1-16]x32-bit primitives, with a type tag.
@@ -749,7 +880,7 @@ struct ReadPixelsResult {
 };
 
 struct ReadPixelsResultIpc final : public ReadPixelsResult {
-  mozilla::ipc::Shmem shmem = {};
+  Maybe<mozilla::ipc::Shmem> shmem = {};
 };
 
 struct VertAttribPointerDesc final {
@@ -759,6 +890,11 @@ struct VertAttribPointerDesc final {
   uint8_t byteStrideOrZero = 0;
   GLenum type = LOCAL_GL_FLOAT;
   uint64_t byteOffset = 0;
+
+  auto MutTiedFields() {
+    return std::tie(intFunc, channels, normalized, byteStrideOrZero, type,
+                    byteOffset);
+  }
 };
 
 struct VertAttribPointerCalculated final {
@@ -769,48 +905,10 @@ struct VertAttribPointerCalculated final {
 
 }  // namespace webgl
 
-/**
- * Represents a block of memory that it may or may not own.  The
- * inner data type must be trivially copyable by memcpy.
- */
-template <typename T = uint8_t>
-class RawBuffer final {
-  const T* mBegin = nullptr;
-  size_t mLen = 0;
-  UniqueBuffer mOwned;
-
- public:
-  using ElementType = T;
-
-  /**
-   * If aTakeData is true, RawBuffer will delete[] the memory when destroyed.
-   */
-  explicit RawBuffer(const Range<const T>& data, UniqueBuffer&& owned = {})
-      : mBegin(data.begin().get()),
-        mLen(data.length()),
-        mOwned(std::move(owned)) {}
-
-  explicit RawBuffer(const size_t len) : mLen(len) {}
-
-  ~RawBuffer() = default;
-
-  Range<const T> Data() const { return {mBegin, mLen}; }
-  const auto& begin() const { return mBegin; }
-  const auto& size() const { return mLen; }
-
-  void Shrink(const size_t newLen) {
-    if (mLen <= newLen) return;
-    mLen = newLen;
-  }
-
-  RawBuffer() = default;
-
-  RawBuffer(const RawBuffer&) = delete;
-  RawBuffer& operator=(const RawBuffer&) = delete;
-
-  RawBuffer(RawBuffer&&) = default;
-  RawBuffer& operator=(RawBuffer&&) = default;
-};
+template <class T>
+inline Range<T> ShmemRange(const mozilla::ipc::Shmem& shmem) {
+  return {shmem.get<T>(), shmem.Size<T>()};
+}
 
 // -
 
@@ -842,6 +940,22 @@ inline Maybe<T> MaybeAs(const U val) {
 
 // -
 
+inline GLenum IsTexImageTarget(const GLenum imageTarget) {
+  switch (imageTarget) {
+    case LOCAL_GL_TEXTURE_2D:
+    case LOCAL_GL_TEXTURE_CUBE_MAP_NEGATIVE_X:
+    case LOCAL_GL_TEXTURE_CUBE_MAP_POSITIVE_X:
+    case LOCAL_GL_TEXTURE_CUBE_MAP_NEGATIVE_Y:
+    case LOCAL_GL_TEXTURE_CUBE_MAP_POSITIVE_Y:
+    case LOCAL_GL_TEXTURE_CUBE_MAP_NEGATIVE_Z:
+    case LOCAL_GL_TEXTURE_CUBE_MAP_POSITIVE_Z:
+    case LOCAL_GL_TEXTURE_3D:
+    case LOCAL_GL_TEXTURE_2D_ARRAY:
+      return true;
+  }
+  return false;
+}
+
 inline GLenum ImageToTexTarget(const GLenum imageTarget) {
   switch (imageTarget) {
     case LOCAL_GL_TEXTURE_CUBE_MAP_NEGATIVE_X:
@@ -851,9 +965,11 @@ inline GLenum ImageToTexTarget(const GLenum imageTarget) {
     case LOCAL_GL_TEXTURE_CUBE_MAP_NEGATIVE_Z:
     case LOCAL_GL_TEXTURE_CUBE_MAP_POSITIVE_Z:
       return LOCAL_GL_TEXTURE_CUBE_MAP;
-    default:
-      return imageTarget;
   }
+  if (IsTexImageTarget(imageTarget)) {
+    return imageTarget;
+  }
+  return 0;
 }
 
 inline bool IsTexTarget3D(const GLenum texTarget) {
@@ -874,6 +990,7 @@ class Element;
 class ImageBitmap;
 class ImageData;
 class OffscreenCanvas;
+class VideoFrame;
 }  // namespace dom
 
 struct TexImageSource {
@@ -888,6 +1005,8 @@ struct TexImageSource {
 
   const dom::OffscreenCanvas* mOffscreenCanvas = nullptr;
 
+  const dom::VideoFrame* mVideoFrame = nullptr;
+
   const dom::Element* mDomElem = nullptr;
   ErrorResult* mOut_error = nullptr;
 };
@@ -910,12 +1029,14 @@ struct PixelPackingState : public DeriveNotEq<PixelPackingState> {
   uint32_t skipRows = 0;
   uint32_t skipImages = 0;
 
-  // C++20's default comparison operators can't come soon enough!
-  bool operator==(const PixelPackingState& rhs) const {
-    return alignmentInTypeElems == rhs.alignmentInTypeElems &&
-           rowLength == rhs.rowLength && imageHeight == rhs.imageHeight &&
-           skipPixels == rhs.skipPixels && skipRows == rhs.skipRows &&
-           skipImages == rhs.skipImages;
+  auto MutTiedFields() {
+    return std::tie(alignmentInTypeElems, rowLength, imageHeight, skipPixels,
+                    skipRows, skipImages);
+  }
+
+  using Self = PixelPackingState;
+  friend bool operator==(const Self& a, const Self& b) {
+    return TiedFields(a) == TiedFields(b);
   }
 
   static void AssertDefaultUnpack(gl::GLContext& gl, const bool isWebgl2) {
@@ -933,6 +1054,13 @@ struct PixelUnpackStateWebgl final : public PixelPackingState {
   bool flipY = false;
   bool premultiplyAlpha = false;
   bool requireFastPath = false;
+  uint8_t padding = {};
+
+  auto MutTiedFields() {
+    return std::tuple_cat(PixelPackingState::MutTiedFields(),
+                          std::tie(colorspaceConversion, flipY,
+                                   premultiplyAlpha, requireFastPath, padding));
+  }
 };
 
 struct ExplicitPixelPackingState final {
@@ -966,6 +1094,8 @@ struct ReadPixelsDesc final {
   uvec2 size;
   PackingInfo pi = {LOCAL_GL_RGBA, LOCAL_GL_UNSIGNED_BYTE};
   PixelPackingState packState;
+
+  auto MutTiedFields() { return std::tie(srcOffset, size, pi, packState); }
 };
 
 }  // namespace webgl
@@ -977,7 +1107,7 @@ struct TexUnpackBlobDesc final {
   uvec3 size;
   gfxAlphaType srcAlphaType = gfxAlphaType::NonPremult;
 
-  Maybe<RawBuffer<>> cpuData;
+  Maybe<Span<const uint8_t>> cpuData;
   Maybe<uint64_t> pboOffset;
 
   Maybe<uvec2> structuredSrcSize;
@@ -1015,19 +1145,6 @@ inline Range<const T> MakeRange(const dom::Sequence<T>& seq) {
   return {seq.Elements(), seq.Length()};
 }
 
-template <typename T>
-inline Range<const T> MakeRange(const RawBuffer<T>& from) {
-  return from.Data();
-}
-
-// abv = ArrayBufferView
-template <typename T>
-inline auto MakeRangeAbv(const T& abv)
-    -> Range<const typename T::element_type> {
-  abv.ComputeState();
-  return {abv.Data(), abv.Length()};
-}
-
 // -
 
 constexpr auto kUniversalAlignment = alignof(std::max_align_t);
@@ -1046,17 +1163,6 @@ inline size_t ByteSize(const Range<T>& range) {
   return range.length() * sizeof(T);
 }
 
-Maybe<Range<const uint8_t>> GetRangeFromView(const dom::ArrayBufferView& view,
-                                             GLuint elemOffset,
-                                             GLuint elemCountOverride);
-
-// -
-
-template <typename T>
-RawBuffer<T> RawBufferView(const Range<T>& range) {
-  return RawBuffer<T>{range};
-}
-
 // -
 
 Maybe<webgl::ErrorInfo> CheckBindBufferRange(
@@ -1089,6 +1195,40 @@ inline void Memcpy(const RangedPtr<uint8_t>& destBytes,
   memcpy(destBytes.get(), srcBytes.get(), byteSize);
 }
 
+template <class T, class U>
+inline void Memcpy(const Range<T>* const destRange,
+                   const RangedPtr<U>& srcBegin) {
+  Memcpy(destRange->begin(), srcBegin, destRange->length());
+}
+template <class T, class U>
+inline void Memcpy(const RangedPtr<T>* const destBegin,
+                   const Range<U>& srcRange) {
+  Memcpy(destBegin, srcRange->begin(), srcRange->length());
+}
+
+template <typename Dst, typename Src>
+inline void Memcpy(const Span<Dst>* const dest, const Span<Src>& src) {
+  MOZ_RELEASE_ASSERT(src.size_bytes() >= dest->size_bytes());
+  MOZ_ASSERT(src.size_bytes() == dest->size_bytes());
+  memcpy(dest->data(), src.data(), dest->size_bytes());
+}
+
+// -
+
+inline bool StartsWith(const std::string_view str,
+                       const std::string_view part) {
+  return str.find(part) == 0;
+}
+
+// -
+
+template <class T>
+Maybe<T> AsValidEnum(const std::underlying_type_t<T> raw_val) {
+  const auto raw_enum = T{raw_val};  // This is the risk we prevent!
+  if (!IsEnumCase(raw_enum)) return {};
+  return Some(raw_enum);
+}
+
 // -
 
 namespace webgl {
@@ -1098,8 +1238,121 @@ namespace webgl {
 // now.
 // (http://opengl.gpuinfo.org/gl_stats_caps_single.php?listreportsbycap=GL_MAX_COLOR_ATTACHMENTS)
 inline constexpr size_t kMaxDrawBuffers = 8;
+
+union UniformDataVal {
+  float f32;
+  int32_t i32;
+  uint32_t u32;
+};
+
+enum class ProvokingVertex : GLenum {
+  FirstVertex = LOCAL_GL_FIRST_VERTEX_CONVENTION,
+  LastVertex = LOCAL_GL_LAST_VERTEX_CONVENTION,
+};
+inline constexpr bool IsEnumCase(const ProvokingVertex raw) {
+  switch (raw) {
+    case ProvokingVertex::FirstVertex:
+    case ProvokingVertex::LastVertex:
+      return true;
+  }
+  return false;
+}
+
+template <class E>
+inline constexpr std::optional<E> AsEnumCase(
+    const std::underlying_type_t<E> raw) {
+  const auto ret = static_cast<E>(raw);
+  if (!IsEnumCase(ret)) return {};
+  return ret;
+}
+
+// -
+
+struct BufferAndIndex final {
+  const WebGLBuffer* buffer = nullptr;
+  uint32_t id = -1;
+};
+
 }  // namespace webgl
 
+struct IndexedBufferBinding final {
+  RefPtr<WebGLBuffer> mBufferBinding;
+  uint64_t mRangeStart = 0;
+  uint64_t mRangeSize = 0;
+
+  IndexedBufferBinding();
+  ~IndexedBufferBinding();
+
+  uint64_t ByteCount() const;
+};
+
+// -
+
+template <class... Args>
+inline std::string PrintfStdString(const char* const format,
+                                   const Args&... args) {
+  const auto nsStr = nsPrintfCString(format, args...);
+  return ToString(nsStr);
+}
+
+inline const char* ToChars(const bool val) {
+  if (val) return "true";
+  return "false";
+}
+
+template <class To>
+struct ReinterpretToSpan {
+  template <class FromT>
+  static inline constexpr Span<To> From(const Span<FromT>& from) {
+    static_assert(sizeof(FromT) == sizeof(To));
+    return {reinterpret_cast<To*>(from.data()), from.size()};
+  }
+};
+
+// -
+
+inline std::string Join(Span<const std::string> ss,
+                        const std::string_view& delim) {
+  if (!ss.size()) return "";
+  auto ret = std::string();
+  {
+    auto chars = delim.size() * (ss.size() - 1);
+    for (const auto& s : ss) {
+      chars += s.size();
+    }
+    ret.reserve(chars);
+  }
+
+  ret = ss[0];
+  ss = ss.subspan(1);
+  for (const auto& s : ss) {
+    ret += delim;
+    ret += s;
+  }
+  return ret;
+}
+
+inline std::string ToStringWithCommas(uint64_t v) {
+  if (!v) return "0";
+  std::vector<std::string> chunks;
+  while (v) {
+    const auto chunk = v % 1000;
+    v /= 1000;
+    chunks.insert(chunks.begin(), std::to_string(chunk));
+  }
+  return Join(chunks, ",");
+}
+
+// -
+
+namespace webgl {
+
+std::unordered_map<GLenum, bool> MakeIsEnabledMap(bool webgl2);
+
+static constexpr uint32_t kMaxClientWaitSyncTimeoutNS =
+    1000 * 1000 * 1000;  // 1000ms in ns.
+
+}  // namespace webgl
 }  // namespace mozilla
 
 #endif
