@@ -31,23 +31,28 @@
 #include "nsIScriptContext.h"
 #include "mozilla/dom/Document.h"
 #include "nsIScriptGlobalObject.h"
+#include "nsQueryObject.h"
 #include "nsVariant.h"
 #include "mozilla/dom/ContentChild.h"
 #include "mozilla/dom/DataTransferBinding.h"
 #include "mozilla/dom/DataTransferItemList.h"
 #include "mozilla/dom/Directory.h"
 #include "mozilla/dom/Element.h"
+#include "mozilla/dom/Event.h"
 #include "mozilla/dom/FileList.h"
+#include "mozilla/dom/IPCBlobUtils.h"
 #include "mozilla/dom/BindingUtils.h"
 #include "mozilla/dom/OSFileSystem.h"
 #include "mozilla/dom/Promise.h"
+#include "mozilla/dom/WindowContext.h"
+#include "mozilla/Unused.h"
 #include "nsComponentManagerUtils.h"
 #include "nsNetUtil.h"
 #include "nsReadableUtils.h"
 
 namespace mozilla::dom {
 
-NS_IMPL_CYCLE_COLLECTION_CLASS(DataTransfer)
+NS_IMPL_CYCLE_COLLECTION_WRAPPERCACHE_CLASS(DataTransfer)
 
 NS_IMPL_CYCLE_COLLECTION_UNLINK_BEGIN(DataTransfer)
   NS_IMPL_CYCLE_COLLECTION_UNLINK(mParent)
@@ -62,7 +67,6 @@ NS_IMPL_CYCLE_COLLECTION_TRAVERSE_BEGIN(DataTransfer)
   NS_IMPL_CYCLE_COLLECTION_TRAVERSE(mDragTarget)
   NS_IMPL_CYCLE_COLLECTION_TRAVERSE(mDragImage)
 NS_IMPL_CYCLE_COLLECTION_TRAVERSE_END
-NS_IMPL_CYCLE_COLLECTION_TRACE_WRAPPERCACHE(DataTransfer)
 
 NS_IMPL_CYCLE_COLLECTING_ADDREF(DataTransfer)
 NS_IMPL_CYCLE_COLLECTING_RELEASE(DataTransfer)
@@ -202,7 +206,7 @@ DataTransfer::DataTransfer(nsISupports* aParent, EventMessage aEventMessage,
                            bool aIsCrossDomainSubFrameDrop,
                            int32_t aClipboardType, DataTransferItemList* aItems,
                            Element* aDragImage, uint32_t aDragImageX,
-                           uint32_t aDragImageY)
+                           uint32_t aDragImageY, bool aShowFailAnimation)
     : mParent(aParent),
       mDropEffect(nsIDragService::DRAGDROP_ACTION_NONE),
       mEffectAllowed(aEffectAllowed),
@@ -215,7 +219,8 @@ DataTransfer::DataTransfer(nsISupports* aParent, EventMessage aEventMessage,
       mClipboardType(aClipboardType),
       mDragImage(aDragImage),
       mDragImageX(aDragImageX),
-      mDragImageY(aDragImageY) {
+      mDragImageY(aDragImageY),
+      mShowFailAnimation(aShowFailAnimation) {
   MOZ_ASSERT(mParent);
   MOZ_ASSERT(aItems);
 
@@ -436,8 +441,20 @@ already_AddRefed<nsINode> DataTransfer::GetMozSourceNode() {
   return sourceNode.forget();
 }
 
+already_AddRefed<WindowContext> DataTransfer::GetSourceTopWindowContext() {
+  nsCOMPtr<nsIDragSession> dragSession = nsContentUtils::GetDragSession();
+  if (!dragSession) {
+    return nullptr;
+  }
+
+  RefPtr<WindowContext> sourceTopWindowContext;
+  dragSession->GetSourceTopWindowContext(
+      getter_AddRefs(sourceTopWindowContext));
+  return sourceTopWindowContext.forget();
+}
+
 already_AddRefed<DOMStringList> DataTransfer::MozTypesAt(
-    uint32_t aIndex, CallerType aCallerType, ErrorResult& aRv) const {
+    uint32_t aIndex, ErrorResult& aRv) const {
   // Only the first item is valid for clipboard events
   if (aIndex > 0 && (mEventMessage == eCut || mEventMessage == eCopy ||
                      mEventMessage == ePaste)) {
@@ -453,10 +470,6 @@ already_AddRefed<DOMStringList> DataTransfer::MozTypesAt(
 
     bool addFile = false;
     for (uint32_t i = 0; i < items.Length(); i++) {
-      if (items[i]->ChromeOnly() && aCallerType != CallerType::System) {
-        continue;
-      }
-
       // NOTE: The reason why we get the internal type here is because we want
       // kFileMime to appear in the types list for backwards compatibility
       // reasons.
@@ -538,10 +551,9 @@ nsresult DataTransfer::GetDataAtInternal(const nsAString& aFormat,
 void DataTransfer::MozGetDataAt(JSContext* aCx, const nsAString& aFormat,
                                 uint32_t aIndex,
                                 JS::MutableHandle<JS::Value> aRetval,
-                                nsIPrincipal& aSubjectPrincipal,
                                 mozilla::ErrorResult& aRv) {
   nsCOMPtr<nsIVariant> data;
-  aRv = GetDataAtInternal(aFormat, aIndex, &aSubjectPrincipal,
+  aRv = GetDataAtInternal(aFormat, aIndex, nsContentUtils::GetSystemPrincipal(),
                           getter_AddRefs(data));
   if (aRv.Failed()) {
     return;
@@ -570,21 +582,17 @@ bool DataTransfer::PrincipalMaySetData(const nsAString& aType,
       return false;
     }
 
-    if (aType.EqualsASCII(kFileMime) || aType.EqualsASCII(kFilePromiseMime)) {
-      NS_WARNING(
-          "Disallowing adding x-moz-file or x-moz-file-promize types to "
-          "DataTransfer");
-      return false;
-    }
-
-    // Disallow content from creating x-moz-place flavors, so that it cannot
-    // create fake Places smart queries exposing user data, but give a free
-    // pass to WebExtensions.
-    auto principal = BasePrincipal::Cast(aPrincipal);
-    if (!principal->AddonPolicy() &&
-        StringBeginsWith(aType, u"text/x-moz-place"_ns)) {
-      NS_WARNING("Disallowing adding moz-place types to DataTransfer");
-      return false;
+    // Don't allow adding internal types of the form */x-moz-*, but
+    // special-case the url types as they are simple variations of urls.
+    // In addition, allow x-moz-place flavors to be added by WebExtensions.
+    if (FindInReadable(kInternal_Mimetype_Prefix, aType) &&
+        !StringBeginsWith(aType, u"text/x-moz-url"_ns)) {
+      auto principal = BasePrincipal::Cast(aPrincipal);
+      if (!principal->AddonPolicy() ||
+          !StringBeginsWith(aType, u"text/x-moz-place"_ns)) {
+        NS_WARNING("Disallowing adding this type to DataTransfer");
+        return false;
+      }
     }
   }
 
@@ -613,49 +621,65 @@ already_AddRefed<DataTransfer> DataTransfer::MozCloneForEvent(
   return dt.forget();
 }
 
-/* static */
-void DataTransfer::GetExternalClipboardFormats(const int32_t& aWhichClipboard,
-                                               const bool& aPlainTextOnly,
-                                               nsTArray<nsCString>* aResult) {
-  MOZ_ASSERT(aResult);
+// The order of the types matters. `kFileMime` needs to be one of the first two
+// types. And the order should be the same as the types order defined in
+// MandatoryDataTypesAsCStrings() for Clipboard API.
+static const nsCString kNonPlainTextExternalFormats[] = {
+    nsLiteralCString(kCustomTypesMime), nsLiteralCString(kFileMime),
+    nsLiteralCString(kHTMLMime),        nsLiteralCString(kRTFMime),
+    nsLiteralCString(kURLMime),         nsLiteralCString(kURLDataMime),
+    nsLiteralCString(kTextMime),        nsLiteralCString(kPNGImageMime),
+    nsLiteralCString(kPDFJSMime)};
 
+void DataTransfer::GetExternalClipboardFormats(const bool& aPlainTextOnly,
+                                               nsTArray<nsCString>& aResult) {
   // NOTE: When you change this method, you may need to change
   //       GetExternalTransferableFormats() too since those methods should
   //       work similarly.
 
+  MOZ_ASSERT(!mAsyncGetClipboardData);
+
+  RefPtr<WindowContext> wc = GetWindowContext();
+  if (NS_WARN_IF(!wc)) {
+    MOZ_ASSERT_UNREACHABLE(
+        "How could this DataTransfer be created with a non-window global?");
+    return;
+  }
+
   nsCOMPtr<nsIClipboard> clipboard =
       do_GetService("@mozilla.org/widget/clipboard;1");
-  if (!clipboard || aWhichClipboard < 0) {
+  if (!clipboard || mClipboardType < 0) {
     return;
   }
 
+  nsresult rv = NS_ERROR_FAILURE;
+  nsCOMPtr<nsIAsyncGetClipboardData> asyncGetClipboardData;
   if (aPlainTextOnly) {
-    bool hasType;
-    AutoTArray<nsCString, 1> unicodeMime = {nsDependentCString(kUnicodeMime)};
-    nsresult rv = clipboard->HasDataMatchingFlavors(unicodeMime,
-                                                    aWhichClipboard, &hasType);
-    NS_SUCCEEDED(rv);
-    if (hasType) {
-      aResult->AppendElement(kUnicodeMime);
-    }
+    rv = clipboard->GetDataSnapshotSync(
+        AutoTArray<nsCString, 1>{nsLiteralCString(kTextMime)}, mClipboardType,
+        wc, getter_AddRefs(asyncGetClipboardData));
+  } else {
+    AutoTArray<nsCString, ArrayLength(kNonPlainTextExternalFormats)> formats;
+    formats.AppendElements(Span<const nsCString>(kNonPlainTextExternalFormats));
+    rv = clipboard->GetDataSnapshotSync(formats, mClipboardType, wc,
+                                        getter_AddRefs(asyncGetClipboardData));
+  }
+
+  if (NS_FAILED(rv) || !asyncGetClipboardData) {
     return;
   }
 
-  // If not plain text only, then instead check all the other types
-  static const char* formats[] = {kCustomTypesMime, kFileMime,    kHTMLMime,
-                                  kRTFMime,         kURLMime,     kURLDataMime,
-                                  kUnicodeMime,     kPNGImageMime};
-
-  for (uint32_t f = 0; f < mozilla::ArrayLength(formats); ++f) {
-    bool hasType;
-    AutoTArray<nsCString, 1> format = {nsDependentCString(formats[f])};
-    nsresult rv =
-        clipboard->HasDataMatchingFlavors(format, aWhichClipboard, &hasType);
-    NS_SUCCEEDED(rv);
-    if (hasType) {
-      aResult->AppendElement(formats[f]);
+  // Order is important for DataTransfer; ensure the returned list items follow
+  // the sequence specified in kNonPlainTextExternalFormats.
+  AutoTArray<nsCString, ArrayLength(kNonPlainTextExternalFormats)> flavors;
+  asyncGetClipboardData->GetFlavorList(flavors);
+  for (const auto& format : kNonPlainTextExternalFormats) {
+    if (flavors.Contains(format)) {
+      aResult.AppendElement(format);
     }
   }
+
+  mAsyncGetClipboardData = asyncGetClipboardData;
 }
 
 /* static */
@@ -675,22 +699,18 @@ void DataTransfer::GetExternalTransferableFormats(
   aTransferable->FlavorsTransferableCanExport(flavors);
 
   if (aPlainTextOnly) {
-    auto index = flavors.IndexOf(nsLiteralCString(kUnicodeMime));
+    auto index = flavors.IndexOf(nsLiteralCString(kTextMime));
     if (index != flavors.NoIndex) {
-      aResult->AppendElement(nsLiteralCString(kUnicodeMime));
+      aResult->AppendElement(nsLiteralCString(kTextMime));
     }
     return;
   }
 
   // If not plain text only, then instead check all the other types
-  static const char* formats[] = {kCustomTypesMime, kFileMime,    kHTMLMime,
-                                  kRTFMime,         kURLMime,     kURLDataMime,
-                                  kUnicodeMime,     kPNGImageMime};
-
-  for (const char* format : formats) {
-    auto index = flavors.IndexOf(nsCString(format));
+  for (const auto& format : kNonPlainTextExternalFormats) {
+    auto index = flavors.IndexOf(format);
     if (index != flavors.NoIndex) {
-      aResult->AppendElement(nsCString(format));
+      aResult->AppendElement(format);
     }
   }
 }
@@ -732,18 +752,17 @@ nsresult DataTransfer::SetDataAtInternal(const nsAString& aFormat,
 
 void DataTransfer::MozSetDataAt(JSContext* aCx, const nsAString& aFormat,
                                 JS::Handle<JS::Value> aData, uint32_t aIndex,
-                                nsIPrincipal& aSubjectPrincipal,
                                 ErrorResult& aRv) {
   nsCOMPtr<nsIVariant> data;
   aRv = nsContentUtils::XPConnect()->JSValToVariant(aCx, aData,
                                                     getter_AddRefs(data));
   if (!aRv.Failed()) {
-    aRv = SetDataAtInternal(aFormat, data, aIndex, &aSubjectPrincipal);
+    aRv = SetDataAtInternal(aFormat, data, aIndex,
+                            nsContentUtils::GetSystemPrincipal());
   }
 }
 
 void DataTransfer::MozClearDataAt(const nsAString& aFormat, uint32_t aIndex,
-                                  nsIPrincipal& aSubjectPrincipal,
                                   ErrorResult& aRv) {
   if (IsReadOnly()) {
     aRv.Throw(NS_ERROR_DOM_NO_MODIFICATION_ALLOWED_ERR);
@@ -762,7 +781,8 @@ void DataTransfer::MozClearDataAt(const nsAString& aFormat, uint32_t aIndex,
     return;
   }
 
-  MozClearDataAtHelper(aFormat, aIndex, aSubjectPrincipal, aRv);
+  MozClearDataAtHelper(aFormat, aIndex, *nsContentUtils::GetSystemPrincipal(),
+                       aRv);
 
   // If we just cleared the 0-th index, and there are still more than 1 indexes
   // remaining, MozClearDataAt should cause the 1st index to become the 0th
@@ -811,48 +831,6 @@ void DataTransfer::UpdateDragImage(Element& aImage, int32_t aX, int32_t aY) {
   }
 }
 
-already_AddRefed<Promise> DataTransfer::GetFilesAndDirectories(
-    nsIPrincipal& aSubjectPrincipal, ErrorResult& aRv) {
-  nsCOMPtr<nsINode> parentNode = do_QueryInterface(mParent);
-  if (!parentNode) {
-    aRv.Throw(NS_ERROR_FAILURE);
-    return nullptr;
-  }
-
-  nsCOMPtr<nsIGlobalObject> global = parentNode->OwnerDoc()->GetScopeObject();
-  MOZ_ASSERT(global);
-  if (!global) {
-    aRv.Throw(NS_ERROR_FAILURE);
-    return nullptr;
-  }
-
-  RefPtr<Promise> p = Promise::Create(global, aRv);
-  if (NS_WARN_IF(aRv.Failed())) {
-    return nullptr;
-  }
-
-  RefPtr<FileList> files = mItems->Files(&aSubjectPrincipal);
-  if (NS_WARN_IF(!files)) {
-    return nullptr;
-  }
-
-  Sequence<RefPtr<File>> filesSeq;
-  files->ToSequence(filesSeq, aRv);
-  if (NS_WARN_IF(aRv.Failed())) {
-    return nullptr;
-  }
-
-  p->MaybeResolve(filesSeq);
-
-  return p.forget();
-}
-
-already_AddRefed<Promise> DataTransfer::GetFiles(
-    bool aRecursiveFlag, nsIPrincipal& aSubjectPrincipal, ErrorResult& aRv) {
-  // Currently we don't support directories.
-  return GetFilesAndDirectories(aSubjectPrincipal, aRv);
-}
-
 void DataTransfer::AddElement(Element& aElement, ErrorResult& aRv) {
   if (IsReadOnly()) {
     aRv.Throw(NS_ERROR_DOM_NO_MODIFICATION_ALLOWED_ERR);
@@ -869,7 +847,7 @@ nsresult DataTransfer::Clone(nsISupports* aParent, EventMessage aEventMessage,
   RefPtr<DataTransfer> newDataTransfer = new DataTransfer(
       aParent, aEventMessage, mEffectAllowed, mCursorState, mIsExternal,
       aUserCancelled, aIsCrossDomainSubFrameDrop, mClipboardType, mItems,
-      mDragImage, mDragImageX, mDragImageY);
+      mDragImage, mDragImageX, mDragImageY, mShowFailAnimation);
 
   newDataTransfer.forget(aNewDataTransfer);
   return NS_OK;
@@ -924,6 +902,17 @@ already_AddRefed<nsITransferable> DataTransfer::GetTransferable(
   }
   transferable->Init(aLoadContext);
 
+  // Set the principal of the global this DataTransfer was created for
+  // on the transferable for ReadWrite events (copy, cut, or dragstart).
+  //
+  // For other events, the data inside the transferable may originate
+  // from another origin or from the OS.
+  if (mMode == Mode::ReadWrite) {
+    if (nsCOMPtr<nsIGlobalObject> global = GetGlobal()) {
+      transferable->SetDataPrincipal(global->PrincipalOrNull());
+    }
+  }
+
   nsCOMPtr<nsIStorageStream> storageStream;
   nsCOMPtr<nsIObjectOutputStream> stream;
 
@@ -938,28 +927,6 @@ already_AddRefed<nsITransferable> DataTransfer::GetTransferable(
   const uint32_t baseLength = sizeof(uint32_t) + 1;
   uint32_t totalCustomLength = baseLength;
 
-  const char* knownFormats[] = {kTextMime,
-                                kHTMLMime,
-                                kNativeHTMLMime,
-                                kRTFMime,
-                                kURLMime,
-                                kURLDataMime,
-                                kURLDescriptionMime,
-                                kURLPrivateMime,
-                                kPNGImageMime,
-                                kJPEGImageMime,
-                                kGIFImageMime,
-                                kNativeImageMime,
-                                kFileMime,
-                                kFilePromiseMime,
-                                kFilePromiseURLMime,
-                                kFilePromiseDestFilename,
-                                kFilePromiseDirectoryMime,
-                                kMozTextInternal,
-                                kHTMLContext,
-                                kHTMLInfo,
-                                kImageRequestMime};
-
   /*
    * Two passes are made here to iterate over all of the types. First, look for
    * any types that are not in the list of known types. For this pass,
@@ -995,8 +962,8 @@ already_AddRefed<nsITransferable> DataTransfer::GetTransferable(
 
       // If the data is of one of the well-known formats, use it directly.
       bool isCustomFormat = true;
-      for (uint32_t f = 0; f < ArrayLength(knownFormats); f++) {
-        if (type.EqualsASCII(knownFormats[f])) {
+      for (const char* format : kKnownFormats) {
+        if (type.EqualsASCII(format)) {
           isCustomFormat = false;
           break;
         }
@@ -1101,8 +1068,8 @@ already_AddRefed<nsITransferable> DataTransfer::GetTransferable(
             nsCOMPtr<nsIInputStream> inputStream;
             storageStream->NewInputStream(0, getter_AddRefs(inputStream));
 
-            RefPtr<nsStringBuffer> stringBuffer =
-                nsStringBuffer::Alloc(totalCustomLength);
+            RefPtr<StringBuffer> stringBuffer =
+                StringBuffer::Alloc(totalCustomLength);
 
             // Subtract off the null terminator when reading.
             totalCustomLength--;
@@ -1116,7 +1083,7 @@ already_AddRefed<nsITransferable> DataTransfer::GetTransferable(
               static_cast<char*>(stringBuffer->Data())[amountRead] = 0;
 
               nsCString str;
-              stringBuffer->ToString(totalCustomLength, str);
+              str.Assign(stringBuffer, totalCustomLength);
               nsCOMPtr<nsISupportsCString> strSupports(
                   do_CreateInstance(NS_SUPPORTS_CSTRING_CONTRACTID));
               strSupports->SetData(str);
@@ -1142,27 +1109,20 @@ already_AddRefed<nsITransferable> DataTransfer::GetTransferable(
           continue;
         }
 
-        // The underlying drag code uses text/unicode, so use that instead of
-        // text/plain
-        const char* format;
-        NS_ConvertUTF16toUTF8 utf8format(type);
-        if (utf8format.EqualsLiteral(kTextMime)) {
-          format = kUnicodeMime;
-        } else {
-          format = utf8format.get();
-        }
+        NS_ConvertUTF16toUTF8 format(type);
 
         // If a converter is set for a format, set the converter for the
         // transferable and don't add the item
         nsCOMPtr<nsIFormatConverter> converter =
             do_QueryInterface(convertedData);
         if (converter) {
-          transferable->AddDataFlavor(format);
+          transferable->AddDataFlavor(format.get());
           transferable->SetConverter(converter);
           continue;
         }
 
-        nsresult rv = transferable->SetTransferData(format, convertedData);
+        nsresult rv =
+            transferable->SetTransferData(format.get(), convertedData);
         if (NS_FAILED(rv)) {
           return nullptr;
         }
@@ -1196,16 +1156,22 @@ bool DataTransfer::ConvertFromVariant(nsIVariant* aVariant,
       return false;
     }
 
-    nsCOMPtr<nsIFlavorDataProvider> fdp = do_QueryInterface(data);
-    if (fdp) {
-      // For flavour data providers, use 0 as the length.
+    // For flavour data providers, use 0 as the length.
+    if (nsCOMPtr<nsIFlavorDataProvider> fdp = do_QueryInterface(data)) {
       fdp.forget(aSupports);
       *aLength = 0;
-    } else {
-      data.forget(aSupports);
-      *aLength = sizeof(nsISupports*);
+      return true;
     }
 
+    // Only use the underlying BlobImpl for transferables.
+    if (RefPtr<Blob> blob = do_QueryObject(data)) {
+      RefPtr<BlobImpl> blobImpl = blob->Impl();
+      blobImpl.forget(aSupports);
+    } else {
+      data.forget(aSupports);
+    }
+
+    *aLength = sizeof(nsISupports*);
     return true;
   }
 
@@ -1238,7 +1204,10 @@ void DataTransfer::Disconnect() {
   }
 }
 
-void DataTransfer::ClearAll() { mItems->ClearAllItems(); }
+void DataTransfer::ClearAll() {
+  mItems->ClearAllItems();
+  mAsyncGetClipboardData = nullptr;
+}
 
 uint32_t DataTransfer::MozItemCount() const { return mItems->MozItemCount(); }
 
@@ -1277,7 +1246,7 @@ void DataTransfer::SetDataWithPrincipalFromOtherProcess(
 
 void DataTransfer::GetRealFormat(const nsAString& aInFormat,
                                  nsAString& aOutFormat) const {
-  // treat text/unicode as equivalent to text/plain
+  // For compatibility, treat text/unicode as equivalent to text/plain
   nsAutoString lowercaseFormat;
   nsContentUtils::ASCIIToLower(aInFormat, lowercaseFormat);
   if (lowercaseFormat.EqualsLiteral("text") ||
@@ -1294,13 +1263,43 @@ void DataTransfer::GetRealFormat(const nsAString& aInFormat,
   aOutFormat.Assign(lowercaseFormat);
 }
 
+already_AddRefed<nsIGlobalObject> DataTransfer::GetGlobal() const {
+  nsCOMPtr<nsIGlobalObject> global;
+  // This is annoying, but DataTransfer may have various things as parent.
+  if (nsCOMPtr<EventTarget> target = do_QueryInterface(mParent)) {
+    global = target->GetOwnerGlobal();
+  } else if (RefPtr<Event> event = do_QueryObject(mParent)) {
+    global = event->GetParentObject();
+  }
+
+  return global.forget();
+}
+
+already_AddRefed<WindowContext> DataTransfer::GetWindowContext() const {
+  nsCOMPtr<nsIGlobalObject> global = GetGlobal();
+  if (!global) {
+    return nullptr;
+  }
+
+  const auto* innerWindow = global->GetAsInnerWindow();
+  if (!innerWindow) {
+    return nullptr;
+  }
+
+  return do_AddRef(innerWindow->GetWindowContext());
+}
+
+nsIAsyncGetClipboardData* DataTransfer::GetAsyncGetClipboardData() const {
+  return mAsyncGetClipboardData;
+}
+
 nsresult DataTransfer::CacheExternalData(const char* aFormat, uint32_t aIndex,
                                          nsIPrincipal* aPrincipal,
                                          bool aHidden) {
   ErrorResult rv;
   RefPtr<DataTransferItem> item;
 
-  if (strcmp(aFormat, kUnicodeMime) == 0) {
+  if (strcmp(aFormat, kTextMime) == 0) {
     item = mItems->SetDataWithPrincipal(u"text/plain"_ns, nullptr, aIndex,
                                         aPrincipal, false, aHidden, rv);
     if (NS_WARN_IF(rv.Failed())) {
@@ -1349,8 +1348,9 @@ void DataTransfer::CacheExternalDragFormats() {
   // all platforms, so just check for the types that can actually be imported
   // XXXndeakin there are some other formats but those are platform specific.
   // NOTE: kFileMime must have index 0
-  const char* formats[] = {kFileMime,    kHTMLMime,    kURLMime,
-                           kURLDataMime, kUnicodeMime, kPNGImageMime};
+  // TODO: should this be `kNonPlainTextExternalFormats` instead?
+  static const char* formats[] = {kFileMime,    kHTMLMime, kURLMime,
+                                  kURLDataMime, kTextMime, kPNGImageMime};
 
   uint32_t count;
   dragSession->GetNumDropItems(&count);
@@ -1390,21 +1390,14 @@ void DataTransfer::CacheExternalClipboardFormats(bool aPlainTextOnly) {
                "caching clipboard data for invalid event");
 
   nsCOMPtr<nsIPrincipal> sysPrincipal = nsContentUtils::GetSystemPrincipal();
-
   nsTArray<nsCString> typesArray;
-
-  if (XRE_IsContentProcess()) {
-    ContentChild::GetSingleton()->SendGetExternalClipboardFormats(
-        mClipboardType, aPlainTextOnly, &typesArray);
-  } else {
-    GetExternalClipboardFormats(mClipboardType, aPlainTextOnly, &typesArray);
-  }
-
+  GetExternalClipboardFormats(aPlainTextOnly, typesArray);
   if (aPlainTextOnly) {
-    // The only thing that will be in types is kUnicodeMime
+    // The only thing that will be in types is kTextMime
     MOZ_ASSERT(typesArray.IsEmpty() || typesArray.Length() == 1);
     if (typesArray.Length() == 1) {
-      CacheExternalData(kUnicodeMime, 0, sysPrincipal, false);
+      MOZ_ASSERT(typesArray.Contains(kTextMime));
+      CacheExternalData(kTextMime, 0, sysPrincipal, false);
     }
     return;
   }
@@ -1427,7 +1420,8 @@ void DataTransfer::CacheExternalData(const nsTArray<nsCString>& aTypes,
   for (const nsCString& type : aTypes) {
     if (type.EqualsLiteral(kCustomTypesMime)) {
       FillInExternalCustomTypes(0, aPrincipal);
-    } else if (type.EqualsLiteral(kFileMime) && XRE_IsContentProcess()) {
+    } else if (type.EqualsLiteral(kFileMime) && XRE_IsContentProcess() &&
+               !StaticPrefs::dom_events_dataTransfer_mozFile_enabled()) {
       // We will be ignoring any application/x-moz-file files found in the paste
       // datatransfer within e10s, as they will fail top be sent over IPC.
       // Because of that, we will unset hasFileData, whether or not it would
@@ -1438,9 +1432,10 @@ void DataTransfer::CacheExternalData(const nsTArray<nsCString>& aTypes,
       // We expect that if kFileMime is supported, then it will be the either at
       // index 0 or at index 1 in the aTypes returned by
       // GetExternalClipboardFormats
-      if (type.EqualsLiteral(kFileMime) && !XRE_IsContentProcess()) {
+      if (type.EqualsLiteral(kFileMime)) {
         hasFileData = true;
       }
+
       // If we aren't the file data, and we have file data, we want to be hidden
       CacheExternalData(
           type.get(), 0, aPrincipal,
