@@ -11,18 +11,26 @@
 #include "mozilla/BasePrincipal.h"
 #include "mozilla/ContentEvents.h"
 #include "mozilla/EventForwards.h"
+#include "mozilla/StaticPrefs_dom.h"
+#include "mozilla/dom/BlobImpl.h"
 #include "mozilla/dom/DataTransferItemBinding.h"
 #include "mozilla/dom/Directory.h"
-#include "mozilla/dom/Event.h"
 #include "mozilla/dom/FileSystem.h"
 #include "mozilla/dom/FileSystemDirectoryEntry.h"
 #include "mozilla/dom/FileSystemFileEntry.h"
+#include "imgIContainer.h"
+#include "imgITools.h"
+#include "nsComponentManagerUtils.h"
+#include "nsGlobalWindowInner.h"
 #include "nsIClipboard.h"
+#include "nsIFile.h"
+#include "nsIInputStream.h"
 #include "nsISupportsPrimitives.h"
 #include "nsIScriptObjectPrincipal.h"
 #include "nsNetUtil.h"
 #include "nsQueryObject.h"
 #include "nsContentUtils.h"
+#include "nsThreadUtils.h"
 #include "nsVariant.h"
 
 namespace {
@@ -39,8 +47,7 @@ FileMimeNameData kFileMimeNameMap[] = {
 
 }  // anonymous namespace
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 NS_IMPL_CYCLE_COLLECTION_WRAPPERCACHE(DataTransferItem, mData, mPrincipal,
                                       mDataTransfer, mCachedFile)
@@ -69,6 +76,7 @@ already_AddRefed<DataTransferItem> DataTransferItem::Clone(
   it->mData = mData;
   it->mPrincipal = mPrincipal;
   it->mChromeOnly = mChromeOnly;
+  it->mDoNotAttemptToLoadData = mDoNotAttemptToLoadData;
 
   return it.forget();
 }
@@ -82,6 +90,8 @@ void DataTransferItem::SetData(nsIVariant* aData) {
     // These are provided by the system, and have guaranteed properties about
     // their kind based on their type.
     MOZ_ASSERT(!mType.IsEmpty());
+    // This type should not be provided by the OS.
+    MOZ_ASSERT(!mType.EqualsASCII(kNativeImageMime));
 
     mKind = KIND_STRING;
     for (uint32_t i = 0; i < ArrayLength(kFileMimeNameMap); ++i) {
@@ -110,6 +120,14 @@ void DataTransferItem::SetData(nsIVariant* aData) {
         nsCOMPtr<nsIFile>(do_QueryInterface(supports))) {
       return KIND_FILE;
     }
+
+    if (StaticPrefs::dom_events_dataTransfer_imageAsFile_enabled()) {
+      // Firefox internally uses imgIContainer to represent images being
+      // copied/dragged. These need to be encoded to PNG files.
+      if (nsCOMPtr<imgIContainer>(do_QueryInterface(supports))) {
+        return KIND_FILE;
+      }
+    }
   }
 
   nsAutoString string;
@@ -127,15 +145,13 @@ void DataTransferItem::SetData(nsIVariant* aData) {
 }
 
 void DataTransferItem::FillInExternalData() {
-  if (mData) {
+  if (mData || mDoNotAttemptToLoadData) {
     return;
   }
 
   NS_ConvertUTF16toUTF8 utf8format(mType);
   const char* format = utf8format.get();
-  if (strcmp(format, "text/plain") == 0) {
-    format = kUnicodeMime;
-  } else if (strcmp(format, "text/uri-list") == 0) {
+  if (strcmp(format, "text/uri-list") == 0) {
     format = kURLDataMime;
   }
 
@@ -158,8 +174,21 @@ void DataTransferItem::FillInExternalData() {
         return;
       }
 
-      nsresult rv = clipboard->GetData(trans, mDataTransfer->ClipboardType());
+      nsCOMPtr<nsIGlobalObject> global = mDataTransfer->GetGlobal();
+      WindowContext* windowContext = nullptr;
+      if (global) {
+        const auto* innerWindow = global->GetAsInnerWindow();
+        windowContext = innerWindow ? innerWindow->GetWindowContext() : nullptr;
+      }
+      MOZ_ASSERT(windowContext);
+      nsresult rv = clipboard->GetData(trans, mDataTransfer->ClipboardType(),
+                                       windowContext);
       if (NS_WARN_IF(NS_FAILED(rv))) {
+        if (rv == NS_ERROR_CONTENT_BLOCKED) {
+          // If the load of this content was blocked by Content Analysis,
+          // do not attempt to load it again.
+          mDoNotAttemptToLoadData = true;
+        }
         return;
       }
     } else {
@@ -197,7 +226,6 @@ void DataTransferItem::FillInExternalData() {
       }
       data = do_QueryObject(file);
     }
-
     variant->SetAsISupports(data);
   } else {
     // We have an external piece of string data. Extract it and store it in the
@@ -288,7 +316,7 @@ already_AddRefed<File> DataTransferItem::GetAsFile(
     if (RefPtr<Blob> blob = do_QueryObject(supports)) {
       mCachedFile = blob->ToFile();
     } else {
-      nsCOMPtr<nsIGlobalObject> global = GetGlobalFromDataTransfer();
+      nsCOMPtr<nsIGlobalObject> global = mDataTransfer->GetGlobal();
       if (NS_WARN_IF(!global)) {
         return nullptr;
       }
@@ -304,6 +332,22 @@ already_AddRefed<File> DataTransferItem::GetAsFile(
         if (NS_WARN_IF(!mCachedFile)) {
           return nullptr;
         }
+      } else if (nsCOMPtr<imgIContainer> img = do_QueryInterface(supports)) {
+        nsCOMPtr<imgITools> imgTools =
+            do_CreateInstance("@mozilla.org/image/tools;1");
+
+        nsCOMPtr<nsIInputStream> inputStream;
+        nsresult rv = imgTools->EncodeImage(img, "image/png"_ns, u""_ns,
+                                            getter_AddRefs(inputStream));
+        if (NS_WARN_IF(NS_FAILED(rv))) {
+          return nullptr;
+        }
+
+        mCachedFile = CreateFileFromInputStream(
+            inputStream, "GenericImageNamePNG", u"image/png"_ns);
+        if (NS_WARN_IF(!mCachedFile)) {
+          return nullptr;
+        }
       } else {
         MOZ_ASSERT(false, "One of the above code paths should be taken");
         return nullptr;
@@ -322,7 +366,7 @@ already_AddRefed<FileSystemEntry> DataTransferItem::GetAsEntry(
     return nullptr;
   }
 
-  nsCOMPtr<nsIGlobalObject> global = GetGlobalFromDataTransfer();
+  nsCOMPtr<nsIGlobalObject> global = mDataTransfer->GetGlobal();
   if (NS_WARN_IF(!global)) {
     return nullptr;
   }
@@ -378,9 +422,15 @@ already_AddRefed<File> DataTransferItem::CreateFileFromInputStream(
     key = "GenericFileName";
   }
 
+  return CreateFileFromInputStream(aStream, key, mType);
+}
+
+already_AddRefed<File> DataTransferItem::CreateFileFromInputStream(
+    nsIInputStream* aStream, const char* aFileNameKey,
+    const nsAString& aContentType) {
   nsAutoString fileName;
   nsresult rv = nsContentUtils::GetLocalizedString(
-      nsContentUtils::eDOM_PROPERTIES, key, fileName);
+      nsContentUtils::eDOM_PROPERTIES, aFileNameKey, fileName);
   if (NS_WARN_IF(NS_FAILED(rv))) {
     return nullptr;
   }
@@ -392,13 +442,13 @@ already_AddRefed<File> DataTransferItem::CreateFileFromInputStream(
     return nullptr;
   }
 
-  nsCOMPtr<nsIGlobalObject> global = GetGlobalFromDataTransfer();
+  nsCOMPtr<nsIGlobalObject> global = mDataTransfer->GetGlobal();
   if (NS_WARN_IF(!global)) {
     return nullptr;
   }
 
   return File::CreateMemoryFileWithLastModifiedNow(global, data, available,
-                                                   fileName, mType);
+                                                   fileName, aContentType);
 }
 
 void DataTransferItem::GetAsString(FunctionStringCallback* aCallback,
@@ -455,19 +505,8 @@ void DataTransferItem::GetAsString(FunctionStringCallback* aCallback,
 
   RefPtr<GASRunnable> runnable = new GASRunnable(aCallback, stringData);
 
-  // DataTransfer.mParent might be EventTarget, nsIGlobalObject, ClipboardEvent
-  // nsPIDOMWindowOuter, null
-  nsISupports* parent = mDataTransfer->GetParentObject();
-  nsCOMPtr<nsIGlobalObject> global = do_QueryInterface(parent);
-  if (parent && !global) {
-    if (nsCOMPtr<dom::EventTarget> target = do_QueryInterface(parent)) {
-      global = target->GetOwnerGlobal();
-    } else if (RefPtr<Event> event = do_QueryObject(parent)) {
-      global = event->GetParentObject();
-    }
-  }
-  if (global) {
-    rv = global->Dispatch(TaskCategory::Other, runnable.forget());
+  if (nsCOMPtr<nsIGlobalObject> global = mDataTransfer->GetGlobal()) {
+    rv = global->Dispatch(runnable.forget());
   } else {
     rv = NS_DispatchToMainThread(runnable);
   }
@@ -558,23 +597,4 @@ already_AddRefed<nsIVariant> DataTransferItem::Data(nsIPrincipal* aPrincipal,
   return variant.forget();
 }
 
-already_AddRefed<nsIGlobalObject>
-DataTransferItem::GetGlobalFromDataTransfer() {
-  nsCOMPtr<nsIGlobalObject> global;
-  // This is annoying, but DataTransfer may have various things as parent.
-  nsCOMPtr<EventTarget> target =
-      do_QueryInterface(mDataTransfer->GetParentObject());
-  if (target) {
-    global = target->GetOwnerGlobal();
-  } else {
-    RefPtr<Event> event = do_QueryObject(mDataTransfer->GetParentObject());
-    if (event) {
-      global = event->GetParentObject();
-    }
-  }
-
-  return global.forget();
-}
-
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
