@@ -5,6 +5,8 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "GlobalKeyListener.h"
+#include "ErrorList.h"
+#include "EventTarget.h"
 
 #include <utility>
 
@@ -12,6 +14,7 @@
 #include "mozilla/EventStateManager.h"
 #include "mozilla/HTMLEditor.h"
 #include "mozilla/KeyEventHandler.h"
+#include "mozilla/NativeKeyBindingsType.h"
 #include "mozilla/Preferences.h"
 #include "mozilla/ShortcutKeys.h"
 #include "mozilla/StaticPtr.h"
@@ -20,13 +23,16 @@
 #include "mozilla/dom/Event.h"
 #include "mozilla/dom/EventBinding.h"
 #include "mozilla/dom/KeyboardEvent.h"
+#include "mozilla/widget/IMEData.h"
 #include "nsAtom.h"
 #include "nsCOMPtr.h"
 #include "nsContentUtils.h"
 #include "nsFocusManager.h"
 #include "nsGkAtoms.h"
 #include "nsIContent.h"
+#include "nsIContentInlines.h"
 #include "nsIDocShell.h"
+#include "nsIWidget.h"
 #include "nsNetUtil.h"
 #include "nsPIDOMWindow.h"
 
@@ -34,7 +40,7 @@ namespace mozilla {
 
 using namespace mozilla::layers;
 
-GlobalKeyListener::GlobalKeyListener(EventTarget* aTarget)
+GlobalKeyListener::GlobalKeyListener(dom::EventTarget* aTarget)
     : mTarget(aTarget), mHandler(nullptr) {}
 
 NS_IMPL_ISUPPORTS(GlobalKeyListener, nsIDOMEventListener)
@@ -51,16 +57,15 @@ static void BuildHandlerChain(nsIContent* aContent, KeyEventHandler** aResult) {
       continue;
     }
 
-    Element* keyElement = key->AsElement();
+    dom::Element* keyElement = key->AsElement();
     // Check whether the key element has empty value at key/char attribute.
     // Such element is used by localizers for alternative shortcut key
     // definition on the locale. See bug 426501.
     nsAutoString valKey, valCharCode, valKeyCode;
     // Hopefully at least one of the attributes is set:
-    keyElement->GetAttr(kNameSpaceID_None, nsGkAtoms::key, valKey) ||
-        keyElement->GetAttr(kNameSpaceID_None, nsGkAtoms::charcode,
-                            valCharCode) ||
-        keyElement->GetAttr(kNameSpaceID_None, nsGkAtoms::keycode, valKeyCode);
+    keyElement->GetAttr(nsGkAtoms::key, valKey) ||
+        keyElement->GetAttr(nsGkAtoms::charcode, valCharCode) ||
+        keyElement->GetAttr(nsGkAtoms::keycode, valKeyCode);
     // If not, ignore this key element.
     if (valKey.IsEmpty() && valCharCode.IsEmpty() && valKeyCode.IsEmpty()) {
       continue;
@@ -83,7 +88,7 @@ static void BuildHandlerChain(nsIContent* aContent, KeyEventHandler** aResult) {
   }
 }
 
-void GlobalKeyListener::WalkHandlers(KeyboardEvent* aKeyEvent) {
+void GlobalKeyListener::WalkHandlers(dom::KeyboardEvent* aKeyEvent) {
   if (aKeyEvent->DefaultPrevented()) {
     return;
   }
@@ -100,7 +105,7 @@ void GlobalKeyListener::WalkHandlers(KeyboardEvent* aKeyEvent) {
     return;
   }
 
-  WalkHandlersInternal(aKeyEvent, true);
+  WalkHandlersInternal(Purpose::ExecuteCommand, aKeyEvent);
 }
 
 void GlobalKeyListener::InstallKeyboardEventListenersTo(
@@ -108,102 +113,68 @@ void GlobalKeyListener::InstallKeyboardEventListenersTo(
   // For marking each keyboard event as if it's reserved by chrome,
   // GlobalKeyListeners need to listen each keyboard events before
   // web contents.
-  aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("keydown"), TrustedEventsAtCapture());
-  aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("keyup"), TrustedEventsAtCapture());
-  aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("keypress"), TrustedEventsAtCapture());
-  aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeydownonplugin"), TrustedEventsAtCapture());
-  aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeyuponplugin"), TrustedEventsAtCapture());
+  aEventListenerManager->AddEventListenerByType(this, u"keydown"_ns,
+                                                TrustedEventsAtCapture());
+  aEventListenerManager->AddEventListenerByType(this, u"keyup"_ns,
+                                                TrustedEventsAtCapture());
+  aEventListenerManager->AddEventListenerByType(this, u"keypress"_ns,
+                                                TrustedEventsAtCapture());
 
   // For reducing the IPC cost, preventing to dispatch reserved keyboard
   // events into the content process.
   aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("keydown"), TrustedEventsAtSystemGroupCapture());
+      this, u"keydown"_ns, TrustedEventsAtSystemGroupCapture());
   aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("keyup"), TrustedEventsAtSystemGroupCapture());
+      this, u"keyup"_ns, TrustedEventsAtSystemGroupCapture());
   aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("keypress"), TrustedEventsAtSystemGroupCapture());
-  aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeydownonplugin"),
-      TrustedEventsAtSystemGroupCapture());
-  aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeyuponplugin"),
-      TrustedEventsAtSystemGroupCapture());
+      this, u"keypress"_ns, TrustedEventsAtSystemGroupCapture());
 
   // Handle keyboard events in bubbling phase of the system event group.
   aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("keydown"), TrustedEventsAtSystemGroupBubble());
+      this, u"keydown"_ns, TrustedEventsAtSystemGroupBubble());
   aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("keyup"), TrustedEventsAtSystemGroupBubble());
+      this, u"keyup"_ns, TrustedEventsAtSystemGroupBubble());
   aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("keypress"), TrustedEventsAtSystemGroupBubble());
+      this, u"keypress"_ns, TrustedEventsAtSystemGroupBubble());
   // mozaccesskeynotfound event is fired when modifiers of keypress event
   // matches with modifier of content access key but it's not consumed by
   // remote content.
   aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("mozaccesskeynotfound"),
-      TrustedEventsAtSystemGroupBubble());
-  aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeydownonplugin"),
-      TrustedEventsAtSystemGroupBubble());
-  aEventListenerManager->AddEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeyuponplugin"),
-      TrustedEventsAtSystemGroupBubble());
+      this, u"mozaccesskeynotfound"_ns, TrustedEventsAtSystemGroupBubble());
 }
 
 void GlobalKeyListener::RemoveKeyboardEventListenersFrom(
     EventListenerManager* aEventListenerManager) {
-  aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("keydown"), TrustedEventsAtCapture());
-  aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("keyup"), TrustedEventsAtCapture());
-  aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("keypress"), TrustedEventsAtCapture());
-  aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeydownonplugin"), TrustedEventsAtCapture());
-  aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeyuponplugin"), TrustedEventsAtCapture());
+  aEventListenerManager->RemoveEventListenerByType(this, u"keydown"_ns,
+                                                   TrustedEventsAtCapture());
+  aEventListenerManager->RemoveEventListenerByType(this, u"keyup"_ns,
+                                                   TrustedEventsAtCapture());
+  aEventListenerManager->RemoveEventListenerByType(this, u"keypress"_ns,
+                                                   TrustedEventsAtCapture());
 
   aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("keydown"), TrustedEventsAtSystemGroupCapture());
+      this, u"keydown"_ns, TrustedEventsAtSystemGroupCapture());
   aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("keyup"), TrustedEventsAtSystemGroupCapture());
+      this, u"keyup"_ns, TrustedEventsAtSystemGroupCapture());
   aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("keypress"), TrustedEventsAtSystemGroupCapture());
-  aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeydownonplugin"),
-      TrustedEventsAtSystemGroupCapture());
-  aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeyuponplugin"),
-      TrustedEventsAtSystemGroupCapture());
+      this, u"keypress"_ns, TrustedEventsAtSystemGroupCapture());
 
   aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("keydown"), TrustedEventsAtSystemGroupBubble());
+      this, u"keydown"_ns, TrustedEventsAtSystemGroupBubble());
   aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("keyup"), TrustedEventsAtSystemGroupBubble());
+      this, u"keyup"_ns, TrustedEventsAtSystemGroupBubble());
   aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("keypress"), TrustedEventsAtSystemGroupBubble());
+      this, u"keypress"_ns, TrustedEventsAtSystemGroupBubble());
   aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("mozaccesskeynotfound"),
-      TrustedEventsAtSystemGroupBubble());
-  aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeydownonplugin"),
-      TrustedEventsAtSystemGroupBubble());
-  aEventListenerManager->RemoveEventListenerByType(
-      this, NS_LITERAL_STRING("mozkeyuponplugin"),
-      TrustedEventsAtSystemGroupBubble());
+      this, u"mozaccesskeynotfound"_ns, TrustedEventsAtSystemGroupBubble());
 }
 
 NS_IMETHODIMP
-GlobalKeyListener::HandleEvent(Event* aEvent) {
-  RefPtr<KeyboardEvent> keyEvent = aEvent->AsKeyboardEvent();
+GlobalKeyListener::HandleEvent(dom::Event* aEvent) {
+  RefPtr<dom::KeyboardEvent> keyEvent = aEvent->AsKeyboardEvent();
   NS_ENSURE_TRUE(keyEvent, NS_ERROR_INVALID_ARG);
 
-  if (aEvent->EventPhase() == Event_Binding::CAPTURING_PHASE) {
+  if (aEvent->EventPhase() == dom::Event_Binding::CAPTURING_PHASE) {
     if (aEvent->WidgetEventPtr()->mFlags.mInSystemGroup) {
       HandleEventOnCaptureInSystemEventGroup(keyEvent);
     } else {
@@ -212,34 +183,9 @@ GlobalKeyListener::HandleEvent(Event* aEvent) {
     return NS_OK;
   }
 
-  WidgetKeyboardEvent* widgetKeyboardEvent =
-      aEvent->WidgetEventPtr()->AsKeyboardEvent();
-  if (widgetKeyboardEvent->IsKeyEventOnPlugin()) {
-    // key events on plugin shouldn't execute shortcut key handlers which are
-    // not reserved.
-    if (!widgetKeyboardEvent->IsReservedByChrome()) {
-      return NS_OK;
-    }
-
-    // If the event is untrusted event or was already consumed, do nothing.
-    if (!widgetKeyboardEvent->IsTrusted() ||
-        widgetKeyboardEvent->DefaultPrevented()) {
-      return NS_OK;
-    }
-
-    // XXX Don't check isReserved here because even if the handler in this
-    //     instance isn't reserved but another instance reserves the key
-    //     combination, it will be executed when the event is normal keyboard
-    //     events...
-    bool isReserved = false;
-    if (!HasHandlerForEvent(keyEvent, &isReserved)) {
-      return NS_OK;
-    }
-  }
-
   // If this event was handled by APZ then don't do the default action, and
   // preventDefault to prevent any other listeners from handling the event.
-  if (widgetKeyboardEvent->mFlags.mHandledByAPZ) {
+  if (aEvent->WidgetEventPtr()->mFlags.mHandledByAPZ) {
     aEvent->PreventDefault();
     return NS_OK;
   }
@@ -249,7 +195,7 @@ GlobalKeyListener::HandleEvent(Event* aEvent) {
 }
 
 void GlobalKeyListener::HandleEventOnCaptureInDefaultEventGroup(
-    KeyboardEvent* aEvent) {
+    dom::KeyboardEvent* aEvent) {
   WidgetKeyboardEvent* widgetKeyboardEvent =
       aEvent->WidgetEventPtr()->AsKeyboardEvent();
 
@@ -257,14 +203,13 @@ void GlobalKeyListener::HandleEventOnCaptureInDefaultEventGroup(
     return;
   }
 
-  bool isReserved = false;
-  if (HasHandlerForEvent(aEvent, &isReserved) && isReserved) {
+  if (HasHandlerForEvent(aEvent).mReservedHandlerForChromeFound) {
     widgetKeyboardEvent->MarkAsReservedByChrome();
   }
 }
 
 void GlobalKeyListener::HandleEventOnCaptureInSystemEventGroup(
-    KeyboardEvent* aEvent) {
+    dom::KeyboardEvent* aEvent) {
   WidgetKeyboardEvent* widgetEvent =
       aEvent->WidgetEventPtr()->AsKeyboardEvent();
 
@@ -278,7 +223,7 @@ void GlobalKeyListener::HandleEventOnCaptureInSystemEventGroup(
     return;
   }
 
-  if (!HasHandlerForEvent(aEvent)) {
+  if (!HasHandlerForEvent(aEvent).mMeaningfulHandlerFound) {
     return;
   }
 
@@ -297,13 +242,12 @@ void GlobalKeyListener::HandleEventOnCaptureInSystemEventGroup(
 // WalkHandlersInternal and WalkHandlersAndExecute
 //
 // Given a particular DOM event and a pointer to the first handler in the list,
-// scan through the list to find something to handle the event. If aExecute =
-// true, the handler will be executed; otherwise just return an answer telling
-// if a handler for that event was found.
+// scan through the list to find something to handle the event. If aPurpose =
+// Purpose::ExecuteHandler, the handler will be executed; otherwise just return
+// an answer telling if a handler for that event was found.
 //
-bool GlobalKeyListener::WalkHandlersInternal(KeyboardEvent* aKeyEvent,
-                                             bool aExecute,
-                                             bool* aOutReservedForChrome) {
+GlobalKeyListener::WalkHandlersResult GlobalKeyListener::WalkHandlersInternal(
+    Purpose aPurpose, dom::KeyboardEvent* aKeyEvent) {
   WidgetKeyboardEvent* nativeKeyboardEvent =
       aKeyEvent->WidgetEventPtr()->AsKeyboardEvent();
   MOZ_ASSERT(nativeKeyboardEvent);
@@ -312,60 +256,63 @@ bool GlobalKeyListener::WalkHandlersInternal(KeyboardEvent* aKeyEvent,
   nativeKeyboardEvent->GetShortcutKeyCandidates(shortcutKeys);
 
   if (shortcutKeys.IsEmpty()) {
-    return WalkHandlersAndExecute(aKeyEvent, 0, IgnoreModifierState(), aExecute,
-                                  aOutReservedForChrome);
+    return WalkHandlersAndExecute(aPurpose, aKeyEvent, 0,
+                                  IgnoreModifierState());
   }
 
-  for (unsigned long i = 0; i < shortcutKeys.Length(); ++i) {
-    ShortcutKeyCandidate& key = shortcutKeys[i];
+  bool foundDisabledHandler = false;
+  for (const ShortcutKeyCandidate& key : shortcutKeys) {
+    const bool skipIfEarlierHandlerDisabled =
+        key.mSkipIfEarlierHandlerDisabled ==
+        ShortcutKeyCandidate::SkipIfEarlierHandlerDisabled::Yes;
+    if (foundDisabledHandler && skipIfEarlierHandlerDisabled) {
+      continue;
+    }
     IgnoreModifierState ignoreModifierState;
-    ignoreModifierState.mShift = key.mIgnoreShift;
-    if (WalkHandlersAndExecute(aKeyEvent, key.mCharCode, ignoreModifierState,
-                               aExecute, aOutReservedForChrome)) {
-      return true;
+    ignoreModifierState.mShift =
+        key.mShiftState == ShortcutKeyCandidate::ShiftState::Ignorable;
+    WalkHandlersResult result = WalkHandlersAndExecute(
+        aPurpose, aKeyEvent, key.mCharCode, ignoreModifierState);
+    if (result.mMeaningfulHandlerFound) {
+      return result;
+    }
+    // Note that if the candidate should not match if an earlier handler is
+    // disabled, the char code of the candidate is a char which may be
+    // introduced with different shift state. In this case, we do NOT find a
+    // disabled handler which **exactly** matches with the keyboard event.
+    // This avoids to override a higher priority handler with a disabled lower
+    // priority handler.
+    if (!skipIfEarlierHandlerDisabled && !foundDisabledHandler) {
+      foundDisabledHandler = result.mDisabledHandlerFound;
     }
   }
-  return false;
+  return {};
 }
 
-bool GlobalKeyListener::WalkHandlersAndExecute(
-    KeyboardEvent* aKeyEvent, uint32_t aCharCode,
-    const IgnoreModifierState& aIgnoreModifierState, bool aExecute,
-    bool* aOutReservedForChrome) {
-  if (aOutReservedForChrome) {
-    *aOutReservedForChrome = false;
-  }
-
+GlobalKeyListener::WalkHandlersResult GlobalKeyListener::WalkHandlersAndExecute(
+    Purpose aPurpose, dom::KeyboardEvent* aKeyEvent, uint32_t aCharCode,
+    const IgnoreModifierState& aIgnoreModifierState) {
   WidgetKeyboardEvent* widgetKeyboardEvent =
       aKeyEvent->WidgetEventPtr()->AsKeyboardEvent();
   if (NS_WARN_IF(!widgetKeyboardEvent)) {
-    return false;
+    return {};
   }
 
   nsAtom* eventType =
       ShortcutKeys::ConvertEventToDOMEventType(widgetKeyboardEvent);
 
   // Try all of the handlers until we find one that matches the event.
+  bool foundDisabledHandler = false;
   for (KeyEventHandler* handler = mHandler; handler;
        handler = handler->GetNextHandler()) {
     bool stopped = aKeyEvent->IsDispatchStopped();
     if (stopped) {
       // The event is finished, don't execute any more handlers
-      return false;
+      return {};
     }
 
-    if (aExecute) {
-      // If the event is eKeyDownOnPlugin, it should execute either keydown
-      // handler or keypress handler because eKeyDownOnPlugin events are
-      // never followed by keypress events.
-      if (widgetKeyboardEvent->mMessage == eKeyDownOnPlugin) {
-        if (!handler->EventTypeEquals(nsGkAtoms::keydown) &&
-            !handler->EventTypeEquals(nsGkAtoms::keypress)) {
-          continue;
-        }
-        // The other event types should exactly be matched with the handler's
-        // event type.
-      } else if (!handler->EventTypeEquals(eventType)) {
+    if (aPurpose == Purpose::ExecuteCommand) {
+      if (!handler->EventTypeEquals(eventType)) {
         continue;
       }
     } else {
@@ -394,17 +341,18 @@ bool GlobalKeyListener::WalkHandlersAndExecute(
     // Before executing this handler, check that it's not disabled,
     // and that it has something to do (oncommand of the <key> or its
     // <command> is non-empty).
-    if (!CanHandle(handler, aExecute)) {
+    if (!CanHandle(handler, aPurpose == Purpose::ExecuteCommand)) {
+      foundDisabledHandler = true;
       continue;
     }
 
-    if (!aExecute) {
+    if (aPurpose == Purpose::LookForCommand) {
       if (handler->EventTypeEquals(eventType)) {
-        if (aOutReservedForChrome) {
-          *aOutReservedForChrome = IsReservedKey(widgetKeyboardEvent, handler);
-        }
-
-        return true;
+        WalkHandlersResult result;
+        result.mMeaningfulHandlerFound = true;
+        result.mReservedHandlerForChromeFound =
+            IsReservedKey(widgetKeyboardEvent, handler);
+        return result;
       }
 
       // If the command is reserved and the event is keydown, check also if
@@ -413,34 +361,28 @@ bool GlobalKeyListener::WalkHandlersAndExecute(
       if (eventType == nsGkAtoms::keydown &&
           handler->EventTypeEquals(nsGkAtoms::keypress)) {
         if (IsReservedKey(widgetKeyboardEvent, handler)) {
-          if (aOutReservedForChrome) {
-            *aOutReservedForChrome = true;
-          }
-
-          return true;
+          WalkHandlersResult result;
+          result.mMeaningfulHandlerFound = true;
+          result.mReservedHandlerForChromeFound = true;
+          return result;
         }
       }
       // Otherwise, we've not found a handler for the event yet.
       continue;
     }
 
-    // This should only be assigned when aExecute is false.
-    MOZ_ASSERT(!aOutReservedForChrome);
-
-    // If it's not reserved and the event is a key event on a plugin,
-    // the handler shouldn't be executed.
-    if (widgetKeyboardEvent->IsKeyEventOnPlugin() &&
-        !IsReservedKey(widgetKeyboardEvent, handler)) {
-      return false;
-    }
-
-    nsCOMPtr<EventTarget> target = GetHandlerTarget(handler);
+    nsCOMPtr<dom::EventTarget> target = GetHandlerTarget(handler);
 
     // XXX Do we execute only one handler even if the handler neither stops
     //     propagation nor prevents default of the event?
     nsresult rv = handler->ExecuteHandler(target, aKeyEvent);
     if (NS_SUCCEEDED(rv)) {
-      return true;
+      WalkHandlersResult result;
+      result.mMeaningfulHandlerFound = true;
+      result.mReservedHandlerForChromeFound =
+          IsReservedKey(widgetKeyboardEvent, handler);
+      result.mDisabledHandlerFound = (rv == NS_SUCCESS_DOM_NO_OPERATION);
+      return result;
     }
   }
 
@@ -449,15 +391,17 @@ bool GlobalKeyListener::WalkHandlersAndExecute(
   // shortcut keys even if the key is pressed.  Therefore, if there is no
   // shortcut key which exactly matches current modifier state, we should
   // retry to look for a shortcut key without the Windows-Logo key press.
-  if (!aIgnoreModifierState.mOS && widgetKeyboardEvent->IsOS()) {
+  if (!aIgnoreModifierState.mMeta && widgetKeyboardEvent->IsMeta()) {
     IgnoreModifierState ignoreModifierState(aIgnoreModifierState);
-    ignoreModifierState.mOS = true;
-    return WalkHandlersAndExecute(aKeyEvent, aCharCode, ignoreModifierState,
-                                  aExecute);
+    ignoreModifierState.mMeta = true;
+    return WalkHandlersAndExecute(aPurpose, aKeyEvent, aCharCode,
+                                  ignoreModifierState);
   }
 #endif
 
-  return false;
+  WalkHandlersResult result;
+  result.mDisabledHandlerFound = foundDisabledHandler;
+  return result;
 }
 
 bool GlobalKeyListener::IsReservedKey(WidgetKeyboardEvent* aKeyEvent,
@@ -470,28 +414,43 @@ bool GlobalKeyListener::IsReservedKey(WidgetKeyboardEvent* aKeyEvent,
     return false;
   }
 
-  if (reserved == ReservedKey_True) {
-    return true;
+  if (reserved != ReservedKey_True &&
+      !nsContentUtils::ShouldBlockReservedKeys(aKeyEvent)) {
+    return false;
   }
 
-  return nsContentUtils::ShouldBlockReservedKeys(aKeyEvent);
+  // Okay, the key handler is reserved, but if the key combination is mapped to
+  // an edit command or a selection navigation command, we should not treat it
+  // as reserved since user wants to do the mapped thing(s) in editor.
+  if (MOZ_UNLIKELY(!aKeyEvent->IsTrusted() || !aKeyEvent->mWidget)) {
+    return true;
+  }
+  widget::InputContext inputContext = aKeyEvent->mWidget->GetInputContext();
+  if (!inputContext.mIMEState.IsEditable()) {
+    return true;
+  }
+  return MOZ_UNLIKELY(!aKeyEvent->IsEditCommandsInitialized(
+             inputContext.GetNativeKeyBindingsType())) ||
+         aKeyEvent
+             ->EditCommandsConstRef(inputContext.GetNativeKeyBindingsType())
+             .IsEmpty();
 }
 
-bool GlobalKeyListener::HasHandlerForEvent(KeyboardEvent* aEvent,
-                                           bool* aOutReservedForChrome) {
+GlobalKeyListener::WalkHandlersResult GlobalKeyListener::HasHandlerForEvent(
+    dom::KeyboardEvent* aEvent) {
   WidgetKeyboardEvent* widgetKeyboardEvent =
       aEvent->WidgetEventPtr()->AsKeyboardEvent();
   if (NS_WARN_IF(!widgetKeyboardEvent) || !widgetKeyboardEvent->IsTrusted()) {
-    return false;
+    return {};
   }
 
   EnsureHandlers();
 
   if (IsDisabled()) {
-    return false;
+    return {};
   }
 
-  return WalkHandlersInternal(aEvent, false, aOutReservedForChrome);
+  return WalkHandlersInternal(Purpose::LookForCommand, aEvent);
 }
 
 //
@@ -502,9 +461,10 @@ bool GlobalKeyListener::HasHandlerForEvent(KeyboardEvent* aEvent,
 // content, then extra work needs to be done to hook it up to the document (XXX
 // WHY??)
 //
-void XULKeySetGlobalKeyListener::AttachKeyHandler(Element* aElementTarget) {
+void XULKeySetGlobalKeyListener::AttachKeyHandler(
+    dom::Element* aElementTarget) {
   // Only attach if we're really in a document
-  nsCOMPtr<Document> doc = aElementTarget->GetUncomposedDoc();
+  nsCOMPtr<dom::Document> doc = aElementTarget->GetUncomposedDoc();
   if (!doc) {
     return;
   }
@@ -534,9 +494,10 @@ void XULKeySetGlobalKeyListener::AttachKeyHandler(Element* aElementTarget) {
 //
 // Removes a key handler added by AttachKeyHandler.
 //
-void XULKeySetGlobalKeyListener::DetachKeyHandler(Element* aElementTarget) {
+void XULKeySetGlobalKeyListener::DetachKeyHandler(
+    dom::Element* aElementTarget) {
   // Only attach if we're really in a document
-  nsCOMPtr<Document> doc = aElementTarget->GetUncomposedDoc();
+  nsCOMPtr<dom::Document> doc = aElementTarget->GetUncomposedDoc();
   if (!doc) {
     return;
   }
@@ -557,14 +518,14 @@ void XULKeySetGlobalKeyListener::DetachKeyHandler(Element* aElementTarget) {
   aElementTarget->RemoveProperty(nsGkAtoms::listener);
 }
 
-XULKeySetGlobalKeyListener::XULKeySetGlobalKeyListener(Element* aElement,
-                                                       EventTarget* aTarget)
+XULKeySetGlobalKeyListener::XULKeySetGlobalKeyListener(
+    dom::Element* aElement, dom::EventTarget* aTarget)
     : GlobalKeyListener(aTarget) {
   mWeakPtrForElement = do_GetWeakReference(aElement);
 }
 
-Element* XULKeySetGlobalKeyListener::GetElement(bool* aIsDisabled) const {
-  RefPtr<Element> element = do_QueryReferent(mWeakPtrForElement);
+dom::Element* XULKeySetGlobalKeyListener::GetElement(bool* aIsDisabled) const {
+  RefPtr<dom::Element> element = do_QueryReferent(mWeakPtrForElement);
   if (element && aIsDisabled) {
     *aIsDisabled = element->AttrValueIs(kNameSpaceID_None, nsGkAtoms::disabled,
                                         nsGkAtoms::_true, eCaseMatters);
@@ -583,7 +544,7 @@ void XULKeySetGlobalKeyListener::EnsureHandlers() {
     return;
   }
 
-  Element* element = GetElement();
+  dom::Element* element = GetElement();
   if (!element) {
     return;
   }
@@ -593,23 +554,23 @@ void XULKeySetGlobalKeyListener::EnsureHandlers() {
 
 bool XULKeySetGlobalKeyListener::IsDisabled() const {
   bool isDisabled;
-  Element* element = GetElement(&isDisabled);
+  dom::Element* element = GetElement(&isDisabled);
   return element && isDisabled;
 }
 
 bool XULKeySetGlobalKeyListener::GetElementForHandler(
-    KeyEventHandler* aHandler, Element** aElementForHandler) const {
+    KeyEventHandler* aHandler, dom::Element** aElementForHandler) const {
   MOZ_ASSERT(aElementForHandler);
   *aElementForHandler = nullptr;
 
-  RefPtr<Element> keyElement = aHandler->GetHandlerElement();
+  RefPtr<dom::Element> keyElement = aHandler->GetHandlerElement();
   if (!keyElement) {
     // This should only be the case where the <key> element that generated the
     // handler has been destroyed. Not sure why we return true here...
     return true;
   }
 
-  nsCOMPtr<Element> chromeHandlerElement = GetElement();
+  nsCOMPtr<dom::Element> chromeHandlerElement = GetElement();
   if (!chromeHandlerElement) {
     NS_WARNING_ASSERTION(keyElement->IsInUncomposedDoc(), "uncomposed");
     keyElement.swap(*aElementForHandler);
@@ -618,7 +579,7 @@ bool XULKeySetGlobalKeyListener::GetElementForHandler(
 
   // We are in a XUL doc.  Obtain our command attribute.
   nsAutoString command;
-  keyElement->GetAttr(kNameSpaceID_None, nsGkAtoms::command, command);
+  keyElement->GetAttr(nsGkAtoms::command, command);
   if (command.IsEmpty()) {
     // There is no command element associated with the key element.
     NS_WARNING_ASSERTION(keyElement->IsInUncomposedDoc(), "uncomposed");
@@ -627,12 +588,12 @@ bool XULKeySetGlobalKeyListener::GetElementForHandler(
   }
 
   // XXX Shouldn't we check this earlier?
-  Document* doc = keyElement->GetUncomposedDoc();
+  dom::Document* doc = keyElement->GetUncomposedDoc();
   if (NS_WARN_IF(!doc)) {
     return false;
   }
 
-  nsCOMPtr<Element> commandElement = doc->GetElementById(command);
+  nsCOMPtr<dom::Element> commandElement = doc->GetElementById(command);
   if (!commandElement) {
     NS_ERROR(
         "A XUL <key> is observing a command that doesn't exist. "
@@ -644,7 +605,8 @@ bool XULKeySetGlobalKeyListener::GetElementForHandler(
   return true;
 }
 
-bool XULKeySetGlobalKeyListener::IsExecutableElement(Element* aElement) const {
+bool XULKeySetGlobalKeyListener::IsExecutableElement(
+    dom::Element* aElement) const {
   if (!aElement) {
     return false;
   }
@@ -655,13 +617,18 @@ bool XULKeySetGlobalKeyListener::IsExecutableElement(Element* aElement) const {
     return false;
   }
 
-  aElement->GetAttr(nsGkAtoms::oncommand, value);
-  return !value.IsEmpty();
+  // Internal keys are defined as <key> elements so that the menu label
+  // and disabled state can be updated properly, but the command is executed
+  // by some other means. This will typically be because the key is defined
+  // as a shortcut defined in ShortcutKeyDefinitions.cpp instead, or on Mac,
+  // some special system defined keys.
+  return !aElement->AttrValueIs(kNameSpaceID_None, nsGkAtoms::internal,
+                                nsGkAtoms::_true, eCaseMatters);
 }
 
-already_AddRefed<EventTarget> XULKeySetGlobalKeyListener::GetHandlerTarget(
+already_AddRefed<dom::EventTarget> XULKeySetGlobalKeyListener::GetHandlerTarget(
     KeyEventHandler* aHandler) {
-  nsCOMPtr<Element> commandElement;
+  nsCOMPtr<dom::Element> commandElement;
   if (!GetElementForHandler(aHandler, getter_AddRefs(commandElement))) {
     return nullptr;
   }
@@ -671,7 +638,12 @@ already_AddRefed<EventTarget> XULKeySetGlobalKeyListener::GetHandlerTarget(
 
 bool XULKeySetGlobalKeyListener::CanHandle(KeyEventHandler* aHandler,
                                            bool aWillExecute) const {
-  nsCOMPtr<Element> commandElement;
+  // If the <key> element itself is disabled, ignore it.
+  if (aHandler->KeyElementIsDisabled()) {
+    return false;
+  }
+
+  nsCOMPtr<dom::Element> commandElement;
   if (!GetElementForHandler(aHandler, getter_AddRefs(commandElement))) {
     return false;
   }
@@ -718,7 +690,7 @@ layers::KeyboardMap RootWindowGlobalKeyListener::CollectKeyboardShortcuts() {
 // content, then extra work needs to be done to hook it up to the document (XXX
 // WHY??)
 //
-void RootWindowGlobalKeyListener::AttachKeyHandler(EventTarget* aTarget) {
+void RootWindowGlobalKeyListener::AttachKeyHandler(dom::EventTarget* aTarget) {
   EventListenerManager* manager = aTarget->GetOrCreateListenerManager();
   if (!manager) {
     return;
@@ -733,7 +705,8 @@ void RootWindowGlobalKeyListener::AttachKeyHandler(EventTarget* aTarget) {
   handler->InstallKeyboardEventListenersTo(manager);
 }
 
-RootWindowGlobalKeyListener::RootWindowGlobalKeyListener(EventTarget* aTarget)
+RootWindowGlobalKeyListener::RootWindowGlobalKeyListener(
+    dom::EventTarget* aTarget)
     : GlobalKeyListener(aTarget) {}
 
 /* static */
@@ -760,8 +733,7 @@ bool RootWindowGlobalKeyListener::IsHTMLEditorFocused() {
     return false;
   }
 
-  Document* doc = htmlEditor->GetDocument();
-  if (doc->HasFlag(NODE_IS_EDITABLE)) {
+  if (htmlEditor->IsInDesignMode()) {
     // Don't need to perform any checks in designMode documents.
     return true;
   }
@@ -769,15 +741,15 @@ bool RootWindowGlobalKeyListener::IsHTMLEditorFocused() {
   nsINode* focusedNode = fm->GetFocusedElement();
   if (focusedNode && focusedNode->IsElement()) {
     // If there is a focused element, make sure it's in the active editing host.
-    // Note that GetActiveEditingHost finds the current editing host based on
+    // Note that ComputeEditingHost finds the current editing host based on
     // the document's selection.  Even though the document selection is usually
     // collapsed to where the focus is, but the page may modify the selection
     // without our knowledge, in which case this check will do something useful.
-    nsCOMPtr<Element> activeEditingHost = htmlEditor->GetActiveEditingHost();
-    if (!activeEditingHost) {
+    dom::Element* editingHost = htmlEditor->ComputeEditingHost();
+    if (!editingHost) {
       return false;
     }
-    return focusedNode->IsInclusiveDescendantOf(activeEditingHost);
+    return focusedNode->IsInclusiveDescendantOf(editingHost);
   }
 
   return false;
