@@ -18,11 +18,14 @@
 #include "nsITimer.h"
 #include "nsTArray.h"
 #include "nsThreadUtils.h"
+#include "nsWindowsHelpers.h"
 
 #include "mozilla/ArrayUtils.h"
 
 #include "mozilla/ipc/BackgroundParent.h"
 #include "mozilla/dom/GamepadPlatformService.h"
+#include "mozilla/dom/GamepadRemapping.h"
+#include "Gamepad.h"
 
 namespace {
 
@@ -33,12 +36,10 @@ using mozilla::ArrayLength;
 // USB HID usage tables, page 1, 0x30 = X
 const uint32_t kAxisMinimumUsageNumber = 0x30;
 // USB HID usage tables, page 1 (Hat switch)
-const uint32_t kDpadMinimumUsageNumber = 0x39;
 const uint32_t kAxesLengthCap = 16;
 
 // USB HID usage tables
 const uint32_t kDesktopUsagePage = 0x1;
-const uint32_t kGameControlsUsagePage = 0x5;
 const uint32_t kButtonUsagePage = 0x9;
 
 // Multiple devices-changed notifications can be sent when a device
@@ -48,8 +49,8 @@ const uint32_t kButtonUsagePage = 0x9;
 const uint32_t kDevicesChangedStableDelay = 200;
 // Both DirectInput and XInput are polling-driven here,
 // so we need to poll it periodically.
-// 50ms is arbitrarily chosen.
-const uint32_t kWindowsGamepadPollInterval = 50;
+// 4ms, or 250 Hz, is consistent with Chrome's gamepad implementation.
+const uint32_t kWindowsGamepadPollInterval = 4;
 
 const UINT kRawInputError = (UINT)-1;
 
@@ -151,7 +152,7 @@ class Gamepad {
 };
 
 // Drop this in favor of decltype when we require a new enough SDK.
-typedef void(WINAPI* XInputEnable_func)(BOOL);
+using XInputEnable_func = void(WINAPI*)(BOOL);
 
 // RAII class to wrap loading the XInput DLL
 class XInputLoader {
@@ -1091,8 +1092,7 @@ class StopWindowsGamepadServiceRunnable final : public Runnable {
 
 }  // namespace
 
-namespace mozilla {
-namespace dom {
+namespace mozilla::dom {
 
 using namespace mozilla::ipc;
 
@@ -1133,5 +1133,4 @@ void SetGamepadLightIndicatorColor(const Tainted<GamepadHandle>& aGamepadHandle,
                                    aGreen, aBlue);
 }
 
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom
