@@ -24,15 +24,16 @@ using mozilla::media::TimeUnit;
 namespace mozilla {
 LazyLogModule gMP4MetadataLog("MP4Metadata");
 
-IndiceWrapper::IndiceWrapper(Mp4parseByteData& aIndice) {
+IndiceWrapper::IndiceWrapper(Mp4parseByteData& aRustIndice) {
   mIndice.data = nullptr;
-  mIndice.length = aIndice.length;
-  mIndice.indices = aIndice.indices;
+  mIndice.length = aRustIndice.length;
+  mIndice.indices = aRustIndice.indices;
 }
 
 size_t IndiceWrapper::Length() const { return mIndice.length; }
 
-bool IndiceWrapper::GetIndice(size_t aIndex, Index::Indice& aIndice) const {
+bool IndiceWrapper::GetIndice(size_t aIndex,
+                              MP4SampleIndex::Indice& aIndice) const {
   if (aIndex >= mIndice.length) {
     MOZ_LOG(gMP4MetadataLog, LogLevel::Error, ("Index overflow in indice"));
     return false;
@@ -306,6 +307,9 @@ MP4Metadata::ResultAndTrackInfo MP4Metadata::GetTrackInfo(
       case MP4PARSE_CODEC_ALAC:
         codecString = "alac";
         break;
+      case MP4PARSE_CODEC_H263:
+        codecString = "h.263";
+        break;
       case MP4PARSE_CODEC_AVC:
         codecString = "h.264";
         break;
@@ -330,13 +334,29 @@ MP4Metadata::ResultAndTrackInfo MP4Metadata::GetTrackInfo(
       case MP4PARSE_CODEC_EC3:
         codecString = "ec-3";
         break;
+      case MP4PARSE_CODEC_HEVC:
+        codecString = "hevc";
+        break;
     }
   }
   MOZ_LOG(gMP4MetadataLog, LogLevel::Debug,
           ("track codec %s (%u)\n", codecString, codecType));
 #endif
 
-  // This specialization interface is crazy.
+  Mp4parseTrackInfo track_info;
+  rv = mp4parse_get_track_info(mParser.get(), trackIndex.value(), &track_info);
+  if (rv != MP4PARSE_STATUS_OK) {
+    MOZ_LOG(gMP4MetadataLog, LogLevel::Warning,
+            ("mp4parse_get_track_info returned error %d", rv));
+    return {MediaResult(NS_ERROR_DOM_MEDIA_METADATA_ERR,
+                        RESULT_DETAIL("Cannot parse %s track #%zu",
+                                      TrackTypeToStr(aType), aTrackNumber)),
+            nullptr};
+  }
+
+  uint32_t timeScale = info.time_scale;
+
+  // This specialization interface is wild.
   UniquePtr<mozilla::TrackInfo> e;
   switch (aType) {
     case TrackInfo::TrackType::kAudioTrack: {
@@ -351,8 +371,17 @@ MP4Metadata::ResultAndTrackInfo MP4Metadata::GetTrackInfo(
                                           TrackTypeToStr(aType), aTrackNumber)),
                 nullptr};
       }
+
+      auto indices = GetTrackIndice(info.track_id);
+      if (!indices.Ref()) {
+        // non fatal
+        MOZ_LOG(gMP4MetadataLog, LogLevel::Warning,
+                ("Can't get index table for audio track, duration might be "
+                 "slightly incorrect"));
+      }
       auto track = mozilla::MakeUnique<MP4AudioInfo>();
-      MediaResult updateStatus = track->Update(&info, &audio);
+      MediaResult updateStatus =
+          track->Update(&info, &audio, indices.Ref().get());
       if (NS_FAILED(updateStatus)) {
         MOZ_LOG(gMP4MetadataLog, LogLevel::Warning,
                 ("Updating audio track failed with %s",
@@ -402,16 +431,23 @@ MP4Metadata::ResultAndTrackInfo MP4Metadata::GetTrackInfo(
               nullptr};
   }
 
+  e->mTimeScale = timeScale;
+
   // No duration in track, use fragment_duration.
   if (e && !e->mDuration.IsPositive()) {
-    Mp4parseFragmentInfo info;
-    auto rv = mp4parse_get_fragment_info(mParser.get(), &info);
+    Mp4parseFragmentInfo fragmentInfo;
+    auto rv = mp4parse_get_fragment_info(mParser.get(), &fragmentInfo);
     if (rv == MP4PARSE_STATUS_OK) {
-      e->mDuration = TimeUnit::FromMicroseconds(info.fragment_duration);
+      // This doesn't use the time scale of the track, but the time scale
+      // indicated in the mvhd box
+      e->mDuration = TimeUnit(fragmentInfo.fragment_duration,
+                              AssertedCast<int64_t>(fragmentInfo.time_scale));
     }
   }
 
   if (e && e->IsValid()) {
+    MOZ_LOG(gMP4MetadataLog, LogLevel::Debug,
+            ("parsed a track info (%s)", e->ToString().get()));
     return {NS_OK, std::move(e)};
   }
   MOZ_LOG(gMP4MetadataLog, LogLevel::Debug, ("TrackInfo didn't validate"));
@@ -428,7 +464,8 @@ MP4Metadata::ResultAndCryptoFile MP4Metadata::Crypto() const {
   return {NS_OK, &mCrypto};
 }
 
-MP4Metadata::ResultAndIndice MP4Metadata::GetTrackIndice(uint32_t aTrackId) {
+MP4Metadata::ResultAndIndice MP4Metadata::GetTrackIndice(
+    uint32_t aTrackId) const {
   Mp4parseByteData indiceRawData = {};
 
   uint8_t fragmented = false;
