@@ -3,16 +3,17 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "H264.h"
-#include <cmath>
-#include <limits>
 #include "AnnexB.h"
 #include "BitReader.h"
 #include "BitWriter.h"
 #include "BufferReader.h"
+#include "ByteStreamsUtils.h"
 #include "ByteWriter.h"
-#include "mozilla/ArrayUtils.h"
+#include "MediaInfo.h"
 #include "mozilla/PodOperations.h"
 #include "mozilla/ResultExtensions.h"
+#include "mozilla/Try.h"
+#include <limits>
 
 #define READSE(var, min, max)     \
   {                               \
@@ -32,6 +33,10 @@
     aDest.var = uval;            \
   }
 
+mozilla::LazyLogModule gH264("H264");
+
+#define LOG(msg, ...) MOZ_LOG(gH264, LogLevel::Debug, (msg, ##__VA_ARGS__))
+
 namespace mozilla {
 
 // Default scaling lists (per spec).
@@ -120,64 +125,6 @@ bool SPSData::operator!=(const SPSData& aOther) const {
   return !(operator==(aOther));
 }
 
-// Described in ISO 23001-8:2016
-// Table 2
-enum class PrimaryID : uint8_t {
-  INVALID = 0,
-  BT709 = 1,
-  UNSPECIFIED = 2,
-  BT470M = 4,
-  BT470BG = 5,
-  SMPTE170M = 6,
-  SMPTE240M = 7,
-  FILM = 8,
-  BT2020 = 9,
-  SMPTEST428_1 = 10,
-  SMPTEST431_2 = 11,
-  SMPTEST432_1 = 12,
-  EBU_3213_E = 22
-};
-
-// Table 3
-enum class TransferID : uint8_t {
-  INVALID = 0,
-  BT709 = 1,
-  UNSPECIFIED = 2,
-  GAMMA22 = 4,
-  GAMMA28 = 5,
-  SMPTE170M = 6,
-  SMPTE240M = 7,
-  LINEAR = 8,
-  LOG = 9,
-  LOG_SQRT = 10,
-  IEC61966_2_4 = 11,
-  BT1361_ECG = 12,
-  IEC61966_2_1 = 13,
-  BT2020_10 = 14,
-  BT2020_12 = 15,
-  SMPTEST2084 = 16,
-  SMPTEST428_1 = 17,
-
-  // Not yet standardized
-  ARIB_STD_B67 = 18,  // AKA hybrid-log gamma, HLG.
-};
-
-// Table 4
-enum class MatrixID : uint8_t {
-  RGB = 0,
-  BT709 = 1,
-  UNSPECIFIED = 2,
-  FCC = 4,
-  BT470BG = 5,
-  SMPTE170M = 6,
-  SMPTE240M = 7,
-  YCOCG = 8,
-  BT2020_NCL = 9,
-  BT2020_CL = 10,
-  YDZDX = 11,
-  INVALID = 255,
-};
-
 static PrimaryID GetPrimaryID(int aPrimary) {
   if (aPrimary < 1 || aPrimary > 22 || aPrimary == 3) {
     return PrimaryID::INVALID;
@@ -351,7 +298,7 @@ class SPSNAL {
       MOZ_ASSERT(mLength / 8 <= mDecodedNAL->Length());
 
       if (memcmp(mDecodedNAL->Elements(), aOther.mDecodedNAL->Elements(),
-                 mLength / 8)) {
+                 mLength / 8) != 0) {
         return false;
       }
 
@@ -697,14 +644,14 @@ bool H264::DecodeSPS(const mozilla::MediaByteBuffer* aSPS, SPSData& aDest) {
   // Determine display size.
   if (aDest.sample_ratio > 1.0) {
     // Increase the intrinsic width
-    aDest.display_width =
-        ConditionDimension(aDest.pic_width * aDest.sample_ratio);
+    aDest.display_width = ConditionDimension(
+        AssertedCast<float>(aDest.pic_width) * aDest.sample_ratio);
     aDest.display_height = aDest.pic_height;
   } else {
     // Increase the intrinsic height
     aDest.display_width = aDest.pic_width;
-    aDest.display_height =
-        ConditionDimension(aDest.pic_height / aDest.sample_ratio);
+    aDest.display_height = ConditionDimension(
+        AssertedCast<float>(aDest.pic_height) / aDest.sample_ratio);
   }
 
   aDest.valid = true;
@@ -939,13 +886,14 @@ uint32_t H264::ComputeMaxRefFrames(const mozilla::MediaByteBuffer* aExtraData) {
 
 /* static */ H264::FrameType H264::GetFrameType(
     const mozilla::MediaRawData* aSample) {
-  if (!AnnexB::IsAVCC(aSample)) {
+  auto avcc = AVCCConfig::Parse(aSample);
+  if (avcc.isErr()) {
     // We must have a valid AVCC frame with extradata.
     return FrameType::INVALID;
   }
   MOZ_ASSERT(aSample->Data());
 
-  int nalLenSize = ((*aSample->mExtraData)[4] & 3) + 1;
+  int nalLenSize = avcc.unwrap().NALUSize();
 
   BufferReader reader(aSample->Data(), aSample->Size());
 
@@ -964,6 +912,8 @@ uint32_t H264::ComputeMaxRefFrames(const mozilla::MediaByteBuffer* aExtraData) {
       case 4:
         nalLen = reader.ReadU32().unwrapOr(0);
         break;
+      default:
+        MOZ_ASSERT_UNREACHABLE("NAL length is up to 4 bytes");
     }
     if (!nalLen) {
       continue;
@@ -972,11 +922,12 @@ uint32_t H264::ComputeMaxRefFrames(const mozilla::MediaByteBuffer* aExtraData) {
     if (!p) {
       return FrameType::INVALID;
     }
-    int8_t nalType = *p & 0x1f;
+    int8_t nalType = AssertedCast<int8_t>(*p & 0x1f);
     if (nalType == H264_NAL_IDR_SLICE) {
       // IDR NAL.
       return FrameType::I_FRAME;
-    } else if (nalType == H264_NAL_SEI) {
+    }
+    if (nalType == H264_NAL_SEI) {
       RefPtr<mozilla::MediaByteBuffer> decodedNAL = DecodeNALUnit(p, nalLen);
       SEIRecoveryData data;
       if (DecodeRecoverySEI(decodedNAL, data)) {
@@ -995,7 +946,8 @@ uint32_t H264::ComputeMaxRefFrames(const mozilla::MediaByteBuffer* aExtraData) {
 
 /* static */ already_AddRefed<mozilla::MediaByteBuffer> H264::ExtractExtraData(
     const mozilla::MediaRawData* aSample) {
-  MOZ_ASSERT(AnnexB::IsAVCC(aSample));
+  auto avcc = AVCCConfig::Parse(aSample);
+  MOZ_ASSERT(avcc.isOk());
 
   RefPtr<mozilla::MediaByteBuffer> extradata = new mozilla::MediaByteBuffer;
 
@@ -1008,7 +960,7 @@ uint32_t H264::ComputeMaxRefFrames(const mozilla::MediaByteBuffer* aExtraData) {
   ByteWriter<BigEndian> ppsw(pps);
   int numPps = 0;
 
-  int nalLenSize = ((*aSample->mExtraData)[4] & 3) + 1;
+  int nalLenSize = avcc.unwrap().NALUSize();
 
   size_t sampleSize = aSample->Size();
   if (aSample->mCrypto.IsEncrypted()) {
@@ -1049,6 +1001,8 @@ uint32_t H264::ComputeMaxRefFrames(const mozilla::MediaByteBuffer* aExtraData) {
         Unused << reader.ReadU32().map(
             [&](uint32_t x) mutable { return nalLen = x; });
         break;
+      default:
+        MOZ_ASSERT_UNREACHABLE("NAL length size is at most 4 bytes");
     }
     const uint8_t* p = reader.Read(nalLen);
     if (!p) {
@@ -1117,25 +1071,14 @@ uint32_t H264::ComputeMaxRefFrames(const mozilla::MediaByteBuffer* aExtraData) {
 }
 
 /* static */
-bool H264::HasSPS(const mozilla::MediaByteBuffer* aExtraData) {
-  return NumSPS(aExtraData) > 0;
+uint8_t H264::NumSPS(const mozilla::MediaByteBuffer* aExtraData) {
+  auto avcc = AVCCConfig::Parse(aExtraData);
+  return avcc.isErr() ? 0 : avcc.unwrap().mNumSPS;
 }
 
 /* static */
-uint8_t H264::NumSPS(const mozilla::MediaByteBuffer* aExtraData) {
-  if (!aExtraData || aExtraData->IsEmpty()) {
-    return 0;
-  }
-
-  BufferReader reader(aExtraData);
-  if (!reader.Read(5)) {
-    return 0;
-  }
-  auto res = reader.ReadU8();
-  if (res.isErr()) {
-    return 0;
-  }
-  return res.unwrap() & 0x1f;
+bool H264::HasSPS(const mozilla::MediaByteBuffer* aExtraData) {
+  return H264::NumSPS(aExtraData) > 0;
 }
 
 /* static */
@@ -1277,6 +1220,8 @@ bool H264::DecodeRecoverySEI(const mozilla::MediaByteBuffer* aSEI,
     bw.WriteUE(1);  // chroma_format_idc -> always set to 4:2:0 chroma format
     bw.WriteUE(0);  // bit_depth_luma_minus8 -> always 8 bits here
     bw.WriteUE(0);  // bit_depth_chroma_minus8 -> always 8 bits here
+    bw.WriteBit(false);  // qpprime_y_zero_transform_bypass_flag
+    bw.WriteBit(false);  // seq_scaling_matrix_present_flag
   }
 
   bw.WriteBits(br.ReadBits(11),
@@ -1348,7 +1293,42 @@ void H264::WriteExtraData(MediaByteBuffer* aDestExtraData,
   aDestExtraData->AppendElements(aPPS.Elements(), aPPS.Length());
 }
 
+/* static */ Result<AVCCConfig, nsresult> AVCCConfig::Parse(
+    const mozilla::MediaRawData* aSample) {
+  if (!aSample || aSample->Size() < 3) {
+    return mozilla::Err(NS_ERROR_FAILURE);
+  }
+  if (aSample->mTrackInfo &&
+      !aSample->mTrackInfo->mMimeType.EqualsLiteral("video/avc")) {
+    LOG("Only allow 'video/avc' (mimeType=%s)",
+        aSample->mTrackInfo->mMimeType.get());
+    return mozilla::Err(NS_ERROR_FAILURE);
+  }
+  return AVCCConfig::Parse(aSample->mExtraData);
+}
+
+/* static */ Result<AVCCConfig, nsresult> AVCCConfig::Parse(
+    const mozilla::MediaByteBuffer* aExtraData) {
+  if (!aExtraData || aExtraData->Length() < 7) {
+    return mozilla::Err(NS_ERROR_FAILURE);
+  }
+  const auto& byteBuffer = *aExtraData;
+  if (byteBuffer[0] != 1) {
+    return mozilla::Err(NS_ERROR_FAILURE);
+  }
+  AVCCConfig avcc{};
+  avcc.mConfigurationVersion = byteBuffer[0];
+  avcc.mAVCProfileIndication = byteBuffer[1];
+  avcc.mProfileCompatibility = byteBuffer[2];
+  avcc.mAVCLevelIndication = byteBuffer[3];
+  avcc.mLengthSizeMinusOne = byteBuffer[4] & 0x3;
+  avcc.mNumSPS = byteBuffer[5] & 0x1F;
+  return avcc;
+}
+
 #undef READUE
 #undef READSE
 
 }  // namespace mozilla
+
+#undef LOG
