@@ -20,27 +20,61 @@
 
 namespace mozilla {
 
-class JsepUuidGenerator {
- public:
-  virtual ~JsepUuidGenerator() {}
-  virtual bool Generate(std::string* id) = 0;
+// JsepSessionImpl members that have default copy c'tors, to simplify the
+// implementation of the copy c'tor for JsepSessionImpl
+class JsepSessionCopyableStuff {
+ protected:
+  struct JsepDtlsFingerprint {
+    std::string mAlgorithm;
+    std::vector<uint8_t> mValue;
+  };
+
+  Maybe<bool> mIsPendingOfferer;
+  Maybe<bool> mIsCurrentOfferer;
+  bool mIceControlling = false;
+  std::string mIceUfrag;
+  std::string mIcePwd;
+  std::string mOldIceUfrag;
+  std::string mOldIcePwd;
+  bool mRemoteIsIceLite = false;
+  std::vector<std::string> mIceOptions;
+  JsepBundlePolicy mBundlePolicy = kBundleBalanced;
+  std::vector<JsepDtlsFingerprint> mDtlsFingerprints;
+  uint64_t mSessionId = 0;
+  uint64_t mSessionVersion = 0;
+  size_t mMidCounter = 0;
+  std::set<std::string> mUsedMids;
+  size_t mTransportIdCounter = 0;
+  std::vector<JsepExtmapMediaType> mRtpExtensions;
+  std::set<uint16_t> mExtmapEntriesEverUsed;
+  std::map<uint16_t, std::string> mExtmapEntriesEverNegotiated;
+  std::string mDefaultRemoteStreamId;
+  std::string mCNAME;
+  // Used to prevent duplicate local SSRCs. Not used to prevent local/remote or
+  // remote-only duplication, which will be important for EKT but not now.
+  std::set<uint32_t> mSsrcs;
+  std::string mLastError;
+  std::vector<std::pair<size_t, std::string>> mLastSdpParsingErrors;
+  bool mEncodeTrackId = true;
+  SsrcGenerator mSsrcGenerator;
+  // !!!NOT INDEXED BY LEVEL!!! The level mapping is done with
+  // JsepTransceiver::mLevel. The keys are UUIDs.
+  std::vector<JsepTransceiver> mTransceivers;
+  // So we can rollback. Not as simple as just going back to the old, though...
+  std::vector<JsepTransceiver> mOldTransceivers;
 };
 
-class JsepSessionImpl : public JsepSession {
+class JsepSessionImpl : public JsepSession, public JsepSessionCopyableStuff {
  public:
   JsepSessionImpl(const std::string& name, UniquePtr<JsepUuidGenerator> uuidgen)
       : JsepSession(name),
-        mIceControlling(false),
-        mRemoteIsIceLite(false),
-        mBundlePolicy(kBundleBalanced),
-        mSessionId(0),
-        mSessionVersion(0),
-        mMidCounter(0),
-        mTransportIdCounter(0),
         mUuidGen(std::move(uuidgen)),
         mSdpHelper(&mLastError),
-        mParser(new HybridSdpParser()),
-        mEncodeTrackId(true) {}
+        mParser(new HybridSdpParser()) {}
+
+  JsepSessionImpl(const JsepSessionImpl& aOrig);
+
+  JsepSession* Clone() const override { return new JsepSessionImpl(*this); }
 
   // Implement JsepSession methods.
   virtual nsresult Init() override;
@@ -54,11 +88,11 @@ class JsepSessionImpl : public JsepSession {
   }
 
   virtual nsresult AddDtlsFingerprint(
-      const std::string& algorithm, const std::vector<uint8_t>& value) override;
+      const nsACString& algorithm, const std::vector<uint8_t>& value) override;
 
-  nsresult AddRtpExtension(JsepMediaType mediaType,
-                           const std::string& extensionName,
-                           SdpDirectionAttribute::Direction direction);
+  virtual nsresult AddRtpExtension(
+      JsepMediaType mediaType, const std::string& extensionName,
+      SdpDirectionAttribute::Direction direction) override;
   virtual nsresult AddAudioRtpExtension(
       const std::string& extensionName,
       SdpDirectionAttribute::Direction direction =
@@ -138,25 +172,23 @@ class JsepSessionImpl : public JsepSession {
   virtual std::set<std::pair<std::string, std::string>> GetLocalIceCredentials()
       const override;
 
-  virtual const std::map<size_t, RefPtr<JsepTransceiver>>& GetTransceivers()
-      const override {
-    return mTransceivers;
-  }
-
-  virtual std::map<size_t, RefPtr<JsepTransceiver>>& GetTransceivers()
-      override {
-    return mTransceivers;
-  }
-
-  virtual nsresult AddTransceiver(RefPtr<JsepTransceiver> transceiver) override;
+  virtual void AddTransceiver(const JsepTransceiver& transceiver) override;
 
   virtual bool CheckNegotiationNeeded() const override;
 
+  virtual void SetDefaultCodecs(
+      const std::vector<UniquePtr<JsepCodecDescription>>& aPreferredCodecs)
+      override;
+
  private:
-  struct JsepDtlsFingerprint {
-    std::string mAlgorithm;
-    std::vector<uint8_t> mValue;
-  };
+  friend class JsepSessionTest;
+  virtual const std::vector<JsepTransceiver>& GetTransceivers() const override {
+    return mTransceivers;
+  }
+
+  virtual std::vector<JsepTransceiver>& GetTransceivers() override {
+    return mTransceivers;
+  }
 
   // Non-const so it can set mLastError
   nsresult CreateGenericSDP(UniquePtr<Sdp>* sdp);
@@ -169,8 +201,6 @@ class JsepSessionImpl : public JsepSession {
                         SdpMediaSection* msection);
   uint16_t GetNeverUsedExtmapEntry();
   nsresult SetupIds();
-  void SetupDefaultCodecs();
-  void SetupDefaultRtpExtensions();
   void SetState(JsepSignalingState state);
   // Non-const so it can set mLastError
   nsresult ParseSdp(const std::string& sdp, UniquePtr<Sdp>* parsedp);
@@ -183,16 +213,17 @@ class JsepSessionImpl : public JsepSession {
   nsresult ValidateOffer(const Sdp& offer);
   nsresult ValidateAnswer(const Sdp& offer, const Sdp& answer);
   nsresult UpdateTransceiversFromRemoteDescription(const Sdp& remote);
-  JsepTransceiver* GetTransceiverForLevel(size_t level) const;
-  JsepTransceiver* GetTransceiverForMid(const std::string& mid) const;
-  JsepTransceiver* GetTransceiverForLocal(size_t level);
-  JsepTransceiver* GetTransceiverForRemote(const SdpMediaSection& msection);
-  JsepTransceiver* GetTransceiverWithTransport(
+  Maybe<JsepTransceiver> GetTransceiverForLevel(size_t level) const;
+  Maybe<JsepTransceiver> GetTransceiverForMid(const std::string& mid) const;
+  Maybe<JsepTransceiver> GetTransceiverForLocal(size_t level);
+  Maybe<JsepTransceiver> GetTransceiverForRemote(
+      const SdpMediaSection& msection);
+  Maybe<JsepTransceiver> GetTransceiverWithTransport(
       const std::string& transportId) const;
   // The w3c and IETF specs have a lot of "magical" behavior that happens when
   // addTrack is used. This was a deliberate design choice. Sadface.
-  JsepTransceiver* FindUnassociatedTransceiver(SdpMediaSection::MediaType type,
-                                               bool magic);
+  Maybe<JsepTransceiver> FindUnassociatedTransceiver(
+      SdpMediaSection::MediaType type, bool magic);
   // Called for rollback of local description
   void RollbackLocalOffer();
   // Called for rollback of remote description
@@ -216,19 +247,17 @@ class JsepSessionImpl : public JsepSession {
                                       SdpSetupAttribute::Role* rolep);
   nsresult MakeNegotiatedTransceiver(const SdpMediaSection& remote,
                                      const SdpMediaSection& local,
-                                     JsepTransceiver* transceiverOut);
+                                     JsepTransceiver& transceiverOut);
   void EnsureHasOwnTransport(const SdpMediaSection& msection,
-                             JsepTransceiver* transceiver);
+                             JsepTransceiver& transceiver);
   void CopyBundleTransports();
 
   nsresult FinalizeTransport(const SdpAttributeList& remote,
                              const SdpAttributeList& answer,
-                             JsepTransport* transport);
+                             JsepTransport* transport) const;
 
   nsresult GetNegotiatedBundledMids(SdpHelper::BundledMids* bundledMids);
 
-  nsresult EnableOfferMsection(SdpMediaSection* msection);
-
   mozilla::Sdp* GetParsedLocalDescription(
       JsepDescriptionPendingOrCurrent type) const;
   mozilla::Sdp* GetParsedRemoteDescription(
@@ -236,38 +265,9 @@ class JsepSessionImpl : public JsepSession {
   const Sdp* GetAnswer() const;
   void SetIceRestarting(bool restarting);
 
-  // !!!NOT INDEXED BY LEVEL!!! The level mapping is done with
-  // JsepTransceiver::mLevel. The keys are opaque, stable identifiers that are
-  // unique within the JsepSession.
-  std::map<size_t, RefPtr<JsepTransceiver>> mTransceivers;
-  // So we can rollback. Not as simple as just going back to the old, though...
-  std::map<size_t, RefPtr<JsepTransceiver>> mOldTransceivers;
+  void InitTransceiver(JsepTransceiver& aTransceiver);
 
-  Maybe<bool> mIsPendingOfferer;
-  Maybe<bool> mIsCurrentOfferer;
-  bool mIceControlling;
-  std::string mIceUfrag;
-  std::string mIcePwd;
-  std::string mOldIceUfrag;
-  std::string mOldIcePwd;
-  bool mRemoteIsIceLite;
-  std::vector<std::string> mIceOptions;
-  JsepBundlePolicy mBundlePolicy;
-  std::vector<JsepDtlsFingerprint> mDtlsFingerprints;
-  uint64_t mSessionId;
-  uint64_t mSessionVersion;
-  size_t mMidCounter;
-  std::set<std::string> mUsedMids;
-  size_t mTransportIdCounter;
-  size_t mTransceiverIdCounter = 0;
-  std::vector<JsepExtmapMediaType> mRtpExtensions;
-  std::set<uint16_t> mExtmapEntriesEverUsed;
   UniquePtr<JsepUuidGenerator> mUuidGen;
-  std::string mDefaultRemoteStreamId;
-  std::string mCNAME;
-  // Used to prevent duplicate local SSRCs. Not used to prevent local/remote or
-  // remote-only duplication, which will be important for EKT but not now.
-  std::set<uint32_t> mSsrcs;
   UniquePtr<Sdp> mGeneratedOffer;   // Created but not set.
   UniquePtr<Sdp> mGeneratedAnswer;  // Created but not set.
   UniquePtr<Sdp> mCurrentLocalDescription;
@@ -275,12 +275,8 @@ class JsepSessionImpl : public JsepSession {
   UniquePtr<Sdp> mPendingLocalDescription;
   UniquePtr<Sdp> mPendingRemoteDescription;
   std::vector<UniquePtr<JsepCodecDescription>> mSupportedCodecs;
-  std::string mLastError;
-  std::vector<std::pair<size_t, std::string>> mLastSdpParsingErrors;
   SdpHelper mSdpHelper;
   UniquePtr<SdpParser> mParser;
-  SsrcGenerator mSsrcGenerator;
-  bool mEncodeTrackId;
 };
 
 }  // namespace mozilla
