@@ -6,8 +6,12 @@
 
 #include "ServiceWorkerScriptCache.h"
 
-#include "js/Array.h"  // JS::GetArrayLength
+#include "js/Array.h"               // JS::GetArrayLength
+#include "js/PropertyAndElement.h"  // JS_GetElement
+#include "js/Utility.h"             // JS::FreePolicy
+#include "mozilla/TaskQueue.h"
 #include "mozilla/Unused.h"
+#include "mozilla/UniquePtr.h"
 #include "mozilla/dom/CacheBinding.h"
 #include "mozilla/dom/cache/CacheStorage.h"
 #include "mozilla/dom/cache/Cache.h"
@@ -39,10 +43,7 @@ using mozilla::dom::cache::Cache;
 using mozilla::dom::cache::CacheStorage;
 using mozilla::ipc::PrincipalInfo;
 
-namespace mozilla {
-namespace dom {
-
-namespace serviceWorkerScriptCache {
+namespace mozilla::dom::serviceWorkerScriptCache {
 
 namespace {
 
@@ -113,7 +114,7 @@ class CompareNetwork final : public nsIStreamLoaderObserver,
     MOZ_ASSERT(NS_IsMainThread());
   }
 
-  nsresult Initialize(nsIPrincipal* aPrincipal, const nsAString& aURL,
+  nsresult Initialize(nsIPrincipal* aPrincipal, const nsACString& aURL,
                       Cache* const aCache);
 
   void Abort();
@@ -122,7 +123,7 @@ class CompareNetwork final : public nsIStreamLoaderObserver,
 
   void CacheFinish(nsresult aRv);
 
-  const nsString& URL() const {
+  const nsCString& URL() const {
     MOZ_ASSERT(NS_IsMainThread());
     return mURL;
   }
@@ -163,7 +164,7 @@ class CompareNetwork final : public nsIStreamLoaderObserver,
 
   nsCOMPtr<nsIChannel> mChannel;
   nsString mBuffer;
-  nsString mURL;
+  nsCString mURL;
   ChannelInfo mChannelInfo;
   RefPtr<InternalHeaders> mInternalHeaders;
   UniquePtr<PrincipalInfo> mPrincipalInfo;
@@ -203,17 +204,17 @@ class CompareCache final : public PromiseNativeHandler,
     MOZ_ASSERT(NS_IsMainThread());
   }
 
-  nsresult Initialize(Cache* const aCache, const nsAString& aURL);
+  nsresult Initialize(Cache* const aCache, const nsACString& aURL);
 
   void Finish(nsresult aStatus, bool aInCache);
 
   void Abort();
 
-  virtual void ResolvedCallback(JSContext* aCx,
-                                JS::Handle<JS::Value> aValue) override;
+  virtual void ResolvedCallback(JSContext* aCx, JS::Handle<JS::Value> aValue,
+                                ErrorResult& aRv) override;
 
-  virtual void RejectedCallback(JSContext* aCx,
-                                JS::Handle<JS::Value> aValue) override;
+  virtual void RejectedCallback(JSContext* aCx, JS::Handle<JS::Value> aValue,
+                                ErrorResult& aRv) override;
 
   const nsString& Buffer() const {
     MOZ_ASSERT(NS_IsMainThread());
@@ -230,7 +231,7 @@ class CompareCache final : public PromiseNativeHandler,
   RefPtr<CompareNetwork> mCN;
   nsCOMPtr<nsIInputStreamPump> mPump;
 
-  nsString mURL;
+  nsCString mURL;
   nsString mBuffer;
 
   enum {
@@ -261,12 +262,14 @@ class CompareManager final : public PromiseNativeHandler {
     MOZ_ASSERT(aRegistration);
   }
 
-  nsresult Initialize(nsIPrincipal* aPrincipal, const nsAString& aURL,
+  nsresult Initialize(nsIPrincipal* aPrincipal, const nsACString& aURL,
                       const nsAString& aCacheName);
 
-  void ResolvedCallback(JSContext* aCx, JS::Handle<JS::Value> aValue) override;
+  void ResolvedCallback(JSContext* aCx, JS::Handle<JS::Value> aValue,
+                        ErrorResult& aRv) override;
 
-  void RejectedCallback(JSContext* aCx, JS::Handle<JS::Value> aValue) override;
+  void RejectedCallback(JSContext* aCx, JS::Handle<JS::Value> aValue,
+                        ErrorResult& aRv) override;
 
   CacheStorage* CacheStorage_() {
     MOZ_ASSERT(NS_IsMainThread());
@@ -323,7 +326,7 @@ class CompareManager final : public PromiseNativeHandler {
 
   void Cleanup();
 
-  nsresult FetchScript(const nsAString& aURL, bool aIsMainScript,
+  nsresult FetchScript(const nsACString& aURL, bool aIsMainScript,
                        Cache* const aCache = nullptr) {
     MOZ_ASSERT(NS_IsMainThread());
 
@@ -361,7 +364,7 @@ class CompareManager final : public PromiseNativeHandler {
       return;
     }
 
-    Optional<RequestOrUSVString> request;
+    Optional<RequestOrUTF8String> request;
     CacheQueryOptions options;
     ErrorResult error;
     RefPtr<Promise> promise = mOldCache->Keys(aCx, request, options, error);
@@ -404,7 +407,7 @@ class CompareManager final : public PromiseNativeHandler {
     mState = WaitingForScriptOrComparisonResult;
 
     bool hasMainScript = false;
-    AutoTArray<nsString, 8> urlList;
+    AutoTArray<nsCString, 8> urlList;
 
     // Extract the list of URLs in the old cache.
     for (uint32_t i = 0; i < len; ++i) {
@@ -420,14 +423,14 @@ class CompareManager final : public PromiseNativeHandler {
         return;
       };
 
-      nsString url;
+      nsCString url;
       request->GetUrl(url);
 
       if (!hasMainScript && url == mURL) {
         hasMainScript = true;
       }
 
-      urlList.AppendElement(url);
+      urlList.AppendElement(std::move(url));
     }
 
     // If the main script is missing, then something has gone wrong.  We
@@ -545,7 +548,8 @@ class CompareManager final : public PromiseNativeHandler {
       return rv;
     }
 
-    RefPtr<InternalResponse> ir = new InternalResponse(200, "OK"_ns);
+    SafeRefPtr<InternalResponse> ir =
+        MakeSafeRefPtr<InternalResponse>(200, "OK"_ns);
     ir->SetBody(body, aCN->Buffer().Length());
     ir->SetURLList(aCN->URLList());
 
@@ -559,10 +563,10 @@ class CompareManager final : public PromiseNativeHandler {
     ir->Headers()->Fill(*(internalHeaders.get()), IgnoreErrors());
 
     RefPtr<Response> response =
-        new Response(aCache->GetGlobalObject(), ir, nullptr);
+        new Response(aCache->GetGlobalObject(), std::move(ir), nullptr);
 
-    RequestOrUSVString request;
-    request.SetAsUSVString().ShareOrDependUpon(aCN->URL());
+    RequestOrUTF8String request;
+    request.SetAsUTF8String().ShareOrDependUpon(aCN->URL());
 
     // For now we have to wait until the Put Promise is fulfilled before we can
     // continue since Cache does not yet support starting a read that is being
@@ -588,7 +592,7 @@ class CompareManager final : public PromiseNativeHandler {
 
   nsTArray<RefPtr<CompareNetwork>> mCNList;
 
-  nsString mURL;
+  nsCString mURL;
   RefPtr<nsIPrincipal> mPrincipal;
 
   // Used for the old cache where saves the old source scripts.
@@ -618,7 +622,7 @@ class CompareManager final : public PromiseNativeHandler {
 NS_IMPL_ISUPPORTS0(CompareManager)
 
 nsresult CompareNetwork::Initialize(nsIPrincipal* aPrincipal,
-                                    const nsAString& aURL,
+                                    const nsACString& aURL,
                                     Cache* const aCache) {
   MOZ_ASSERT(aPrincipal);
   MOZ_ASSERT(NS_IsMainThread());
@@ -630,7 +634,7 @@ nsresult CompareNetwork::Initialize(nsIPrincipal* aPrincipal,
   }
 
   mURL = aURL;
-  mURLList.AppendElement(NS_ConvertUTF16toUTF8(mURL));
+  mURLList.AppendElement(mURL);
 
   nsCOMPtr<nsILoadGroup> loadGroup;
   rv = NS_NewLoadGroup(getter_AddRefs(loadGroup), aPrincipal);
@@ -665,7 +669,18 @@ nsresult CompareNetwork::Initialize(nsIPrincipal* aPrincipal,
   nsCOMPtr<nsICookieJarSettings> cookieJarSettings =
       mozilla::net::CookieJarSettings::Create(aPrincipal);
 
-  net::CookieJarSettings::Cast(cookieJarSettings)->SetPartitionKey(uri);
+  // Populate the partitionKey by using the given prinicpal. The ServiceWorkers
+  // are using the foreign partitioned principal, so we can get the partitionKey
+  // from it and the partitionKey will only exist if it's in the third-party
+  // context. In first-party context, we can still use the uri to set the
+  // partitionKey.
+  if (!aPrincipal->OriginAttributesRef().mPartitionKey.IsEmpty()) {
+    net::CookieJarSettings::Cast(cookieJarSettings)
+        ->SetPartitionKey(aPrincipal->OriginAttributesRef().mPartitionKey);
+  } else {
+    net::CookieJarSettings::Cast(cookieJarSettings)
+        ->SetPartitionKey(uri, false);
+  }
 
   // Note that because there is no "serviceworker" RequestContext type, we can
   // use the TYPE_INTERNAL_SCRIPT content policy types when loading a service
@@ -678,6 +693,16 @@ nsresult CompareNetwork::Initialize(nsIPrincipal* aPrincipal,
     return rv;
   }
 
+  // Set the IsInThirdPartyContext for the channel's loadInfo according to the
+  // partitionKey of the principal. The worker is foreign if it's using
+  // partitioned principal, i.e. the partitionKey is not empty. In this case,
+  // we need to set the bit to the channel's loadInfo.
+  if (!aPrincipal->OriginAttributesRef().mPartitionKey.IsEmpty()) {
+    nsCOMPtr<nsILoadInfo> loadInfo = mChannel->LoadInfo();
+    rv = loadInfo->SetIsInThirdPartyContext(true);
+    MOZ_ASSERT(NS_SUCCEEDED(rv));
+  }
+
   nsCOMPtr<nsIHttpChannel> httpChannel = do_QueryInterface(mChannel);
   if (httpChannel) {
     // Spec says no redirects allowed for top-level SW scripts.
@@ -788,7 +813,7 @@ void CompareNetwork::Abort() {
     mState = Finished;
 
     MOZ_ASSERT(mChannel);
-    mChannel->Cancel(NS_BINDING_ABORTED);
+    mChannel->CancelWithReason(NS_BINDING_ABORTED, "CompareNetwork::Abort"_ns);
     mChannel = nullptr;
 
     if (mCC) {
@@ -920,7 +945,7 @@ CompareNetwork::OnStreamComplete(nsIStreamLoader* aLoader,
       mURLList.AppendElement(channelURLSpec);
     }
 
-    char16_t* buffer = nullptr;
+    UniquePtr<char16_t[], JS::FreePolicy> buffer;
     size_t len = 0;
 
     rv = ScriptLoader::ConvertToUTF16(channel, aString, aLen, u"UTF-8"_ns,
@@ -929,7 +954,7 @@ CompareNetwork::OnStreamComplete(nsIStreamLoader* aLoader,
       return rv;
     }
 
-    mBuffer.Adopt(buffer, len);
+    mBuffer.Adopt(buffer.release(), len);
 
     rv = NS_OK;
     return NS_OK;
@@ -990,7 +1015,7 @@ CompareNetwork::OnStreamComplete(nsIStreamLoader* aLoader,
     // (in that case the originalURL is the resolved jar URI and so we have to
     // look to the channel principal instead).
     if (channelPrincipal->SchemeIs("moz-extension")) {
-      char16_t* buffer = nullptr;
+      UniquePtr<char16_t[], JS::FreePolicy> buffer;
       size_t len = 0;
 
       rv = ScriptLoader::ConvertToUTF16(channel, aString, aLen, u"UTF-8"_ns,
@@ -999,7 +1024,7 @@ CompareNetwork::OnStreamComplete(nsIStreamLoader* aLoader,
         return rv;
       }
 
-      mBuffer.Adopt(buffer, len);
+      mBuffer.Adopt(buffer.release(), len);
 
       return NS_OK;
     }
@@ -1030,7 +1055,7 @@ CompareNetwork::OnStreamComplete(nsIStreamLoader* aLoader,
     ServiceWorkerManager::LocalizeAndReportToAllClients(
         mRegistration->Scope(), "ServiceWorkerRegisterNetworkError",
         nsTArray<nsString>{NS_ConvertUTF8toUTF16(mRegistration->Scope()),
-                           statusAsText, mURL});
+                           statusAsText, NS_ConvertUTF8toUTF16(mURL)});
 
     rv = NS_ERROR_FAILURE;
     return NS_OK;
@@ -1063,7 +1088,8 @@ CompareNetwork::OnStreamComplete(nsIStreamLoader* aLoader,
     ServiceWorkerManager::LocalizeAndReportToAllClients(
         mRegistration->Scope(), "ServiceWorkerRegisterMimeTypeError2",
         nsTArray<nsString>{NS_ConvertUTF8toUTF16(mRegistration->Scope()),
-                           NS_ConvertUTF8toUTF16(mimeType), mURL});
+                           NS_ConvertUTF8toUTF16(mimeType),
+                           NS_ConvertUTF8toUTF16(mURL)});
     rv = NS_ERROR_DOM_SECURITY_ERR;
     return rv;
   }
@@ -1085,7 +1111,7 @@ CompareNetwork::OnStreamComplete(nsIStreamLoader* aLoader,
     mURLList.AppendElement(channelURLSpec);
   }
 
-  char16_t* buffer = nullptr;
+  UniquePtr<char16_t[], JS::FreePolicy> buffer;
   size_t len = 0;
 
   rv = ScriptLoader::ConvertToUTF16(httpChannel, aString, aLen, u"UTF-8"_ns,
@@ -1094,13 +1120,13 @@ CompareNetwork::OnStreamComplete(nsIStreamLoader* aLoader,
     return rv;
   }
 
-  mBuffer.Adopt(buffer, len);
+  mBuffer.Adopt(buffer.release(), len);
 
   rv = NS_OK;
   return NS_OK;
 }
 
-nsresult CompareCache::Initialize(Cache* const aCache, const nsAString& aURL) {
+nsresult CompareCache::Initialize(Cache* const aCache, const nsACString& aURL) {
   MOZ_ASSERT(NS_IsMainThread());
   MOZ_ASSERT(aCache);
   MOZ_DIAGNOSTIC_ASSERT(mState == WaitingForInitialization);
@@ -1110,8 +1136,8 @@ nsresult CompareCache::Initialize(Cache* const aCache, const nsAString& aURL) {
   AutoJSAPI jsapi;
   jsapi.Init();
 
-  RequestOrUSVString request;
-  request.SetAsUSVString().ShareOrDependUpon(aURL);
+  RequestOrUTF8String request;
+  request.SetAsUTF8String().ShareOrDependUpon(aURL);
   ErrorResult error;
   CacheQueryOptions params;
   RefPtr<Promise> promise = aCache->Match(jsapi.cx(), request, params, error);
@@ -1143,7 +1169,7 @@ void CompareCache::Abort() {
     mState = Finished;
 
     if (mPump) {
-      mPump->Cancel(NS_BINDING_ABORTED);
+      mPump->CancelWithReason(NS_BINDING_ABORTED, "CompareCache::Abort"_ns);
       mPump = nullptr;
     }
   }
@@ -1164,7 +1190,7 @@ CompareCache::OnStreamComplete(nsIStreamLoader* aLoader, nsISupports* aContext,
     return aStatus;
   }
 
-  char16_t* buffer = nullptr;
+  UniquePtr<char16_t[], JS::FreePolicy> buffer;
   size_t len = 0;
 
   nsresult rv = ScriptLoader::ConvertToUTF16(nullptr, aString, aLen,
@@ -1174,14 +1200,15 @@ CompareCache::OnStreamComplete(nsIStreamLoader* aLoader, nsISupports* aContext,
     return rv;
   }
 
-  mBuffer.Adopt(buffer, len);
+  mBuffer.Adopt(buffer.release(), len);
 
   Finish(NS_OK, true);
   return NS_OK;
 }
 
 void CompareCache::ResolvedCallback(JSContext* aCx,
-                                    JS::Handle<JS::Value> aValue) {
+                                    JS::Handle<JS::Value> aValue,
+                                    ErrorResult& aRv) {
   MOZ_ASSERT(NS_IsMainThread());
 
   switch (mState) {
@@ -1196,7 +1223,8 @@ void CompareCache::ResolvedCallback(JSContext* aCx,
 }
 
 void CompareCache::RejectedCallback(JSContext* aCx,
-                                    JS::Handle<JS::Value> aValue) {
+                                    JS::Handle<JS::Value> aValue,
+                                    ErrorResult& aRv) {
   MOZ_ASSERT(NS_IsMainThread());
 
   if (mState != Finished) {
@@ -1265,7 +1293,9 @@ void CompareCache::ManageValueResult(JSContext* aCx,
   if (rr) {
     nsCOMPtr<nsIEventTarget> sts =
         do_GetService(NS_STREAMTRANSPORTSERVICE_CONTRACTID);
-    rv = rr->RetargetDeliveryTo(sts);
+    RefPtr<TaskQueue> queue =
+        TaskQueue::Create(sts.forget(), "CompareCache STS Delivery Queue");
+    rv = rr->RetargetDeliveryTo(queue);
     if (NS_WARN_IF(NS_FAILED(rv))) {
       mPump = nullptr;
       Finish(rv, false);
@@ -1275,7 +1305,7 @@ void CompareCache::ManageValueResult(JSContext* aCx,
 }
 
 nsresult CompareManager::Initialize(nsIPrincipal* aPrincipal,
-                                    const nsAString& aURL,
+                                    const nsACString& aURL,
                                     const nsAString& aCacheName) {
   MOZ_ASSERT(NS_IsMainThread());
   MOZ_ASSERT(aPrincipal);
@@ -1332,7 +1362,8 @@ nsresult CompareManager::Initialize(nsIPrincipal* aPrincipal,
 // 4. Put the value in the cache.
 // For this reason we have mState to know what callback we are handling.
 void CompareManager::ResolvedCallback(JSContext* aCx,
-                                      JS::Handle<JS::Value> aValue) {
+                                      JS::Handle<JS::Value> aValue,
+                                      ErrorResult& aRv) {
   MOZ_ASSERT(NS_IsMainThread());
   MOZ_ASSERT(mCallback);
 
@@ -1362,7 +1393,8 @@ void CompareManager::ResolvedCallback(JSContext* aCx,
 }
 
 void CompareManager::RejectedCallback(JSContext* aCx,
-                                      JS::Handle<JS::Value> aValue) {
+                                      JS::Handle<JS::Value> aValue,
+                                      ErrorResult& aRv) {
   MOZ_ASSERT(NS_IsMainThread());
   switch (mState) {
     case Finished:
@@ -1410,23 +1442,6 @@ void CompareManager::Cleanup() {
   }
 }
 
-class NoopPromiseHandler final : public PromiseNativeHandler {
- public:
-  NS_DECL_ISUPPORTS
-
-  NoopPromiseHandler() { AssertIsOnMainThread(); }
-
-  void ResolvedCallback(JSContext* aCx, JS::Handle<JS::Value> aValue) override {
-  }
-  void RejectedCallback(JSContext* aCx, JS::Handle<JS::Value> aValue) override {
-  }
-
- private:
-  ~NoopPromiseHandler() { AssertIsOnMainThread(); }
-};
-
-NS_IMPL_ISUPPORTS0(NoopPromiseHandler)
-
 }  // namespace
 
 nsresult PurgeCache(nsIPrincipal* aPrincipal, const nsAString& aCacheName) {
@@ -1452,10 +1467,9 @@ nsresult PurgeCache(nsIPrincipal* aPrincipal, const nsAString& aCacheName) {
     return rv.StealNSResult();
   }
 
-  // Add a no-op promise handler to ensure that if this promise gets rejected,
+  // Set [[PromiseIsHandled]] to ensure that if this promise gets rejected,
   // we don't end up reporting a rejected promise to the console.
-  RefPtr<NoopPromiseHandler> promiseHandler = new NoopPromiseHandler();
-  promise->AppendNativeHandler(promiseHandler);
+  MOZ_ALWAYS_TRUE(promise->SetAnyPromiseIsHandled());
 
   // We don't actually care about the result of the delete operation.
   return NS_OK;
@@ -1486,7 +1500,7 @@ nsresult GenerateCacheName(nsAString& aName) {
 
 nsresult Compare(ServiceWorkerRegistrationInfo* aRegistration,
                  nsIPrincipal* aPrincipal, const nsAString& aCacheName,
-                 const nsAString& aURL, CompareCallback* aCallback) {
+                 const nsACString& aURL, CompareCallback* aCallback) {
   MOZ_ASSERT(NS_IsMainThread());
   MOZ_ASSERT(aRegistration);
   MOZ_ASSERT(aPrincipal);
@@ -1503,7 +1517,4 @@ nsresult Compare(ServiceWorkerRegistrationInfo* aRegistration,
   return NS_OK;
 }
 
-}  // namespace serviceWorkerScriptCache
-
-}  // namespace dom
-}  // namespace mozilla
+}  // namespace mozilla::dom::serviceWorkerScriptCache
