@@ -5,11 +5,10 @@
 Apply some defaults and minor modifications to the pgo jobs.
 """
 
-
-from gecko_taskgraph.transforms.base import TransformSequence
-
 import logging
 
+from taskgraph.transforms.base import TransformSequence
+
 logger = logging.getLogger(__name__)
 
 transforms = TransformSequence()
@@ -29,7 +28,7 @@ def run_profile_data(config, jobs):
         else:
             artifact = "target.tar.bz2"
         job.setdefault("fetches", {})[instr] = [
-            artifact,
+            {"artifact": artifact, "extract": not artifact.endswith((".dmg", ".apk"))},
             "target.crashreporter-symbols.zip",
         ]
         yield job
