@@ -122,15 +122,12 @@ UNITTEST_PLATFORM_PRETTY_NAMES = {
         "linux1804-64-asan",
     ],
     "x64": ["linux64", "linux64-asan", "linux1804-64", "linux1804-64-asan"],
-    "Android 7.0 Moto G5 32bit": ["android-hw-g5-7.0-arm7"],
     "Android 7.0 Samsung A51 32bit": ["android-hw-a51-11.0-arm7"],
     "Android 7.0 Samsung A51 64bit": ["android-hw-a51-11.0-aarch64"],
-    "Android 8.0 Google Pixel 2 32bit": ["android-hw-p2-8.0-arm7"],
-    "Android 8.0 Google Pixel 2 64bit": ["android-hw-p2-8.0-android-aarch64"],
     "Android 13.0 Google Pixel 5 32bit": ["android-hw-p5-13.0-arm7"],
     "Android 13.0 Google Pixel 5 64bit": ["android-hw-p5-13.0-android-aarch64"],
-    "Windows 7": ["windows7-32"],
-    "Windows 7 VM": ["windows7-32-vm"],
+    "Android 13.0 Google Pixel 6 64bit": ["android-hw-p6-13.0-android-aarch64"],
+    "Android 13.0 Samsung S21 64bit": ["android-hw-s21-13.0-android-aarch64"],
     "Windows 10": ["windows10-64"],
 }
 
@@ -532,10 +529,8 @@ class TryOptionSyntax:
         # This might be somewhat confusing but we parse the string _backwards_ so
         # there is no ambiguity over what state we are in.
         for char in reversed(input_str):
-
             # , indicates exiting a state
             if char == ",":
-
                 # Exit a particular platform.
                 if in_platforms:
                     add_platform(token)
